/*
table - convert symbol stroke table into code for symbol.c
*/

static const char rcsid_table_c[] = "$Id: table.c,v 1.4 2004/05/17 17:21:10 ccrawford Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define MCHAR 256		/* maximum number of characters */
#define MVERTEX 2048		/* maximum number of vertices */

static int line=0;		/* line number in file */

static void
ap(name,a,n,t)
	char *name;
	int *a;	
	int n;
	int t;
{
	int i;
	printf("static %s %s[%d] = {\n",
		(t)? "unsigned char" : "short int",
		name,n);

	for(i=0;i<n;i++){
		printf("%5u",a[i]);
		if(i != n-1){
			putchar(',');
			if(!((i+1)%8)) putchar('\n');
		}
	}
	printf("};\n");
}

static int
hex(c)
	char c;
{
	if(c >= '0' && c <= '9'){
		return(c - '0');
	}else if(c >= 'a' && c <= 'f'){
		return(c - 'a' + 10);
	}else if(c >= 'A' && c <= 'F'){
		return(c - 'A' + 10);
	}else{
		fprintf(stderr,"table: unexpected charcacter '%c'. line %d\n",
			c,line);
		exit(1);
	}
	return(-1);	/* can't get here; make compiler happier */
}

int main()
{
	int nxny[MVERTEX];		/* vertices of characters */
	int charp[MCHAR];		/* pointers to nxny array */
	int charc[MCHAR];		/* count in nxny array */
	int n = 0;			/* counter for charp arrary */
	int np = 0;			/* counter for nxny array */
	char s[256],*sp;		/* input buffer and pointer */
	int index;			/* start of char in nxny */
	int count;			/* # nodes for character */
	int nx,ny;			/* x,y locations of node */
	char *tableid=NULL;		/* rcs id from input table */

	while(line++ , fgets(s,256,stdin) != NULL){
		if(!strncmp(s,"$Id:",4)){
			tableid = malloc(strlen(s));
			if(tableid == NULL){
				fputs("table: malloc failure\n",stderr);
				exit(1);
			}
			s[strlen(s)-1] = 0;
			strcpy(tableid,s);
			continue;
		}
		if(*s && *s != '#' && *s != ' ' && *s != '\t' && *s != '\n'){
			index = np;
			sp = s;
			count = 0;
			while(isxdigit(*sp)){
				nx = hex(*sp++);
				ny = hex(*sp++);
				if(np == MVERTEX){
					fputs("table: too many nodes\n",stderr);
					exit(1);
				}
				nxny[np++] = (nx << 4) | ny;
				count++;
			}
			if(*sp != '\t' && *sp != ' ' && *sp != 0){
				fprintf(stderr,"table: syntax error. line %d\n",
					line);
				exit(1);
			}
			if(n == MCHAR){
				fputs("table: too many characters\n",stderr);
				exit(1);
			}
			charp[n] = index; 
			charc[n] = count;
			n++;
		}
	}

	printf("/* symbol.h - automatically generated by table.c */\n");
	printf("static const char rcsid_table_c[] =\n\"%s\";\n",
		rcsid_table_c);
	if(tableid != NULL){
		printf("static const char rcsid_table_as[] =\n\"%s\";\n",
			tableid);
	}
	printf("#define MCHAR %d\n",n);
	printf("#define MNXNY %d\n",np);
	ap("charp",charp,n,0);
	ap("charc",charc,n,0);
	ap("nxny",nxny,np,1);
	exit(0);
}

