/*
accept - accept (read) objects from stdin or units

parse_aci - accept an int from stdin
parse_acimm - accept an int from stdin with min/max checking
parse_acf - accept a float from stdin
parse_acfmm - accept a float from stdin with min/max checking
parse_acs - accept a string or filename from stdin
parse_act - accpet a toggle (yes/no) from stdin
parse_acm - accpet a mode (one of string) from stdin
*/

#ident "$Id: accept.c,v 1.6 2007/05/22 15:51:51 ccrawford Exp $"

#include <string.h>
#include <stdlib.h>
#include "parse.h"

void 
parse_aci (char *message, int d, int *i)
{
	char s[PARSE_MAX_LINE];

	parse_atof_args(stderr,1);

	while(1){
		printf(">> %s [%d]: ",message,d);
		if(parse_urline(stdin,s)) exit(1);
		if(strlen(s)){
			*i = (int) parse_atof(s);
			if(!parse_atof_error()) break;
		}else{
			*i = d;
			break;
		}
		printf("---------");
	}
}

void 
parse_acimm (char *message, int amin, int amax, int d, int *i)
{
	char s[PARSE_MAX_LINE];

	parse_atof_args(stderr,1);

	while(1){
		printf(">> %s <%d,%d>[%d]: ",message,amin,amax,d);
		if(parse_urline(stdin,s)) exit(1);
		if(strlen(s)){
			*i = (int) parse_atof(s);
			if(!parse_atof_error()&& *i <= amax && *i >= amin)break;
		}else{
			*i = d;
			if(*i <= amax && *i >= amin)break;
		}
		printf("---------");
	}
}

void 
parse_acf (char *message, double d, float *i)
{
	char s[PARSE_MAX_LINE];

	parse_atof_args(stderr,1);

	while(1){
		printf(">> %s [%f]: ",message,d);
		if(parse_urline(stdin,s)) exit(1);
		if(strlen(s)){
			*i = parse_atof(s);
			if(!parse_atof_error()) break;
		}else{
			*i = d;
			break;
		}
		printf("---------");
	}
}

void 
parse_acfmm (char *message, double amin, double amax, double d, float *i)
{
	char s[PARSE_MAX_LINE];

	parse_atof_args(stderr,1);

	while(1){
		printf(">> %s <%f,%f>[%f]: ",message,amin,amax,d);
		if(parse_urline(stdin,s)) exit(1);
		if(strlen(s)){
			*i = parse_atof(s);
			if(!parse_atof_error() && *i <= amax && *i>= amin)break;
		}else{
			*i = d;
			if(*i <= amax && *i >= amin)break;
		}
		printf("---------");
	}
}

char *
parse_acs (char *message, char *d, char *fe, char *i)
{
	char s[PARSE_MAX_LINE];
	char *r;

	printf(">> %s",message);
	if(d != NULL && strlen(d)) printf(" [%s",d);
	if(fe != NULL && strlen(fe)){
		if(!strlen(d))puts(" ");
		printf("[.%s]",fe);
	}
	if(d != NULL && strlen(d)) fputs("]",stdout);
	fputs(": ",stdout);

        if(parse_urline(stdin,s)) exit(1);
	if(!strlen(s) && d == NULL){
		*s = 0;
	}else{
		strncpy(s,strlen(s)?s:d,PARSE_MAX_LINE);
	}

	if(i != NULL){
		r = i;
	}else{
		r = parse_malloc_char(strlen(s)+1);
	}
	strcpy(r,s);
	return(r);
}

void 
parse_act (char *message, int d, int *i)
{
	char s[PARSE_MAX_LINE];

	printf(">> %s [%s]: ",message,d?"yes":"no");
        if(parse_urline(stdin,s)) exit(1);

	*i = d;
	if(strlen(s)){
		if(*s == 'n' || *s == 'N') *i = 0;
		if(*s == 'y' || *s == 'Y') *i = 1;
	}
}

void 
parse_acm (char *message, int d, char *modes, int *i)
{
	char *c,s[PARSE_MAX_LINE];

	if(d < 0 || d >= (int) strlen(modes)){
		parse_pri("acm error: illegal default mode",d);
		exit(1);
	}

	while(1){
		printf(">> %s [%c]: ",message,modes[d]);
		if(parse_urline(stdin,s)) exit(1);
		if(!strlen(s)){
			*i = d;
			return;
		}
		if(NULL != (c = strchr(modes,*s))){
			*i = c - modes;
			return;
		}
		printf("---------");
	}
}
