/*
parse_atof - atof with expression handling - has fair error detection
parse_atof_args - setup the operational parameters for parse_atof
parse_atof_error - report errors caused during calls to parse_atof
*/

#ident "$Id: atof.c,v 1.6 2009/11/19 13:07:49 carl Exp $"

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include "parse.h"

static int flags_set;		/* 1=next 2 flags set */
static FILE *parse_atof_fd;
static int atof_error;
static int parse_atof_stop_error;
static double e1(char **s),e2(char **s),e3(char **s);

static void 
w (char **s)
{
	while(**s==' '||**s=='\t'||**s=='\n')(*s)++;
}

static double 
p10 (	/* do exp10, don't force user to use -lm */
    int p
)
{
	int i,ap;
	double r;

	ap = abs(p);
	r = 1;
	for(i=0;i<ap;i++)r *= 10;
	return((p>0)? r : 1/r);
}

static void 
q (int t)
{

	if(!flags_set) parse_atof_args(stderr,1);
	if(!atof_error){
		fputs("expression error: ",parse_atof_fd);
		if(t == 1000)
			fputs("missing right parenthesis\n",parse_atof_fd);
		else if(t == 1001)
			fputs("divide by zero\n",parse_atof_fd);
		else if(!isprint(t))
			fputs("unknown syntax\n",parse_atof_fd);
		else
			fprintf(parse_atof_fd,"unexpected character '%c'\n",t);
		atof_error = 1;
		if(!parse_atof_stop_error)exit(1);
	}
}

static double 
e1 (char **s)
{
	double r;
	r = e2(s);
	while(1){
		w(s);
		switch(*(*s)++){
		case '+':
			r += e2(s);
			break;
		case '-':
			r -= e2(s);
			break;
		default:
			(*s)--;
			return r;
		}
	}
}

static double 
e2 (char **s)
{
	double r,d;

	r = e3(s);
	while(1){
		w(s);
		switch(*(*s)++){
		case '*':
			r *= e3(s);
			break;
		case '/':
			d = e3(s);
			if(d == 0.0) q(1001);
			r /= d;
			break;
		default:
			(*s)--;
			return r;
		}
	}
}

static double 
e3 (char **s)
{
	int c,sp,p;
	double frac,n;

	w(s);
	c= *(*s)++;
	if(isdigit(c) || c =='.' || c== 'e' || c=='E'){
		if(c == '.'){
			n = 0.0;
		}else if(c== 'e' || c=='E'){
			n = 1.0;
			(*s)--;
		}else{
			n = c - '0';
			while(isdigit(**s))n=10*n+*(*s)++-'0';
			if('.' == (c = **s)) (*s)++;
		}

		if(c == '.'){
			frac = 10.;
			while(isdigit(**s)){
				n += (*(*s)++-'0')/frac;
				frac *= 10.0;
			}
			c = **s;
		}

		if(c == 'e' || c == 'E'){
			(*s)++;
			if(**s == '-'){
				sp = -1;
				(*s)++;
			}else{
				sp = 1;
			}
			p=0;
			while(isdigit(**s))p=10*p+*(*s)++-'0';
			n *= p10(sp*p);
		}
		return n;
	}else if(c=='-'){
		return -e3(s);
	}else if(c=='+'){
		return e3(s);
	}else if(c=='('){
		double r=e1(s);
		if(*(*s)++!=')'){
			q(1000);
			(*s)--;
		}
		return r;
	}else{
		q(c);
		return 0;
	}
}
double 
parse_atof (char *s)
{
	char *p;
	double r;

	p = s;
	atof_error = 0;
	r = e1(&p);
	if(*p != '\0' && *p != ' ' && *p != '\t')q(*p);
	return(r);
}

int 
parse_atof_error (void)
{
	return(atof_error);
}

void 
parse_atof_args (
    FILE *fd,
    int stop	/* 0=stop on error 1=return, set atof_error */
)
{
	parse_atof_fd = fd;
	parse_atof_stop_error = stop;
	flags_set = 1;
}

#ifdef DEBUG
int 
main (void){
	double parse_atof();
	char b[PARSE_MAX_LINE];
	parse_atof_args(stderr,1);
	while(!parse_rline(b))printf("%f\n",parse_atof(b));
}
#endif
