/*
da - handle da files

parse_vwrite - write vector out as a da file (short or float)
parse_vwrite_float - write float vector to da file
parse_vread - read vector from a da file
parse_wda - write multiple records of float to a da file
parse_rda - read multiple records of float from a da file
*/

#ident "$Id: da.c,v 1.7 2007/05/22 15:51:51 ccrawford Exp $"

#include "parse.h"
#include <string.h>
#include <limits.h>

void 
parse_vwrite (float *data, int n, int type, char *file, char *message)
{
	char *filep;
	int nblk;
	short head[256];
	FILE *u;
	int j,np,sp;
	float *datap;

	if(type)parse_err("parse_vwrite: int not supported:",file);

	nblk = BLOCK_FLOAT(n);
	sp = sizeof(float);

	np = nblk * 512 / sp;
	filep = parse_suffix_malloc(file,"da");
	u = parse_fopen(filep,"wb");
	parse_zbuf(head,256);
	parse_free(filep);
	head[0] = n;
	head[1] = 1;
	head[2] = type;
	if(message != NULL)strcpy(((char *)head) + 48,message);
	parse_to_can_shortv(head,256);
	parse_wblock(u,0,(char *)head,1);

	datap = parse_malloc_float(n);
	for(j=0;j<n;j++) datap[j] = data[j];
	parse_to_can_floatv(datap,n);
	parse_fwrite((char *)datap,sp,n,u);
	parse_free((char *)datap);
	parse_fwrite((char *)head,sp,np-n,u);
	parse_fclose(u);
}

void 
parse_vwrite_float (float *data, int n, char *file)
{
	parse_vwrite(data,n,0,file,"");
}

float *
parse_vread (float *data, int *n, int mn, int type, char *file)
{
	char *filep;
	int m,nr;
	short head[256];
	FILE *u;

	filep = parse_suffix_malloc(file,"da");
	if(type)parse_err("parse_vread: 'int' not supported:",filep);

	u = parse_fopen(filep,"rb");
	parse_rblock(u,0,(char *)head,1);
	parse_from_can_shortv(head,256);
	*n = head[0];
	nr = head[1];
	m = head[2];

	if(m != type)parse_err("parse_vread: type mismatch in: ",filep);
	if(*n < 1)parse_err("parse_vread: data underflow in: ",filep);
	if(nr != 1)parse_err("parse_vread: bad number records in: ",filep);

	if(data == NULL){
		data = parse_malloc_float(*n);
	}else{
		if(*n > mn)parse_err("parse_vread: data overflow in: ",filep);
	}

	parse_fread_block((char *)data,sizeof(float),*n,u,1);
	parse_from_can_floatv(data,*n);
	parse_fclose(u);
	parse_free(filep);
	return(data);
}

void 
parse_wda (float *data, int col, int row, char *file)
{
	int colp,i,j,nblk;
	short head[256];
	FILE *u;
	char *filep;
	float *datap;

	if(col < 0) parse_pri("*parse_wda: column < 0:",col);
	if(row < 0) parse_pri("*parse_wda: row < 0:",row);
	if(col > SHRT_MAX) parse_pri("*parse_wda: column > max short:",col);
	if(row > SHRT_MAX) parse_pri("*parse_wda: row > max short:",row);

	nblk = BLOCK_FLOAT(col);
	colp = nblk * 512 / sizeof(float);

	filep = parse_suffix_malloc(file,"da");
	u = parse_fopen(filep,"wb");
	parse_free(filep);
	parse_zbuf(head,256);
	head[0] = col;
	head[1] = row;
	head[2] = 0;
	parse_to_can_shortv(head,256);
	parse_wblock(u,0,(char *)head,1);

	datap = parse_malloc_float(col);
	for(i=0;i<row;i++){
		for(j=0;j<col;j++) datap[j] = data[j+i*col];
		parse_to_can_floatv(datap,col);
		parse_fwrite((char *)datap,sizeof(float),col,u);
		parse_fwrite((char *)head,sizeof(float),colp-col,u);
	}
	parse_fclose(u);
	parse_free((char *)datap);
}

float *
parse_rda (float *data, int *rcol, int *rrow, char *file)
{
	int colp,i,nblk,col,row;
	short head[256];
	FILE *u;
	char *filep;
	float *dp;

	filep = parse_suffix_malloc(file,"da");
	u = parse_fopen(filep,"rb");

	parse_rblock(u,0,(char *)head,1);
	parse_from_can_shortv(head,256);
	*rcol = col = head[0];
	*rrow = row = head[1];
	if(head[2] != 0)parse_err("parse_rda: can't read int:",filep);

	nblk = BLOCK_FLOAT((col));
	colp = nblk * 512 / sizeof(float);

	if(data == NULL) 
		dp = parse_malloc_float(row * col);
	else
		dp = data;

	for(i=0;i<(row);i++){
		parse_fread((char *)(dp+i*(col)),sizeof(float),col,u);
		parse_fread((char *)head,sizeof(float),colp-(col),u);
	}

	parse_from_can_floatv(dp,col*row);
	parse_fclose(u);
	parse_free(filep);
	return(dp);
}
