/*
parse_file_proc - parse initialization file using external proc's
*/

#ident "$Id: file.c,v 1.8 2007/05/22 15:51:51 ccrawford Exp $"

#include <string.h>
#include <stdlib.h>
#include "parse.h"

#ifndef NT
#include <unistd.h>
#include <sys/param.h>
#endif

#define N 256 		/* max # arguments */

int 
parse_file_proc (
    struct PARSE_STRUCT *options,
    struct PARSE_FLAG_TABLE *flags,
    FILE *fd,
    char *file,
    int (*proc)(),
    int proc_type,		// 0=argc/argv 1=strucutures 
    int home,			// 1=check for files in home and relative to home, 0=open file as is
    int efile			// 1 = check for file open/read errors
)
{
	int r,j,nf,i,argc;
	char *argv[N+1];
	char buf[PARSE_MAX_LINE];
	char *s;
	FILE *dfd;
	int largc;
	char *largv[N];
	char *c;
	char pn[MAXPATHLEN];
	int debug=0;		//1=print debug info

#ifdef NT
	home = 0;
#endif

	if(debug) printf("parse_file_proc: file: %s home: %d efile %d\n",file,home, efile);
	if(home == 1){
		if(*file != '/'){
			nf = 2;
		}else{
			nf = 1;
		}
	}else{
		nf = 1;
	}

	for(j=0;j<nf;j++){
		if(home == 1 && *file != '/'){
			if(j == 0){
				if((s = getenv("HOME")) == NULL){
					j++;
				}else{
					strcpy(buf,s);
					strcat(buf,"/");
					strcat(buf,file);
				}
			}

			if(j == 1){
				if((s = getenv("HOME")) == NULL) return(0);
				if(NULL == getcwd(pn,MAXPATHLEN)) return(0);
				if(!strcmp(pn,s)) return(0);
				strcpy(buf,file);
			}
		}else{
			strcpy(buf,file);
		}

		if((dfd = fopen(buf,"r")) == NULL){
			if(efile){
				fprintf(fd,"Can't open: %s\n",buf);
				if(debug) parse_prs("parse_file_proc: returning on can't open:",buf);
				return(-1);
			}else{
				if(debug) parse_prs("parse_file_proc: ignore can't open:",buf);
			}
		}else{
			if(debug) parse_prs("parse_file_proc: processing:",buf);
			argc = 1;
			while(!parse_urline(dfd,buf)){
				if(NULL != (s = strchr(buf,'#'))) *s = 0;
				if(strlen(buf)){
					if(buf[0]== '/' && buf[1] == '/') break;
					s = buf + strlen(buf) - 1;
					while(1){
						if(*s == ' '){
							*s = 0;
							if(s == buf){
								break;
							}else{
								s--;
							}
						}else{
							break;
						}
					}
					c = parse_malloc_char(strlen(buf) + 1);
					strcpy(c,buf);
					parse_words(c,&largc,largv,N);
					for(i=0;i<largc;i++){
						if(argc > N){
							parse_uwtext(fd,"parse_file_proc: too many args");
							return(-1);
						}else{
							argv[argc++] = largv[i];
						}
					}
				}
			}
			if(argc > 1){
				switch(proc_type){

				case 0:
					(* (int (*)(int,char **)) proc)(argc,argv);
					break;

				case 1:
					if(( r = 
					(* (int (*)
					(
					    struct PARSE_STRUCT *,
					    struct PARSE_FLAG_TABLE *,
					    FILE *,
					    int,
					    char **

					)) proc)
					 (options,flags,fd,argc,argv))
					 != 0) return(r);
					 break;

				default:
					parse_er("parse_file_proc: illegal proc type");
				}
			}
			fclose(dfd);
		}
	}
	return(0);
}
