/*
parse.h - include file for routines that use parse.a
*/


#ifndef _parse_h
#define _parse_h

static const char rcsid_parse_h[] = "$Id: parse.h,v 1.18 2006/03/02 15:59:09 ccrawford Exp $";

#include <stdio.h>
#include <stdlib.h>

/*
	stuff for mi-files 
*/
			/* pointers in 65K header */
#define PARSE_MI_LOC_MAGIC 0	/* location of magic number */
#define PARSE_MI_LOC_N 1	/* location of number of images */
#define PARSE_MI_LOC_FIRST 2	/* location of first block */

#define PARSE_MI_MAGIC_65K 12431	/* magic number for 65K files */

			/* pointers in image header */
#define PARSE_MI_LOC_NCOL 40	/* location of ncol */
#define PARSE_MI_LOC_NROW 41	/* location of nrow */
#define PARSE_MI_LOC_LABEL 0	/* location of label */
#define PARSE_MI_LEN_LABEL 60	/* length of label in bytes */

/*
	stuff for parse_omi
*/
			/* header returned by parse_omi */
typedef struct {
	int type;		/* type of header: 256, 1024, 65K */
	int n;			/* number of images */
	int same;		/* nonzero: images have same size */
	int contig;		/* nonzero: images block contiguous */
	int order;		/* nonzero: images sequentially ordered */
	int max_ncol;		/* max ncol for all images */
	int max_nrow;		/* max nrow for all images */
	int min_ncol;		/* min ncol for all images */
	int min_nrow;		/* min nrow for all images */
	int *ncol;		/* array of ncol's per image */
	int *nrow;		/* array of nrow's per image */
	char **label;		/* array of labels per image */
	unsigned long *bptr;	/* array of block pointers to image plus header */
	unsigned long *nblk;	/* array of size of images (in blocks) w/o header */
	unsigned long first;	/* first block used */
	unsigned long tblk;	/* total blocks in file */
} PARSE_MIHEAD;

				/* mi-file types for PARSE_MIHEAD.type */
#define PARSE_MI_256 0		/* 256 entry/one block header */
#define PARSE_MI_1024 1		/* 1024 entry/four block header */
#define PARSE_MI_65K 2		/* header for infinite */
/*
	stuff for command table
*/
#define PARSE_INT     0
#define PARSE_FLOAT   1
#define PARSE_LONG    2
#define PARSE_CHAR    3		/* pointer */
#define PARSE_STRING  4		/* buffer */
#define PARSE_PROC    5		
#define PARSE_TOGGLE  11 	/* yes/no/y/n/0/1 */

		/* all int, for additional flag setting */
#define PARSE_SET 6
#define PARSE_CLR 7
#define PARSE_INC 8
#define PARSE_DCR 9
#define PARSE_VAL 10

#define PARSE_SHIFT 5
#define PARSE_VALUE(x) ((x)<< PARSE_SHIFT)|PARSE_VAL

#define PARSE_CHAR_POINTER PARSE_CHAR
#define PARSE_CHAR_BUF PARSE_STRING

typedef void *PARSE_ARG;

struct PARSE_OPTION_TABLE {
	char *label;
	int type;
	PARSE_ARG pointer;
	char *help;
	};

#define PARSE_STRUCT PARSE_OPTION_TABLE

/*
	stuff for flag table 
*/
struct PARSE_FLAG_TABLE{
	char flag;
	int *var;
	int type;
	char *help;
};
#define PARSE_FLAG_SET 0
#define PARSE_FLAG_CLR 1
#define PARSE_FLAG_INC 2
#define PARSE_FLAG_DCR 3
#define PARSE_FLAG_VAL 4
#define PARSE_FLAG_PROC 5

#define PARSE_FLAG_SHIFT 3
#define PARSE_FLAG_VALUE(x) ((x)<< PARSE_FLAG_SHIFT)|PARSE_FLAG_VAL
/*
	other defines
*/
#define PARSE_MAX_LINE	256	/* max # of characters read */
#define PARSE_ITOA	12	/* max # of characters in itoa call */
/*
	macros
*/
#define BLOCK_SHORT(x) (((x)+(512/sizeof(short)-1))/(512/sizeof(short)))
#define BLOCK_FLOAT(x) (((x)+(512/sizeof(float)-1))/(512/sizeof(float)))

#define PARSE_BLOCK_SHORT(x) (((x)+(512/sizeof(short)-1))/(512/sizeof(short)))
#define PARSE_BLOCK_FLOAT(x) (((x)+(512/sizeof(float)-1))/(512/sizeof(float)))
/*
	function defintions
*/
void *parse_malloc(size_t n);
float *parse_malloc_float(int n);
char *parse_malloc_char(int n);
int *parse_malloc_int(int n);
long *parse_malloc_long(int n);
double *parse_malloc_double(int n);
short *parse_malloc_short(int n);
char *parse_root(char *s);
char *parse_itoa(int i,char *s);
double parse_atof(char *s);
int parse_atof_error(void);
float *parse_vread(float *data,int *n,int mn,int type,char *file);
FILE *parse_fopen(char *s,char *mode);
FILE *parse_popen(char *s,char *mode);
void parse_pclose(FILE *s);
unsigned short int *parse_rmi(char *file,unsigned short *image,int *rncol,int *rnrow);
char *parse_fn_name(FILE *fd);
char *parse_suffix_malloc(char *n,char *s);
void parse_acf(char *message,double d,float *i);
void parse_acfmm(char *message,double amin,double amax,double d,float *i);
void parse_aci(char *message,int d,int *i);
void parse_acimm(char *message,int amin,int amax,int d,int *i);
void parse_acm(char *message,int d,char *modes,int *i);
char *parse_acs(char *message,char *d,char *fe,char *i);
void parse_act(char *message,int d,int *i);
void parse_args(char *arg);
void parse_args_basic(char *arg);
void parse_atof_args(FILE *fd,int stop);
void parse_basic(int argc,char **argv);
void parse_fclose(FILE *s);
void parse_err(char *s1,char *s2);
void parse_er(char *s1);
void parse_fcheck(double i,double imin,double imax,char *s);
void parse_file(char *file);
void parse_file_basic(char *file);
int parse_file_proc(struct PARSE_STRUCT *options,struct PARSE_FLAG_TABLE *flags,
	FILE *fd,char *file,int (*proc)(),int proc_type,int home,int efile);
int parse_args_general(struct PARSE_OPTION_TABLE *options,
	struct PARSE_FLAG_TABLE *flags,FILE *fd,char *arg);
int parse_file_general(struct PARSE_OPTION_TABLE *options,
	struct PARSE_FLAG_TABLE *flags,FILE *fd,char *file);
void parse_flags(char *s);
int parse_fflags(struct PARSE_FLAG_TABLE *table,char *s,FILE *fd);
void parse_flag_help(struct PARSE_FLAG_TABLE *table,FILE *fd);
void parse_fread(void *ptr,int size,int n,FILE *unit);
void parse_fread_block(void *ptr,int size,int n,FILE *unit,int blk);
void parse_free(void *p);
void parse_fwrite(void *ptr,int size,int n,FILE *unit);
void parse_pcheck(int i,int imin,int imax,char *s);
void parse_option_help(struct PARSE_STRUCT *parse_table,FILE *fd);
void parse_pdt(char *s);
void parse_prf(char *s,double i);
void parse_pri(char *s,int i);
void parse_prm(char *s,char *modes,int i);
void parse_prs(char *s,char *i);
void parse_uprs(FILE *u,char *s,char *i);
void parse_prc(char *s,char i);
void parse_prt(char *s,int i);
void parse_updt(FILE *u,char *s);
void parse_uprf(FILE *u,char *s,double i);
void parse_upri(FILE *u,char *s,int i);
void parse_uprm(FILE *u,char *s,char *modes,int i);
void parse_uprc(FILE *u,char *s,char i);
void parse_uprt(FILE *u,char *s,int i);
void parse_rblock(FILE *unit,int start,void *data,int count);
void parse_remove_suffix(char *n);
void parse_fseek(FILE *unit,long byte,char *type);
void parse_suffix(char *n,char *s);
void parse_tolower(char *s);
void parse_toupper(char *s);
void parse_uwtext(FILE *u,char *s);
void parse_vwrite(float *data,int n,int type,char *file,char *message);
void parse_vwrite_float(float *data,int n,char *file);
void parse_wblock(FILE *unit,int start,void *data,int count);
void parse_wda(float *data,int col,int row,char *file);
float *parse_rda(float *data,int *rcol,int *rrow,char *file);
void parse_wmi(char *file,unsigned short *image,int ncol,int nrow);
void parse_wmi_l(char *file,unsigned short *image,int ncol,int nrow,char *label);
void parse_words(char *s,int *argc,char **argv,int mword);
void parse_wtext(char *s);
void parse_zbuf(short int *a,int n);
void parse_zsbuf(short int *a,int n);
void parse_zrbuf(float *a,int n);
void parse_zfbuf(float *a,int n);
void parse_zibuf(int *a,int n);
void parse_help_basic(void);
void parse_all_basic(char *file,char *arg,int argc,char **argv);
void parse_help_general(struct PARSE_OPTION_TABLE *options,
	struct PARSE_FLAG_TABLE *flags,FILE *fd);
int parse_general(struct PARSE_OPTION_TABLE *options,struct PARSE_FLAG_TABLE *flags,
	FILE *fd,int argc,char **argv);
int parse_args_proc(struct PARSE_STRUCT *options,struct PARSE_FLAG_TABLE *flags,
	FILE *fd,char *arg,int (*proc)(),int proc_type);
void parse_fft(float *a,float *b,int m,int iopt);
void parse_fft2d(float *a,float *b,int mx,int my,int iopt);
void parse_dami(float *a,short int *mi,int n,int m);
int parse_opexist(char *op);
int parse_urline(FILE *u,char *s);
int parse_rline(char *s);
int parse_pparse_aa(struct PARSE_STRUCT *parse_table,FILE *fd,int *argc,char ***argv);
int parse_pparse(struct PARSE_STRUCT *parse_table,char *s,FILE *fd);
int parse_comm_aa(int *argc,char ***argv);
int parse_comm(char *s);

long parse_from_can_long(long n);
short parse_from_can_short(short n);
unsigned short parse_from_can_ushort(unsigned short n);
float parse_from_can_float(int n);
long parse_to_can_long(long n);
short parse_to_can_short(short n);
unsigned short parse_to_can_ushort(unsigned short n);
int parse_to_can_float(float n);
void parse_from_can_longv(long *v,int n);
void parse_from_can_shortv(short *v,int n);
void parse_from_can_ushortv(unsigned short *v,int n);
void parse_from_can_floatv(void *v,int n);
void parse_to_can_longv(long *v,int n);
void parse_to_can_shortv(short *v,int n);
void parse_to_can_ushortv(unsigned short *v,int n);
void parse_to_can_floatv(void *v,int n);

long parse_from_ican_long(long n);
short parse_from_ican_short(short n);
float parse_from_ican_float(int n);
long parse_to_ican_long(long n);
short parse_to_ican_short(short n);
int parse_to_ican_float(float n);
void parse_from_ican_longv(long *v,int n);
void parse_from_ican_shortv(short *v,int n);
void parse_from_ican_floatv(void *v,int n);
void parse_to_ican_longv(long *v,int n);
void parse_to_ican_shortv(short *v,int n);
void parse_to_ican_floatv(void *v,int n);

char *parse_host(void);
PARSE_MIHEAD *parse_omi(FILE *iu,int writeonly);
PARSE_MIHEAD *parse_omi_fe(FILE *iu,int writeonly,FILE *ou,int err);
void parse_omi_checklimit(int limit);
void parse_frmi(PARSE_MIHEAD *mihead);
void parse_fn_error(char *s,FILE *fd);
void parse_file_default(char *file);
void parse_arg_default(char *arg);
char * parse_extract_suffix(char *n);
void parse_rcsid_print(char *id);
void parse_time_print(FILE *ofd,char *s);
void parse_time_init(void);

#endif /* _parse_h */
