/*
print - print different kinds of variables
*/

#ident "$Id: print.c,v 1.7 2007/05/22 15:51:51 ccrawford Exp $"

#include <sys/types.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <stdlib.h>
#include "parse.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

		/* stuff for timing functions */
static int eTimeu,eTimes;	/* elapsed time in usec and s */
static int sTimeu,sTimes;	/* system time in usec and s */
static int uTimeu,uTimes;	/* user time in usec and s */
static struct timeval timep;	/* for call to gettimeofday */
static struct rusage usage;	/* for call to getrusage */
static int time_init_run = 0;	/* 1 after call to parse_time_init */

#ifdef __sun__
int getrusage(int, struct rusage *);    /* missing in solaris */
#endif


void 
parse_upri (FILE *u, char *s, int i)
{
	int e;

	if(s != NULL){
		e = *s == '*';
		fprintf(u,"%s %d\n",s+e,i);
		if(e) exit(1);
	}else{
		fprintf(u,"%d\n",i);
	}
}

void 
parse_pri (char *s, int i)
{
	parse_upri(stdout,s,i);
}

void 
parse_uprf (FILE *u, char *s, double i)
{
	int e;

	if(s != NULL){
		e = *s == '*';
		if(fabs(i) > .01 || i == 0.0)
			fprintf(u,"%s %f\n",s+e,i);
		else
			fprintf(u,"%s %e\n",s+e,i);
		if(e) exit(1);
	}else{
		if(fabs(i) > .01)
			fprintf(u,"%f\n",i);
		else
			fprintf(u,"%e\n",i);
	}
}

void 
parse_prf (char *s, double i)
{
	parse_uprf(stdout,s,i);
}

void 
parse_uprs (FILE *u, char *s, char *i)
{
	int e;

	if(s != NULL){
		e = *s == '*';
		fprintf(u,"%s %s\n",s+e,i);
		if(e) exit(1);
	}else{
		fprintf(u,"%s\n",i);
	}
}

void 
parse_prs (char *s, char *i)
{
	parse_uprs(stdout,s,i);
}

void 
parse_err (char *s1, char *s2)
{
	if(s2 != NULL)
		fprintf(stderr,"%s%s\n",s1,s2);
	else
		fprintf(stderr,"%s\n",s1);
	exit(1);
}

void 
parse_er (char *s1)
{
	parse_err(s1,(char *)NULL);
}

void 
parse_uwtext (FILE *u, char *s)
{
	if(s != NULL) fputs(s,u);
	fputc('\n',u);
}

void 
parse_wtext (char *s)
{
	parse_uwtext(stdout,s);
}

void 
parse_updt (FILE *u, char *s)
{
	time_t my_clock;
	int e;

	if(s != NULL){
		e = *s == '*';
		fputs(s+e,u);
		fputs(": ",u);
	}else{
		e = 0;
	}

	time(&my_clock);
	fputs(ctime(&my_clock),u);
	if(e) exit(1);
}

void 
parse_pdt (char *s)
{
	parse_updt(stdout,s);
}

void 
parse_uprt (FILE *u, char *s, int i)
{
	int e;

	if(s != NULL){
		e = *s == '*';
		fprintf(u,"%s: %s\n",s+e,i? "yes" : "no");
		if(e) exit(1);
	}else{
		fprintf(u,"%s\n",i? "yes" : "no");
	}
}

void 
parse_prt (char *s, int i)
{
	parse_uprt(stdout,s,i);
}

void 
parse_uprm (FILE *u, char *s, char *modes, int i)
{
	int e;

	e = *s == '*';
	if(i < 0 || i >= (int) strlen(modes)){
		parse_pri("uprm error: illegal mode",i);
		exit(1);
	}

	if(s != NULL){
		fprintf(u,"%s: %c\n",s+e,modes[i]);
		if(e) exit(1);
	}else{
		fprintf(u,"%c\n",modes[i]);
	}
}

void 
parse_prm (char *s, char *modes, int i)
{
	parse_uprm(stdout,s,modes,i);
}

void parse_uprc(FILE *u,char *s,char i)
{
	int e;

	if(s != NULL){
		e = *s == '*';
		fprintf(u,"%s %c\n",s+e,i);
		if(e) exit(1);
	}else{
		fprintf(u,"%c\n",i);
	}
}

void parse_prc(char *s,char i)
{
	parse_uprc(stdout,s,i);
}

void 
parse_rcsid_print(char *id)
{
	int argc;
	char *argv[20];
	char *s;
	char *idp;

	idp = (char *) parse_malloc(strlen(id)+1);
	strcpy(idp,id);

	parse_words(idp,&argc,argv,20);

	if(argc != 8 && argc != 9){
		parse_uwtext(stderr,"parse_rcsid_print: bad word count");
		parse_uprs(stderr,"rcsid:",id);
		exit(1);
	}
	if(strcmp(argv[0],"$Id:")){
		parse_uwtext(stderr,"parse_rcsid_print: 'Id' not found");
		parse_uprs(stderr,"rcsid:",id);
		exit(1);
	}
	if(strcmp(argv[6],"Exp")){
		parse_uwtext(stderr,"parse_rcsid_print: 'Exp' not found");
		parse_uprs(stderr,"rcsid:",id);
		exit(1);
	}

	s = (char *) parse_malloc(strlen(argv[2])+20);
	strcpy(s,"rcs revision: ");
	strcat(s,argv[2]);
	if(argc == 9) strcat(s,"L");
	parse_wtext(s);
	parse_free(s);
	parse_free(idp);
}

void
parse_time_init(void)
{
	gettimeofday(&timep,NULL);
	eTimeu = timep.tv_usec;
	eTimes = timep.tv_sec;
	getrusage(RUSAGE_SELF,&usage);
	uTimeu = usage.ru_utime.tv_usec;
	uTimes = usage.ru_utime.tv_sec;
	sTimeu = usage.ru_stime.tv_usec;
	sTimes = usage.ru_stime.tv_sec;
	time_init_run = 1;
}

void
parse_time_print(FILE *ofd,char *s)
{
	int eTimem;		/* elapses time in minutes */

	if(!time_init_run){
		parse_er("parse_time_print: parse_time_init not run");
	}

	gettimeofday(&timep,NULL);
	eTimeu = timep.tv_usec - eTimeu;
	eTimes = timep.tv_sec - eTimes;
	if(eTimeu < 0){
		eTimeu += 1000000;
		eTimes -= 1;
	}
	getrusage(RUSAGE_SELF,&usage);
	uTimeu = usage.ru_utime.tv_usec - uTimeu;
	uTimes = usage.ru_utime.tv_sec - uTimes;
	sTimeu = usage.ru_stime.tv_usec - sTimeu;
	sTimes = usage.ru_stime.tv_sec - sTimes;
	if(sTimeu < 0){
		sTimeu += 1000000;
		sTimes -= 1;
	}
	if(uTimeu < 0){
		uTimeu += 1000000;
		uTimes -= 1;
	}
				/* convert to hundredths of a second */
	eTimeu /= 10000;
	uTimeu /= 10000;
	sTimeu /= 10000;

	eTimem = eTimes / 60;
	eTimes %= 60;

	if(s != NULL && strlen(s) > 0){
		fprintf(ofd,"%s ",s);
	}
	fprintf(ofd,"%d.%02duser %d.%02dsystem %d:%02d.%02delapsed\n",
		uTimes,uTimeu,sTimes,sTimeu,eTimem,eTimes,eTimeu);
	parse_time_init();
}
