/*
block - block storage routines for qplot input routines
*/

#ident "$Id: block.c,v 1.3 2007/05/22 15:54:48 ccrawford Exp $"

#include <parse.h>
#include "qplot.h"

#ifdef TESTBLOCK
#define S 2048
#else
#define S 2048
#endif
#define EXTRA 4

static struct bd{
	float *x;
	float *y;
	struct bd *next;
} *root,*active;

static int c,m,n;

void block_init(int cc)
{
	m = n = 0;
	c = cc;
	if(c != 1 && c != 2)parse_err("block: bad vector length","");
	c--;
}

void block_add(float a,float b)
{
	struct bd *p;

	if(!n || n == S){
		p = (struct bd *)parse_malloc(sizeof(struct bd));
		if(m){
			active->next = p;
		}else{
			root = p;
		}
		active = p;
		active->next = NULL;
		active->x = (float *)parse_malloc(S*sizeof(float));
		if(c) active->y = (float *)parse_malloc(S*sizeof(float));
		n = 0;
		m++;
	}
	active->x[n] = a;
	if(c)active->y[n] = b;
	n++;
}

int block_flush(float **xv,float **yv)
{
	float *x,*xp,*yp=NULL,*y=NULL;
	struct bd *p;
	int i,j,nn,t;

	if(!m && !n) return(0);
	t = (m-1)*S+n;
	xp = x = (float *)parse_malloc((t+EXTRA) * sizeof(float));
	if(c)yp = y = (float *)parse_malloc((t+EXTRA) * sizeof(float));

	for(i=0;i<m;i++){
		if(i){
			p = active;
			active = active->next;
			parse_free((char *)p->x);
			if(c)parse_free((char *)p->y);
			parse_free((char *)p);
		}else{
			active = root;
		}
		nn = (i == m-1)? n : S;
		for(j=0;j<nn;j++){
			*xp++ = active->x[j];
			if(c)*yp++ = active->y[j];
		}
	}
	parse_free((char *)active->x);
	if(c)parse_free((char *)active->y);
	parse_free((char *)active);
	*xv = x;
	if(c) *yv = y;
	return(t);
}

#ifdef TESTBLOCK
main()
{
	float *x,*y;
	int j,i,n,c;
	j = 0;
	c= 0;

	while(++j < 3){
		block_init(c+1);
		block_add(2.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-1.0,1.23);
		block_add(-4.0,1.222);
		block_add((float)j,0.0);
		n = block_flush(&x,&y);
		printf("\n%d\n",n);
		if(c)
			for(i=0;i<n;i++) printf("%f %f\n",x[i],y[i]);
		else
			for(i=0;i<n;i++) printf("%f %f\n",x[i]);
	}
}
#endif
