/*
qplot - quickly plot vectors
*/

#ident "$Id: qplot.c,v 1.8 2011/09/14 16:03:53 carl Exp $"

#include <string.h>
#include <math.h>
#include "parse.h"
#include "crcplot.h"
#include "qplot.h"

#define FOVX	10.0		/* horizontal size of field-of-view */
#define FOVY	FOVX		/* vertical size of field-of-view */

#define MAX(x,y) ((x > y)? x: y)
#define MIN(x,y) ((x < y)? x: y)

static int paxis=1;		/*0=don't plot axes*/
static int erase=1;		/*0=don't 1=do earse screen*/
static int usey=1;		/*1=use y file only*/
static int pgrid=0;		/*1=plot grid*/
static int use_stdout;		/*1=send output to stdout */
static int nicescale=0;		/*1=use nice scales*/
static float *xd;		/*x vector (+2 = xmin, xmax) (+2 = nicescale)*/
static float *yd;		/*y vector (+2 = ymin, ymax)*/
static int label=0;		/*1=get labels from sys$input*/
static float scfac=1.0;		/*scale factor*/
static int lmode=0;		/*line-mode 0=line 1=dline*/
static float xxs=0.0;		/*x starting location of plot*/
static float yys=0.0;		/*y starting location of plot*/
static int saves=0;		/*1=save final scale informaton*/
static float height=0.14;	/*char heights*/
static float xmax_cl;		/* xmax on command line */
static float xmin_cl;		/* xmin on command line */
static float ymax_cl;		/* ymax on command line */
static float ymin_cl;		/* ymin on command line */
static int skip=0;		/*skip 'skip' points on input*/
static int begin=0;		/*starting point in input vectors*/
static int count= -1;		/*number of points to use*/
static float xlen=3.0;		/*length of x axis*/
static float ylen=3.0;		/*length of y axis*/
static int frame=0;		/*1=frame plot with border*/
static int online=0;		/*1=use online symbols*/
static int sym=4;		/*which on-line symbol*/
static int jol=0;		/*mark flag for on-line symbols*/
static float ds=0.1;		/*dash part of dashed line*/
static float gp=0.1;		/*gap part of dashed line*/
static int vtype;		/*vector type 0=da 1=binary 2=ascii*/
static int nx=1;		/*position of x-data*/
static int ny=1;		/*position of y-data*/
static int cxym=0;		/*1=use combined xy-data file*/
static char *ctype="f";		/* pointer to type string */
static int hexant,quadrant,nanant;	/* !=0 -> set subregion */
static int filelabel;		/* 1= use file names as labels */
static char *host;		/* host to connect to */
static int kludge=0;		/* turn on kludges */

static char *xl;		/*x axis label*/
static char *yl;		/*y axis label*/
static char *tl;		/*top label*/
static char *bl;		/*bottom label*/
static char *xf = "x";		/*x input vector name*/
static char *yf = "y";		/*y input vector name*/
static char *xyf = "xy";	/*scale file name*/
static char *cxyf = "cxy";	/*combined xy-data file name*/
static char *el;		/* end label */

struct PARSE_STRUCT parse_table[] = {
	{"x",PARSE_CHAR,(PARSE_ARG)&xf,"name of x-vector"},
	{"x",PARSE_CLR,(PARSE_ARG)&usey},
	{"y",PARSE_CHAR,(PARSE_ARG)&yf,"name of y-vector"},
	{"=",PARSE_CHAR,(PARSE_ARG)&yf,"name of y-vector"},
	{"s",PARSE_CHAR,(PARSE_ARG)&xyf,"name of file with scale factors"},
	{"xp",PARSE_FLOAT,(PARSE_ARG)&xxs,"horizontal offset to axes origin"},
	{"yp",PARSE_FLOAT,(PARSE_ARG)&yys,"vertical offset to axes origin"},
	{"xl",PARSE_CHAR,(PARSE_ARG)&xl,"label for x-axis"},
	{"yl",PARSE_CHAR,(PARSE_ARG)&yl,"label for y-axis"},
	{"skip",PARSE_INT,(PARSE_ARG)&skip,"points to skip between points"},
	{"begin",PARSE_INT,(PARSE_ARG)&begin,"first point, <0 is records"},
	{"xmin",PARSE_FLOAT,(PARSE_ARG)&xmin_cl,"minimum value on x-axis"},
	{"xmax",PARSE_FLOAT,(PARSE_ARG)&xmax_cl,"maximum value on x-axis"},
	{"ymin",PARSE_FLOAT,(PARSE_ARG)&ymin_cl,"minimum value on y-axis"},
	{"ymax",PARSE_FLOAT,(PARSE_ARG)&ymax_cl,"maximum value on y-axis"},
	{"count",PARSE_INT,(PARSE_ARG)&count,"number of input points"},
	{"scfac",PARSE_FLOAT,(PARSE_ARG)&scfac,"plotting scale factor"},
	{"scale",PARSE_FLOAT,(PARSE_ARG)&scfac,"plotting scale factor"},
	{"xlen",PARSE_FLOAT,(PARSE_ARG)&xlen,"length of x-axis"},
	{"ylen",PARSE_FLOAT,(PARSE_ARG)&ylen,"length of y-axis"},
	{"len",PARSE_FLOAT,(PARSE_ARG)&ylen,"length of both axes"},
	{"len",PARSE_FLOAT,(PARSE_ARG)&xlen},
	{"sym",PARSE_INT,(PARSE_ARG)&sym,"symbol number for on-center-symbols"},
	{"sym",PARSE_SET,(PARSE_ARG)&online},
	{"j",PARSE_INT,(PARSE_ARG)&jol,"point number for symbol, <0 no line"},
	{"j",PARSE_SET,(PARSE_ARG)&online},
	{"dash",PARSE_FLOAT,(PARSE_ARG)&ds,"length of dash for dashed line"},
	{"dash",PARSE_SET,(PARSE_ARG)&lmode},
	{"gap",PARSE_FLOAT,(PARSE_ARG)&gp,"length of gap for dashed line"},
	{"gap",PARSE_SET,(PARSE_ARG)&lmode},
	{"tl",PARSE_CHAR,(PARSE_ARG)&tl,"top label"},
	{"bl",PARSE_CHAR,(PARSE_ARG)&bl,"bottom label"},
	{"nx",PARSE_INT,(PARSE_ARG)&nx,"column number for x-vector"},
	{"nx",PARSE_VALUE(2),(PARSE_ARG)&vtype},
	{"nx",PARSE_CLR,(PARSE_ARG)&usey},
	{"ny",PARSE_INT,(PARSE_ARG)&ny,"colmn number for y-vector"},
	{"ny",PARSE_VALUE(2),(PARSE_ARG)&vtype},
	{"cxy",PARSE_CHAR,(PARSE_ARG)&cxyf,"name xy-vector in ascii format"},
	{"cxy",PARSE_VALUE(2),(PARSE_ARG)&vtype},
	{"cxy",PARSE_SET,(PARSE_ARG)&cxym},
	{"cxy",PARSE_CLR,(PARSE_ARG)&usey},
	{"type",PARSE_CHAR,(PARSE_ARG)&ctype,"type of binary data"},
	{"type",PARSE_VALUE(1),(PARSE_ARG)&vtype},
	{"el",PARSE_CHAR,(PARSE_ARG)&el,"end label"},
	{"quadrant",PARSE_INT,(PARSE_ARG)&quadrant,"plotting quadrant"},
	{"hexant",PARSE_INT,(PARSE_ARG)&hexant,"plotting haxant"},
	{"nanant",PARSE_INT,(PARSE_ARG)&nanant,"plotting nanant"},
	{"host",PARSE_CHAR,(PARSE_ARG)&host,"host[:port]"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'a',&paxis,PARSE_FLAG_CLR,"don't plot axes"},
	{'b',&vtype,PARSE_FLAG_VALUE(1),"binary mode"},
	{'c',&vtype,PARSE_FLAG_VALUE(2),"ascii/character mode"},
	{'d',&lmode,PARSE_FLAG_SET,"use dashed vectors"},
	{'e',&erase,PARSE_FLAG_CLR,"don't erase screen"},
	{'f',&frame,PARSE_FLAG_SET,"frame plot"},
	{'g',&pgrid,PARSE_FLAG_SET,"plot grid"},
	{'k',&kludge,PARSE_FLAG_SET,"enable kludge mode"},
	{'L',&filelabel,PARSE_FLAG_SET,"use file names as labels"},
	{'l',&label,PARSE_FLAG_SET,"get labels from command line"},
	{'m',&online,PARSE_FLAG_SET,"mark with online symbol"},
	{'n',&nicescale,PARSE_FLAG_SET,"nice scales"},
	{'o',&use_stdout,PARSE_FLAG_SET,"send plot output to stdout"},
	{'s',&saves,PARSE_FLAG_SET,"save scale information"},
	{'x',&usey,PARSE_FLAG_CLR,"open x-vector"},
	{'?',(int *)parse_help_basic,PARSE_PROC},
	{0,0,0,0}
};

int
main(argc, argv)
int argc;
char **argv;
{
	FILE *fd;
	int iy,ix,i,j=-1;
	float xtmp,ytmp,dx;
	float xmin,xmax,ymin,ymax,xp=-1,yp;
	int cticx,cticy;
/*
	initialize parameters
*/
	parse_all_basic(".qplot","QPLOTARGS",argc,argv);
/*
	set up other flags
*/
	if(xlen < 0){
		xlen = -xlen;
		cticx = -1;
	}else{
		cticx = 1;
	}

	if(ylen < 0){
		ylen = -ylen;
		cticy = -1;
	}else{
		cticy = 1;
	}
/*
	set up sub region
*/
	i = 0;
	if(quadrant){
		parse_pcheck(abs(quadrant),1,4,"bad quadrant");
		i = 2;
		j = quadrant;
	}
	if(nanant){
		parse_pcheck(abs(nanant),1,9,"bad nanant");
		i = 3;
		j = nanant;
	}
	if(hexant){
		parse_pcheck(abs(hexant),1,16,"bad hexant");
		i = 4;
		j = hexant;
	}
	if(i){
		if((quadrant&&hexant)||(quadrant&&nanant)||(hexant&nanant))
			parse_err("can't set more than one subregion","");

		if(j < 0){
			xxs += FOVX;
			j = -j;
		}
		yys += (i-1-(j-1)/i) * FOVY / ((float)i);
		xxs += (j-1-((j-1)/i)*i) * FOVX / ((float) i);
		scfac /= i;
	}
/*
	set up default labels
*/
	if(filelabel){
		if(usey){
			xl = yf;
		}else{
			xl = xf;
			yl = yf;
		}
	}
/*
	input x and y vectors
*/
	if(vtype == 2){
		iy = inputa(xf,yf,cxyf,cxym,&xd,&yd,
			begin,skip,count,nx,ny,usey);
		if(cxym)usey = 0;
	}else if(vtype == 0){
		iy = input(yf,&yd,begin,skip,count);
	}else{
		iy = inputb(yf,&yd,begin,skip,count,*ctype);
	}
	ix = iy;
	if(!usey){
		if(vtype == 2){
			;
		}else if(vtype == 0){
			ix = input(xf,&xd,begin,skip,count);
		}else{
			ix = inputb(xf,&xd,begin,skip,count,*ctype);
		}
	}else{
		xd = (float *)parse_malloc((ix+4) * sizeof(float));
	}

	if(ix != iy) parse_err("x and y vectors have different length","");
/*
	set up scale information
*/
	xyf = parse_suffix_malloc(xyf,"as");
	if(parse_opexist("s")){
		fd = parse_fopen(xyf,"r");
		if(0 == fscanf(fd,"%f %f %f %f",&xmin,&xmax,&ymin,&ymax)) parse_er("qplot: fscanf error reading scale");
		parse_fclose(fd);
	}else{
		xmax = xmin = xd[0];
		ymax = ymin = yd[0];
		for(i=0;i<iy;i++){
			yp = yd[i];
			if(!usey) xp = xd[i];
			if(yp > ymax)ymax = yp;
			if(yp < ymin)ymin = yp;
			if(!usey){
				if(xp > xmax)xmax = xp;
				if(xp < xmin)xmin = xp;
			}
		}
	}
/*
	set up max/min overrides
*/
	if(parse_opexist("xmin")) xmin = xmin_cl;
	if(parse_opexist("xmax")) xmax = xmax_cl;
	if(parse_opexist("ymin")) ymin = ymin_cl;
	if(parse_opexist("ymax")) ymax = ymax_cl;
/*
	set up x vector
*/
	if(usey == 1){
		begin = MAX(0,begin);
		if(parse_opexist("s") || 
		  (parse_opexist("xmin") && parse_opexist("xmax"))){
			dx = (xmax_cl - xmin_cl)/(iy-1);
			for(i=0;i<iy;i++) xd[i] = xmin_cl + i*dx;
		}else if(parse_opexist("xmin")){
			j = -skip - 1;
			for(i=0;i<iy;i++){
				j += skip + 1;
				xd[i] = xmin_cl + j;
			}
			xmin = xmin_cl;
			xmax = xd[iy-1];
		}else{
			j = -skip - 1;
			for(i=0;i<iy;i++){
				j += skip + 1;
				xd[i] = begin + j;
			}
			xmin = begin;
			xmax = xd[iy-1];
		}
	}else{
		i = monotonicity(xd,iy);
		if(i == 2 || i == 4){
			xtmp = xmax;
			xmax = xmin;
			xmin = xtmp;
		}
	}
	if(kludge){
		xtmp = xmax;
		xmax = xmin;
		xmin = xtmp;
	}
/*
	save scale file
*/
	if(saves){
		fd = parse_fopen(xyf,"w");
		(void) fprintf(fd,"%f %f %f %f\n",xmin,xmax,ymin,ymax);
		parse_fclose(fd);
	}
/*
	misc. scaling
*/
	for(i=0;i<iy;i++) yd[i] = MAX(MIN(yd[i],ymax),ymin);

	if(ymin == ymax) parse_uprf(stderr,"*ymin = ymax =",ymin);
	if(xmin == xmax) parse_uprf(stderr,"*xmin = xmax =",xmin);
/*
	do scaling 
*/
	if(nicescale){
		xd[iy] = xmin;
		xd[iy+1] = xmax;
		yd[iy] = ymin;
		yd[iy+1] = ymax;

		scale(xd,xlen,iy+2,1);
		scale(yd,ylen,iy+2,1);

		xd[iy] = xd[iy+2];
		xd[iy+1] = xd[iy+3];
		yd[iy] = yd[iy+2];
		yd[iy+1] = yd[iy+3];
	}else{
		xd[iy] = xmin;
		xd[iy+1] = (xmax - xmin) / xlen;
		yd[iy] = ymin;
		yd[iy+1] = (ymax - ymin) / ylen;
	}
/*
	get labels
*/
	if(label){
		xl = parse_acs("x label",xl,(char *)NULL,(char *)NULL);
		yl = parse_acs("y label",yl,(char *)NULL,(char *)NULL);
		bl = parse_acs("bottom label",bl,(char *)NULL,(char *)NULL);
		tl = parse_acs("top label",tl,(char *)NULL,(char *)NULL);
	}
/*	
	do the plotting
*/
	if(host){
		plots_host(host);
	}else if(use_stdout){
		plots_host("stdout");
	}else{
		plots();
	}

	if(erase == 1)clear();
	plot(xxs,yys,-3);
	factor(scfac * 2.0);
	plot(.75,1.0,-3);

	if(paxis == 1){
		axis(0.0,0.0,xl,-999,cticx*xlen,0.0,xd[iy],xd[iy+1]);
		axis(0.0,0.0,yl,999,cticy*ylen,90.0,yd[iy],yd[iy+1]);
	}

	if(bl && (i = strlen(bl))){
		symbol(xlen*0.5-(3./7.*i*height),-.6,height,bl,0.0,i);
	}

	if(frame == 1){
		plot(xlen,0.0,3);
		plot(xlen,ylen,2);
		plot(0.0,ylen,2);
		if(paxis == 1){
			plot(0.0,0.0,2);
			plot(xlen,0.0,2);
		}
	}

	if(!lmode){
		if(online && !jol)jol = 1;
		line(xd,yd,iy,1,jol,sym);
	}else{
		dline(xd,yd,iy,&ds,&gp,1);
	}

	if(el){        				/* end label */
		i = strlen(el);
		xtmp = (xd[iy-1] - xd[iy]) / xd[iy+1] +.25;
		ytmp = (yd[iy-1] - yd[iy]) / yd[iy+1];
		symbol(xtmp,ytmp,height,el,0.0,i);
	}

	if(tl && (i = strlen(tl))){
		symbol(xlen*0.5-(3./7.*i*height),ylen+height,height,tl,0.0,i);
	}

	if(pgrid){
		nx = xlen;
		ny = ylen;
		for(i=0;i<MAX(nx,ny);i++) xd[i] = yd[i] = 1;
		xd[nx] = xlen - nx;            /*fractional tic mark*/
		yd[ny] = ylen - ny;
		grid(0.,0.,1001+nx,xd,1001+ny,yd);
	}
/*
	clean up and exit
*/
	plot(0.0,0.0,999);
	exit(0);
}
