/*
asda - convert ascii (.as) files to .da files
*/

static char rcsid[] = "$Id: asda.c,v 1.7 2011/09/14 14:09:03 carl Exp $";

#include <string.h>
#include <math.h>
#include <parse.h>

#define S 2048			/* block size for block functions */
#define MWORD 200		/*max # words on a line*/

static int print=1;		/*flag for printing program info*/
static char *ofn;		/*output filename for y*/
static int ofns;		/*output file name set*/
static char *xfn;		/*output filename for x*/
static char *ifn="y.as";	/*input filename*/
static int count,counts;	/*number of points and set*/
static int begin;		/*first point, origin=0*/
static int skip;		/*skip points*/
static int nx=1,ny=1;		/*columbns where to find data*/
static int usex;		/*toggle to generate x file */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input filename"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,
		"output filename (default from input file)"},
	{"skip",PARSE_INT,(PARSE_ARG)&skip,"number skip points"},
	{"ny",PARSE_INT,(PARSE_ARG)&ny,"[relative] column for y vector"},
	{"nx",PARSE_INT,(PARSE_ARG)&nx,"column for x-vector"},
	{"begin",PARSE_INT,(PARSE_ARG)&begin,"first point (origin=0)"},
	{"count",PARSE_INT,(PARSE_ARG)&count,"number of points (default all)"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

static struct bd{
	float *x;
	float *y;
	struct bd *next;
} *root,*active;

static int c,m,n;

static void block_init(int cc)
{
	m = n = 0;
	c = cc;
	if(c != 1 && c != 2)parse_err("block: bad vector length","");
	c--;
}

static void block_add(double a,double b)
{
	struct bd *p;

	if(!n || n == S){
		p = (struct bd *)parse_malloc(sizeof(struct bd));
		if(m){
			active->next = p;
		}else{
			root = p;
		}
		active = p;
		active->next = NULL;
		active->x = (float *)parse_malloc(S*sizeof(float));
		if(c) active->y = (float *)parse_malloc(S*sizeof(float));
		n = 0;
		m++;
	}
	active->x[n] = a;
	if(c)active->y[n] = b;
	n++;
}

static int block_flush(float **xv,float **yv)
{
	float *x,*xp,*yp=NULL,*y=NULL;
	struct bd *p;
	int i,j,nn,t;

	if(!m && !n) return(0);
	t = (m-1)*S+n;
	xp = x = parse_malloc_float(t);
	if(c)yp = y = parse_malloc_float(t);

	for(i=0;i<m;i++){
		if(i){
			p = active;
			active = active->next;
			parse_free((char *)p->x);
			if(c)parse_free((char *)p->y);
			parse_free((char *)p);
		}else{
			active = root;
		}
		nn = (i == m-1)? n : S;
		for(j=0;j<nn;j++){
			*xp++ = active->x[j];
			if(c)*yp++ = active->y[j];
		}
	}
	parse_free((char *)active->x);
	if(c)parse_free((char *)active->y);
	parse_free((char *)active);
	*xv = x;
	if(c) *yv = y;
	return(t);
}

int
main(argc,argv)
	int argc;
	char **argv;
{
	float *x=NULL,*y=NULL;
	FILE *ifd;
	int i;
        char s[PARSE_MAX_LINE];		/*line from input files*/
        char *urgv[MWORD];		/*pointers to words*/
        int urgc;			/*# words found on a line*/
	int iy;				/*number of points returned*/
	int loop,isb;
/*
	parse command line and fix up defaults
*/
	parse_basic(argc,argv);

	counts = parse_opexist("count");
	ofns = parse_opexist("of");
	usex = parse_opexist("nx");
/*
	fix file names and open input file
*/
	if(!ofns)ofn = strcpy(parse_malloc_char(strlen(ifn)+1),ifn);
        ifn = parse_suffix_malloc(ifn,"as");
        parse_remove_suffix(ofn);
        ofn = parse_suffix_malloc(ofn,"da");
	if(usex){
		xfn = parse_malloc_char(strlen(ofn)+4);
		strcpy(xfn,ofn);
		parse_remove_suffix(xfn);
		strcat(xfn,"_x");
		xfn = parse_suffix_malloc(xfn,"da");
	}
        ifd = parse_fopen(ifn,"r");
/*
	print program status
*/
	if(print){
		parse_wtext("****************** asda *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_pri("relative y-column:",ny);
		if(usex){
			parse_prs("x-vector:",xfn);
			parse_pri("x-column:",nx);
		}
		parse_pri("skip:",skip);
		if(counts)parse_pri("count:",count);
		parse_pri("begin:",begin);
	}
/*
	do the reading
*/
	block_init(2);
	loop = 1;
	iy = 0;
	while(loop){
		if(counts && iy == count)break;
		if(loop == 1){
			isb = begin;             
			loop = 2;
		}else{
			isb = skip;
		}

		for(i=0;i<isb;i++) if(parse_urline(ifd,s))break;
		if(parse_urline(ifd,s))break;
		parse_words(s,&urgc,urgv,MWORD);

		if(usex){
			if(nx+ny < 0 || nx + ny > urgc)break;
			if(nx < 0 || nx > urgc)break;
			block_add(parse_atof(urgv[nx-1]),
				parse_atof(urgv[nx+ny-1]));
		}else{
			if(ny < 0 || ny > urgc)break;
			block_add(0.0,parse_atof(urgv[ny-1]));
		}
		iy++; 
	}
	if(iy <= 0)parse_err("no points read from input file","");
/*
	write vectors
*/
	block_flush(&x,&y);
	parse_vwrite(y,iy,0,ofn,"");
	if(usex)parse_vwrite(x,iy,0,xfn,"");
/*
	clean up and exit
*/
        if(print){
		parse_pri("points:",iy);
		parse_pdt("---> stop time");
	}
	exit(0);
}
