/*
cmi - combine mi-files through weighted combinations
*/

static char rcsid[] = "$Id: cmi.c,v 1.3 2011/09/14 14:09:03 carl Exp $";

#include <parse.h>
#include <limits.h>

static char *ofn="out.mi";	/*output filename*/
static char *ifn1="in1.mi";	/*input filename*/
static char *ifn2="in2.mi";	/*input filename*/
static int print=1;		/*flag for printing program info*/
static float w1=1.0;		/*weight*/
static float w2=1.0;		/*weight*/
static int cmode=1;		/*0=add 1=subtract */
static int ncf=0;		/*# files on command line w/o =*/
static float offset=0;		/* offset after operation */
static int onefile=0;		/* 1=process only one file */

static void eq_proc(char *option,char *arg);

struct PARSE_OPTION_TABLE parse_table[] = {
	{"i1",PARSE_CHAR,(PARSE_ARG)&ifn1,"first input file"},
	{"i2",PARSE_CHAR,(PARSE_ARG)&ifn2,"second input file"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"w1",PARSE_FLOAT,(PARSE_ARG)&w1,"weight for first vector"},
	{"w2",PARSE_FLOAT,(PARSE_ARG)&w2,"weight for second vector"},
	{"offset",PARSE_FLOAT,(PARSE_ARG)&offset,"offset after operation"},
	{"=",PARSE_PROC,(PARSE_ARG)eq_proc,"set files"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'o',&onefile,PARSE_FLAG_SET,"process one file"},
	{'a',&cmode,PARSE_FLAG_VALUE(0),"add images"},
	{'s',&cmode,PARSE_FLAG_VALUE(1),"subtract images (default)"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(argc,argv)
	int argc;
	char **argv;
{
	int nc,nr;		/* columns and rows */
	int ncp,nrp;		/* columns and rows */
	unsigned short *d1,*d2,*d3;	/* data arrays */
	int min1,min2,min3;	/* min values in input and output files */
	int max1,max2,max3;	/* max values in input and output files */
	int n;			/* total pixels */
	int i;			/* loopers */
	int res;		/* result of operation */
/*
	parse command line and check for errors
*/
	parse_basic(argc,argv);
	if(ncf == 1 && onefile == 0)parse_err("cmi: only one file set","");
/*
	open files and buffers
*/
        ofn = parse_suffix_malloc(ofn,"mi");

        d1 = parse_rmi(ifn1,NULL,&nc,&nr);
	n = nc * nr;

        if(onefile == 0){
		d2 = parse_rmi(ifn2,NULL,&ncp,&nrp);
		if(ncp != nc || nrp != nr) parse_er("cmi: input files don't match");
	}else{
		d2 = (unsigned short *) parse_malloc_short(n);
		parse_zbuf((short *) d2,n);
	}

	d3 = (unsigned short *) parse_malloc_short(n);
/*
	print program status
*/
	if(print){
		parse_wtext("****************** cmi *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("first input file:",ifn1);
		if(onefile == 0){
			parse_prs("second input file:",ifn2);
		}else{
			parse_wtext("second input file: not used");
		}
		parse_prs("output file:",ofn);
		parse_pri("colums =",nc);
		parse_pri("rows =",nr);
		parse_prf("w1 =",w1);
		if(onefile == 0)parse_prf("w2 =",w2);
		parse_prf("offset =",offset);
		if(cmode == 0)parse_wtext("mode: add");
		if(cmode == 1)parse_wtext("mode: subtract");
	}
/*
	do it
*/
	min1 = min2 = min3 = USHRT_MAX + 1;
	max1 = max2 = max3 = -1;

	for(i=0;i<n;i++){
		if(d1[i] < min1) min1 = d1[i];
		if(d1[i] > max1) max1 = d1[i];
		if(d2[i] < min2) min2 = d2[i];
		if(d2[i] > max2) max2 = d2[i];

		if(cmode == 0){
			res = w1 * d1[i] + w2 * d2[i] + offset;
		}else if(cmode == 1){
			res = w1 * d1[i] - w2 * d2[i] + offset;
		}else{
			res = -1;		/* internal code trap */
		}

		if(res > USHRT_MAX) res = USHRT_MAX;
		if(res < 0) res = 0;
		if(res < min3) min3 = res;
		if(res > max3) max3 = res;
		d3[i] = res;
	}
/*
	write output file and min/max results
*/
	parse_wmi(ofn,d3,nc,nr);

	if(print){
		parse_pri("min file1 =",min1);
		parse_pri("max file1 =",max1);
		if(onefile == 0){
			parse_pri("min file2 =",min2);
			parse_pri("max file2 =",max2);
		}
		parse_pri("min output file =",min3);
		parse_pri("max output file =",max3);
	}
/*
	clean up and exit
*/
        if(print)parse_pdt("---> stop time");
	exit(0);
}

static void eq_proc(option,arg)
char *option,*arg;
{
	switch(ncf++){

	case 0:
		ifn1 = arg;
		break;
	case 1:
		ifn2 = arg;
		break;
	default:
		parse_err("cmi: too many files","");
	}
}
