/*
dft - take dft of da-file
*/
static char rcsid[] = "$Id: dft.c,v 1.7 2002/03/06 22:48:14 ccrawford Exp $";

#include <parse.h>
#include <math.h>

static char *ifn="y.da";	/* input file name */
static char *ofn="z.da";	/* output file name */
static char *lfn="zl.da";	/* log file name */
static int logit=0;		/* take log of data */
static int print=1;		/* flag for printing program info */
static int m;			/* number of output points */
static int rowcol=0;		/* 0=row 1=col dft's */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file name"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output filename"},
	{"lf",PARSE_CHAR,(PARSE_ARG)&lfn,"log filename"},
	{"lf",PARSE_SET,(PARSE_ARG)&logit,},
	{"m",PARSE_INT,(PARSE_ARG)&m,"output points (default: n/2)"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'l',&logit,PARSE_FLAG_SET,"take log of data"},
	{'c',&rowcol,PARSE_FLAG_SET,"column dft's"},
	{'r',&rowcol,PARSE_FLAG_CLR,"row dft's (default)"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

static void 
dft(float *dr,float *di,float *fr,float *fi,int n,int n2)
{
	int i,k;
	double arg,rr,ri;
	float a,b,c,d;
	double pi = M_PI;

	for(k=0;k<n2;k++){
		arg = 2 * pi * k / n;
		rr = ri = 0;
		for(i=0;i<n;i++){
			a = dr[i];
			b = di[i];
			c = cos(arg * i);
			d = sin(arg * i);

			rr += a*c - b*d;
			ri += a*d + b*c;
		}
		fr[k] = rr;
		fi[k] = ri;
	}
}

int
main(int argc,char **argv)
{
	float *fb,*fa;	/* buffers for call to dft */
	float *dr;	/* real part of input data */
	float *drp;	/* vector of real part of input data */
	float *di;	/* imaginary part of input data (forced to zero) */
	int ms;		/* 1= m set on the command line */
	int i,j;	/* loopers */
	int n;		/* points per row */
	int r;		/* rows */
	float *lod;	/* logged output data */
	float *rod;	/* real output data */
	int vl;		/* input vector length */
	int nv;		/* number output vectors */
/*
	parse command line
*/
	parse_basic(argc,argv);
	ms = parse_opexist("m");
/*
	set up filenames 
*/
	ifn = parse_suffix_malloc(ifn,"da");
	ofn = parse_suffix_malloc(ofn,"da");
	lfn = parse_suffix_malloc(lfn,"da");
/*
	read data and allocate buffers
*/
	dr = parse_rda(NULL,&n,&r,ifn);

	if(rowcol == 0){	/* row mode */
		vl = n;
		nv = r;
	}else{
		vl = r;
		nv = n;
	}

	if(ms && m > vl/2)parse_er("dft: too many output points");
	m = (ms)? m : vl/2;

	di = parse_malloc_float(vl);
	drp = parse_malloc_float(vl);
	fa = parse_malloc_float(vl);
	fb = parse_malloc_float(vl);

	rod = parse_malloc_float(nv*m);
	lod = parse_malloc_float(nv*m);
/*
	print program information
*/
        if(print){
                parse_wtext("******************** dft ********************");
                parse_pdt("---start time");
		parse_rcsid_print(rcsid);
                parse_prs("input file:",ifn);
                parse_prs("output file:",ofn);
                if(logit) parse_prs("log file:",lfn);
                parse_pri("n (points/row) =",n);
                parse_pri("r (rows) =",r);
                parse_pri("m (dft points/row) =",m);
                parse_prs("dft mode:",rowcol? "column":"row");
	}
/*
	do it
*/
	if(rowcol == 0){		/* rows */
		for(j=0;j<r;j++){
			parse_zrbuf(di,n);

			dft(dr+n*j,di,fa,fb,n,m);

			for(i=0;i<m;i++) rod[i+j*m] = hypot(fa[i],fb[i]);

			if(logit){
				for(i=0;i<m;i++) lod[i+j*m] = log(rod[i+j*m]);
			}
		}
	}else{
		for(j=0;j<n;j++){
			parse_zrbuf(di,vl);
			for(i=0;i<vl;i++) drp[i] = dr[i*n+j];

			dft(drp,di,fa,fb,vl,m);

			for(i=0;i<m;i++) rod[j*m+i] = hypot(fa[i],fb[i]);

			if(logit){
				for(i=0;i<m;i++) lod[j*m+i] = log(rod[j*m+i]);
			}
		}
	}
/*
	write files
*/
        parse_wda(rod,m,nv,ofn);
	if(logit) parse_wda(lod,m,nv,lfn);
/*
	clean up and exit
*/
        if(print)parse_pdt("---stop time");
	exit(0);
}
