/*
fmm - find min/max of da-file
*/

#ident "$Id: fmm.c,v 1.9 2007/05/22 15:58:18 ccrawford Exp $"

#include "parse.h"
#include <math.h>

static int qplot;
static char *prog = "fmm";
static int many;

static void
syntax()
{
	fprintf(stderr,"syntax: %s [-q] file[.da] ...\n",prog);
	exit(1);
}

static void 
parse(s)
	char *s;
{
	while(*++s){
		switch(*s){

		case '?':
			syntax();
			break;
		case 'q':
			qplot = 1;
			break;
		default:
			fprintf(stderr,"bad flag: -%c\n",*s);
			syntax();
		}
	}
}

int
main(argc,argv)
	int argc;
	char **argv;
{
	char *fn;
	FILE *fd;
	short int head[256];
	int *lfr;
	float *fr;
	int i,j,t,n,k,iblk,nblk;
	float amin=0,amax=0,tmin=0,tmax=0;
	int aminmaxs=0;
	int tminmaxs=0;
	int minr=-1,maxr=-1,minp=-1,maxp=-1;

	--argc; ++argv;
	if(argc && **argv == '-'){
		parse(*argv);
		--argc; ++argv;
	}
	if(argc == 0) syntax();
	many = argc > 1;

	while(argc){
		aminmaxs = 0;
		fn = parse_suffix_malloc(*argv,"da");
		fd = parse_fopen(fn,"r");

		parse_rblock(fd,0,(char *)head,1);
		parse_from_can_shortv(head,256);
		n = head[0];
		k = head[1];
		t = head[2];
		if(t)parse_err("fmm: can't handle integer da files:",fn);

		nblk = BLOCK_FLOAT(n);
		lfr = parse_malloc_int(nblk*128);
		fr = parse_malloc_float(nblk*128);
		for(i=0,iblk=1;i<k;i++){
			parse_rblock(fd,iblk,(char *)lfr,nblk);
			iblk += nblk;

			for(j=0;j<n;j++){
				fr[j] = parse_from_can_float(lfr[j]);
				if(aminmaxs){
					if(fr[j] > amax){
						amax = fr[j];
						maxr = i;
						maxp = j;
					}
					if(fr[j] < amin){
						amin = fr[j];
						minr = i;
						minp = j;
					}
				}else{
					aminmaxs = 1;
					amax = fr[j];
					maxr = i;
					maxp = j;
					amin = fr[j];
					minr = i;
					minp = j;
				}
			}
		}

		if(tminmaxs){
			if(amin < tmin) tmin = amin;
			if(amax > tmax) tmax = amax;
		}else{
			tminmaxs = 1;
			tmin = amin;
			tmax = amax;
		}

		if(!qplot)printf("%s (%d,%d): min: %f (%d,%d) max: %f (%d,%d)\n",
			fn,k,n,amin,minr,minp,amax,maxr,maxp);
		parse_fclose(fd);
		parse_free(fn);
		--argc; ++argv;
	}
	if(!qplot){
		if(many)printf("min: %f max: %f\n",tmin,tmax);
	}else{
		printf("ymin=%f ymax=%f\n",tmin,tmax);
	}
	exit(0);
}
