/*
mag - magnify mi-images x2 using bi-cubic iterpolation 
*/

static char rcsid[] = "$Id: mag.c,v 1.5 2011/09/14 14:09:03 carl Exp $";

#include "parse.h"
#include <string.h>

typedef unsigned short PIXEL;
#define SMALL 10		/* minimun image size; need intrinsic padding */

static int print=1;		/* 1=print program status */
static char *ifn="pic.mi";	/* input file name*/
static char *ofn;		/* output file name*/

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,
		"output filename (default: input filename)"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

static int
inter(a,b,c,d)
PIXEL a,b,c,d;
{
	int p;

	p = .5625*(a + b)-.0625*(c + d)+ .5;
	return((p<0)? 0 : p);
}

int
main(int argc,char **argv)
{
	int nrow,ncol,nrowp,ncolp;
	int loop,lin,col;
	PIXEL *ip,*in,*op,*out,*outm1,*outm3,*outp1,*out0,*out2,*out1,*outp3;
	char *cp;
/*
	parse command line 
*/
	parse_all_basic(".mag","MAG",argc,argv);
/*
	open files
*/
	if(!parse_opexist("of")){
		ofn = parse_malloc_char(strlen(ifn)+1);
		strcpy(ofn,ifn);
		if((cp = strchr(ofn,':')) != NULL) *cp = 0;
	}
/*
	get image and allocate arrays
*/
	ip = parse_rmi(ifn,NULL,&ncol,&nrow);
/*
	check for small images
*/
	if(ncol <= SMALL) parse_pri("*mag: bad number of columns:",ncol);
	if(nrow <= SMALL) parse_pri("*mag: bad number of rows:",nrow);
/*
	set up output picture
*/
	nrowp = nrow * 2;
	ncolp = ncol * 2;
	op = (PIXEL *)parse_malloc(sizeof(PIXEL)*ncolp*nrowp);
/*
	print what's going on
*/
	if(print){
		ifn = parse_suffix_malloc(ifn,"mi");
		ofn = parse_suffix_malloc(ofn,"mi");

		parse_wtext("****************** mag *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_pri("input columns:",ncol);
		parse_pri("input rows:",nrow);
	}
/*
	magnify columns
*/
	for(lin=0;lin<nrowp;lin+=2){
		in = ip + ncol*lin/2;
		out = lin*ncolp + op;

					/* replicate even columns */
		for(col=0;col<ncolp;col += 2) out[col]=in[col/2];

					/* set up boundaries */
		out[1] = out[0];
		out[ncolp-3] = out[ncolp-1] = out[ncolp-2];

					/* interpolate odd columns */
		for(col=3,loop=ncolp-3;col<loop;col += 2){
	        	out[col] = inter(out[col-1],out[col+1],
				out[col-3],out[col+3]);
		}
	}
/*
	magnify rows
*/
	out0 = 0*ncolp + op;
	out1 = 1*ncolp + op;
					/* set up top boundary */
	for(col=0;col<ncolp;col++){
		out1[col] = out0[col];
	}

					/* set up bottom boundary */
	out0 = (nrowp-2)*ncolp + op;
	out1 = (nrowp-1)*ncolp + op;
	out2 = (nrowp-3)*ncolp + op;
	for(col=0;col<ncolp;col++){
		out2[col] = out1[col] = out0[col];
	}

					/* interpolate odd rows */
	for(lin=3,loop=nrowp-3;lin<loop;lin += 2){
		out = lin*ncolp + op;
		outm1 = (lin-1)*ncolp + op;
		outm3 = (lin-3)*ncolp + op;
		outp1 = (lin+1)*ncolp + op;
		outp3 = (lin+3)*ncolp + op;
		for(col=0;col<ncolp;col++){
			out[col]= inter(outm1[col],outp1[col],
				outm3[col],outp3[col]);
		}
	}
/*
	write output image
*/
	parse_wmi(ofn,op,ncolp,nrowp);
/*
	clean up and exit
*/
	if(print)parse_pdt("---> stop time");
	exit(0);
}
