/*
box - handle box cursor
*/

#ident "$Id: box.c,v 1.17 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static struct BOX_TAB{
	int x;
	int y;
	int xw;
	int yw;
	int type;	/* 0 = box 1 = cursor */
	struct BOX_TAB *next;
} *broot=NULL;

void
box_init(void)
{
	box.x = 0;
	box.y = 0;
	box.xw = 40;
	box.yw = 40;
	box.lx = 0;
	box.ly = 0;
	box.status = 0;
	box.vis = 0;
}

static void
draw_box_really(int x,int y,int xw,int yw,int gc)
{
	int xh,yh;
	int x1[4],y1[4],x2[4],y2[4];
	int i;

	xh = xw / 2;
	yh = yw / 2;

	x1[0] = x - xh;
	y1[0] = y - yh;
	x2[0] = x - xh + xw - 1;
	y2[0] = y - yh;

	x1[1] = x - xh + xw;
	y1[1] = y - yh;
	x2[1] = x - xh + xw;
	y2[1] = y - yh + yw - 1;

	x1[2] = x - xh + xw;
	y1[2] = y - yh + yw;
	x2[2] = x - xh + 1;
	y2[2] = y - yh + yw;

	x1[3] = x - xh;
	y1[3] = y - yh + yw;
	x2[3] = x - xh;
	y2[3] = y - yh + 1;

	for(i=0;i<4;i++){
		display_drawline(x1[i],y1[i],x2[i],y2[i],gc);
	}

	if(gc == LINE_ERASE){
		annotate_refresh();
	}

	// editor tag: peep
}

void
roi(int mode)
{
	char line[80];
	double mean,std,t;
	int exclude;			// number of excluded pixels
	CPIXEL pmin,pmax;
	register int i,j;
	CPIXEL *mip;
	int x2,y2,offset;
	int rmin,rmax,cmin,cmax;	/* locations of min/max pixels */
	int area,area2;
	int inroi;
	float cx,cy,ea,eb;
	int firstin;

	x2 = box.x - box.xw / 2;
	y2 = box.y - box.yw / 2;

	if(x2 < 0 || x2 + box.xw >= canvas_size_x ||
	   y2 < 0 || y2 + box.yw >= canvas_size_y){
		fprintf(ttyout,"ROI outside canvas\n");
		return;
	}

	if(box.xw == 0 || box.yw == 0){
		fprintf(ttyout,"Zero box width or height!\n");
		return;
	}

	if(mode)ctext("Calculating ROI");
	offset = y2 * canvas_size_x + x2;
	mip=image16+offset;
	rmin = rmax = cmin = cmax = 0;

	area = 0;
	cx = (box.xw - 1.0)/2.0;
	cy = (box.yw - 1.0)/2.0;
	ea = box.xw/2.0;
	eb = box.yw/2.0;
	firstin = 1;
	mean = 0.0;
	pmin = pmax = 0.0;
	exclude = 0;

	for(i=0;i<box.yw;i++){
		for(j=0;j<box.xw;j++){
			if(roiellipse){
				if(pow((j-cx)/ea,2.0)+pow((i-cy)/eb,2.0)<=1){
					inroi = 1;
				}else{
					inroi = 0;
				}
			}else{
				inroi = 1;
			}

			if(inroi){
				if(*mip < roimin){
					exclude++;
				}else{
					area++;
					if(firstin){
						pmin = *mip;
						rmin = i;
						cmin = j;

						pmax = *mip;
						rmax = i;
						cmax = j;

						firstin = 0;
					}else{
						if(*mip < pmin){
							pmin = *mip;
							rmin = i;
							cmin = j;
						}
						if(*mip > pmax){
							pmax = *mip;
							rmax = i;
							cmax = j;
						}
					}
					mean += *mip;
				}
			}
			mip++;
		}
		mip += canvas_size_x - box.xw;
	}

	if(area <= 0){
		fprintf(ttyout,"zero area!\n");
		fflush(ttyout);
		return;
	}

	mean /= area;
	rmin += y2;
	rmax += y2;
	cmin += x2;
	cmax += x2;
	area2 = 0;

	for(std=0.0,mip=image16+offset,i=0;i<box.yw;i++){
		for(j=0;j<box.xw;j++){
			if(roiellipse){
				if(pow((j-cx)/ea,2.0)+pow((i-cy)/eb,2.0)<=1){
					inroi = 1;
				}else{
					inroi = 0;
				}
			}else{
				inroi = 1;
			}

			if(inroi){
				if(*mip >= roimin){
					area2++;
					t = *mip - mean;
					std += t * t;
				}
			}
			mip++;
		}
		mip += canvas_size_x - box.xw;
	}

	if(area2 != area){
		fprintf(ttyout,"internal error: areas do not match!\n");
		return;
	}

	if(area == 1){
		fprintf(ttyout,"only one pixel in ROI; cannot calculate standard deviation!\n");
		return;
	}

	std = sqrt(std / (area - 1));

	if(mode){
		sprintf(line,"%7.1f %7.2f  ",mean-zero_offset,std);
		ctext(line);
	}
	sprintf(line,"mean: %.1f std: %.2f\n",mean-zero_offset,std);
	display_tty_out(line);
	sprintf(line,"min: %d (%d,%d) max: %d (%d,%d)\n",
		pmin-zero_offset,cmin,rmin,pmax-zero_offset,cmax,rmax);
	display_tty_out(line);
	sprintf(line,"center: (%d,%d) size: %d x %d area: %d%s\n",
		box.x,box.y,box.xw,box.yw,area,roiellipse?" (ellipse)": " (box)");
	display_tty_out(line);
	if(roimin > 0){
		sprintf(line,"roiminimum: %d exclusions: %d\n",roimin,exclude);
		display_tty_out(line);
	}
}

static void
roi_syntax()
{
	fputs("Synopsis: calculate region-of-interest statistics in box\n",ttyout);
	min_abbreviation();
	fputs("Related: box, cursor, panel\n",ttyout);
	fputs("Syntax: roi [-e] [x y [xw yw]]\n",ttyout);
	fputs("    * -e: use ellipse inscribed in box\n",ttyout);
	fputs("    * -b: use box inscribed in box\n",ttyout);
	fputs("    * x y: center box at (x,y)\n",ttyout);
	fputs("    * xw yw: set size of box to xw by yw\n",ttyout);
}

void
cmd_roi(void)
{
	int x,y,xw,yw;

	if(argc > 1 && !strcmp(argv[1],"-e")){
		roiellipse = 1;
		argc--;
	}else if(argc > 1 && !strcmp(argv[1],"-b")){
		roiellipse = 0;
		argc--;
	}

	if(argc != 1 && argc != 3 && argc != 5){ 
		roi_syntax();
		return;
	}

	if(argc == 1){
		if(!cursor_stat || !box.status){
			fputs("Box cursor not enabled\n",ttyout);
			return;
		}
		if(!cursor_in_win){
			fputs("Box not in window\n",ttyout);
			return;
		}
	}else{
		if(box.vis){
			draw_box(LINE_ERASE);
			box.vis = 0;
		}

		x = parse_atof(argv[1+roiellipse]);
		if(parse_atof_error()) return;
		if(x < 0 || x >= canvas_size_x){
			fputs("x outside of canvas\n",ttyout);
			return;
		}

		y = parse_atof(argv[2+roiellipse]);
		if(parse_atof_error()) return;
		if(y < 0 || y >= canvas_size_y){
			fputs("y outside of canvas\n",ttyout);
			return;
		}

		display_set_mouse(x,y);
		if(argc == 5){
			xw = parse_atof(argv[3+roiellipse]);
			if(parse_atof_error()) return;
			if(xw <= 0){
				fputs("xw zero or negative\n",ttyout);
				return;
			}

			yw = parse_atof(argv[4+roiellipse]);
			if(parse_atof_error()) return;
			if(yw <= 0){
				fputs("yw zero or negative\n",ttyout);
				return;
			}

			box.xw = xw;
			box.yw = yw;
		}
		box.x = x;
		box.y = y;
		cursor_stat = 1;
		box.status = 1;
		cur_on_off();
	}
	roi(0);
}

static void
box_syntax(void)
{
	fputs("Synopsis: manage box cursors\n",ttyout);
	min_abbreviation();
	fputs("Related: cursor\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("box on            * turn box cursor on\n",ttyout);
	fputs("box off           * turn box cursor off\n",ttyout);
	fputs("box               * change from off/on to on/off\n",ttyout);
	fputs("box -s x y        * set size to x by y\n",ttyout);
	fputs("box -p x y        * position box at (x,y)\n",ttyout);
}

void
cmd_box(void)
{
	int next_box,x,y,xw,yw;

	if(argc < 1 || argc > 4){
		box_syntax();
		return;
	}

	next_box = 0;
	if(argc == 1){
		next_box = box.status + 1;
		box.status = !box.status;
	}else{
		if(!strcmp("on",argv[1])){
			next_box = 1;
		}else if(!strcmp("off",argv[1])){
			next_box = 2;
		}else if(!strcmp("-s",argv[1])){
			if(argc != 4){
				fputs("syntax: box -s x y\n",ttyout);
				return;
			} 
			if(box.vis){
				draw_box(LINE_ERASE);
				box.vis = 0;
			}
			xw = parse_atof(argv[2]);
			if(parse_atof_error()) return;

			yw = parse_atof(argv[3]);
			if(parse_atof_error()) return;

			if(xw <= 0 || yw <= 0){
				fputs("Zero or negative width or height\n",ttyout);
				return;
			}

			box.xw = xw;
			box.yw = yw;

			next_box = 1;
		}else if(!strcmp("-p",argv[1])){
			if(argc != 4){
				fputs("syntax: box -p x y\n",ttyout);
				return;
			} 
			x = parse_atof(argv[2]);
			if(parse_atof_error()) return;

			y = parse_atof(argv[3]);
			if(parse_atof_error()) return;

			if(x < 0 || y < 0 || x >= canvas_size_x || y >= canvas_size_y){
				fputs("Box position outside canvas\n",ttyout);
				return;
			}

			display_set_mouse(x,y);
			next_box = 1;
		}else{
			fputs("On or off please!\n",ttyout);
			return;
		}
	}
	if(next_box == 1){
		cursor_stat = 1;
		box.status = 1;
		cur_on_off();
		fputs("Box cursor on\n",ttyout);
	}else if(next_box == 2){
		box.status = 0;
		cur_on_off();
		fputs("Box cursor off\n",ttyout);
	}
}

void
box_add(int x,int y,int xw,int yw,int type)
{
	struct BOX_TAB *bp,*bpp;

	if(NULL == (bp = (struct BOX_TAB *)malloc(sizeof(struct BOX_TAB)))){
		fputs("Can't allocate space for annotation\n",ttyout);
		return;
	}

	bp->x = x;
	bp->y = y;
	bp->xw = xw;
	bp->yw = yw;
	bp->type = type;
	bp->next = NULL;

	if(broot == NULL){
		broot = bp;
	}else{
		bpp = broot;
		while(bpp->next != NULL) bpp = bpp->next;
		bpp->next = bp;
	}
}

int
box_exist(void)
{
	return(broot != NULL);
}

void
box_refresh(void)
{
	struct BOX_TAB *bpp;

	bpp = broot;
	while(bpp != NULL){
		if(bpp->type){
			draw_cursor(bpp->x,bpp->y);
		}else{
			draw_box_really(bpp->x,bpp->y,bpp->xw,bpp->yw,LINE_HARD);
		}
		bpp = bpp->next;
	}
}

void
box_table_clear(void)
{
	struct BOX_TAB *bp,*bpp;

	bpp = broot;
	while(bpp != NULL){
		bp = bpp;
		bpp = bp->next;
		free(bp);
	}
	broot = NULL;
}

void
draw_box(int gc)
{
	draw_box_really(box.x,box.y,box.xw,box.yw,gc);
}
