/*
cursor - track cursor and perform some operations
*/

#ident "$Id: cursor.c,v 1.9 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

			/* cursor stack */
void
draw_cursor(int x,int y)
{
	display_drawline(x-8,y,x+8,y,LINE_HARD);
	display_drawline(x,y-8,x,y+8,LINE_HARD);
}

void
deposit_cursor(void)
{
	int x,y;

	if(!cursor_stat){
		fputs("Cursor not enabled\n",ttyout);
		return;
	}
	if(!cursor_in_win){
		fputs("Cursor not in window\n",ttyout);
		return;
	}
	if(box.status == 0){
		display_get_mouse(&x,&y);
		draw_cursor(x,y);
		box_add(x,y,0,0,1);
	}else{
		box.vis = 0;
		box_add(box.x,box.y,box.xw,box.yw,0);
		draw_box(LINE_HARD);
	}
}


void
ctext(char *s)
{
	if(opt_panel){
		display_panel_message(s);
	}
}

static void
cursor_syntax(void)
{
	fputs("Synopsis: manage cursors\n",ttyout);
	min_abbreviation();
	fputs("Related: box\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("cursor on            * turn cursor on\n",ttyout);
	fputs("cursor off           * turn cursor off\n",ttyout);
	fputs("cursor               * change from off/on to on/off\n",ttyout);
	fputs("cursor -p x y        * position cursor at (x,y)\n",ttyout);
	fputs("cursor -d            * deposit cursor\n",ttyout);
}

void
cmd_cursor(void)
{
	int x,y;
	int value;

	if(argc < 1 || argc > 4){
		cursor_syntax();
		return;
	}

	if(argc == 1){
		cursor_stat = !cursor_stat;
	}else{
		if(!strcmp("on",argv[1])){
			cursor_stat = 1;
		}else if(!strcmp("off",argv[1])){
			cursor_stat = 0;
		}else if(!strcmp("-d",argv[1])){
			deposit_cursor();
			return;
		}else if(!strcmp("-p",argv[1])){
			if(argc != 4){
				fputs("syntax: cursor -p x y\n",ttyout);
				return;
			} 
			x = parse_atof(argv[2]);
			if(parse_atof_error()) return;

			y = parse_atof(argv[3]);
			if(parse_atof_error()) return;

			if(x < 0 || y < 0 || x >= canvas_size_x || y >= canvas_size_y){
				fputs("Cursor position outside canvas\n",ttyout);
				return;
			}

			display_set_mouse(x,y);
			cursor_stat = 1;

			value = image16[y*canvas_size_x + x] - zero_offset;
			fprintf(ttyout,"(%d,%d) %d\n",x,y,value);
		}else{
			fputs("On or off please!\n",ttyout);
			return;
		}
	}
	cur_on_off();

	if(cursor_stat){
		fputs("Cursor on\n",ttyout); 
	}else{
		fputs("Cursor off\n",ttyout);
	}
}

void
cur_on_off(void)
{
	int x,y;

	if(cursor_stat){
		if(box.status){
			display_cross_off();
			if(cursor_in_win){
				if(!box.vis){
					display_get_mouse(&x,&y);
					box.x = x;
					box.y = y;
					draw_box(LINE_SOFT);
					box.vis = 1;
				}
			}
		}else{
			display_cross_on();
			if(box.vis){
				draw_box(LINE_ERASE);
				box.vis = 0;
			}
		}
	}else{
		display_cross_on();
		if(box.vis){
			draw_box(LINE_ERASE);
			box.vis = 0;
		}

		if(opt_panel){
			ctext("               ");
		}else{
			da2_x = 0;
			da2_y = canvas_size_y - font_height;
			da2_xw = font_width*CTEXT_LEN;
			da2_yw = font_height;
			load_pix(2);
		}
	}
}
