/*
display - handle all interfacing to the display system (X/xview in this case)
*/

#ident "$Id: display.c,v 1.48 2011/09/14 11:23:38 carl Exp $"

/*
	defines
*/
#define CMS_ALIGN 32		/* align background, 32 is for lut=64*/
#define SLIDER_SLOP 75		/* room needed on line slider for other atoms */
#define SLIDER_FINE_RANGE 2047	/* fine range for sliders */
#define MAIN_NOTIFY_CLIENT 1357 /* arbitrary signal id */
#define FILE_FILTER_STRING_LEN 50	/*length of file filter string*/
#define FILE_FILTER_DISPLAY_LEN 25	/*length of file filter display*/


/*
  xview include files
*/
#define OWTOOLKIT_WARNING_DISABLED
#include <xview/xview.h> 
#include <xview/panel.h> 
#include <xview/canvas.h>
#include <xview/tty.h>
#include <xview/termsw.h>
#include <xview/cursor.h>
#include <xview/icon.h>
#include <xview/svrimage.h>
#include <xview/cms.h>
#include <xview/font.h>
#include <xview/file_chsr.h>
#include <xview/hist.h>
#include <xview/notice.h>
#include <xview/file_list.h>
#include <xview/notify.h>

/*X11 libraries*/
#include <X11/Xutil.h>

/* standard includes */
#include <errno.h>
#include <netdb.h>
#include <signal.h>

#include "xpic.h"
/* gnu libraries */
#include <fnmatch.h>  //wildcard matching

#include <libgen.h>   //basename, dirname

#include <sys/stat.h> //file stat structure
#include <sys/types.h>
/* #include <termios.h> */
#include <unistd.h>

/*
	debugging stuff
*/
#ifdef DEBUG
#define WTEXT(x) parse_wtext(x)
#else
#define WTEXT(x)
#endif
/*
	stuff for timing commands
*/
#ifdef TIME
#include <sys/time.h>
#include <sys/resource.h>
static int etimeu,etimes;
static int stimeu,stimes;
static int utimeu,utimes;
static struct timeval timep;
static struct rusage usage;
#endif
/*
	stuff for mit's shared memory extension
*/
#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
static int rmajor,rminor;
static Bool shm_pixmap;
static XShmSegmentInfo shminfo;
#endif
/*
	statics for required for display subroutines
*/
static Display *dpy;				/* x display */
static GC canvas_gc;				/* gc for imaging */
static GC gc_l_src,gc_l_xor;			/* gc's for drawing lines in display*/
static GC gc_L;					/* gc for drawing lines in pixmap*/
static GC gc_a; 				/* gc for drawing text-display*/
static GC gc_A; 				/* gc for drawing text-pixmap*/
static Visual *vis;				/* display's visual */
static Window canvas_win;			/* XID for canvas */
static Window canvas_paint;			/* painting window for canvas */
static Pixmap pixmap;				/* pixmap "behind" canvas */
static XImage *xi;				/* X image on client */
static Frame BaseFrame;				/* Top level window frame */
static Frame ControlFrame;			/* window frame for tty and panel */
static Canvas DrawingCanvas;			/* Main drawing window for text */
static Tty tty;					/* input/ouput tile */
static Xv_Cursor cross_hair;			/* handle for cursor */
static Xv_Cursor no_cursor;			/* handle for blank */
static XImage *txi;				/* temp image for get/rel image */
static Panel_item message_item;			/* status line in panel */
static Panel_item message_win;			/* window slider message */
static Panel_item message_lev;			/* level slider message */
static Panel panel;				/* panel */
static Panel_item win_slider;			/* slider for window */
static Panel_item lev_slider;			/* slider for level */
static Panel_item cf_select;			/* coarse/fine select button */
static int cf_mode=0;				/* 0=fine 1=coarse w/l adj */
static Xv_Font font_panel;			/* font for panel */
static int frfd[2],tofd[2];			/* pipe to/from keyboard proc */
static Region region;				/* for annotation clipping */
static int mitshm;				/* 0=no, 1=image, 2=image+pixmap */
static int canvas_shm;				/* 1=canvas uses shared memory */
static int canvas_id;				/* canvas's shared memory id */
static DPIXEL32 red_mask,green_mask,blue_mask;	/* bit masks for color table */
static VisualID visid;				/* id of default visual */
static File_chooser	open=0;			/*open dialog*/
static History_list	open_history;		/*history of opened files*/
static Attr_attribute	FILE_TYPE_KEY;		/*key for file type chooser*/
static Attr_attribute	REFRESH_BTN_KEY;	/*key for refresh button in file chooser*/
static Attr_attribute	PERSIST_KEY;		/*key for persistance checkbox*/
static Attr_attribute	ENV_MENU_KEY;		/*key for env drop down menu*/
static Attr_attribute	FILE_FILTER_KEY;	/*key for file filter text box*/
static Attr_attribute	AUTO_OPEN_KEY;		/*key for auto open check box*/
static Attr_attribute	HIDDEN_FILE_KEY;	/*key for hidden file check box*/
static Attr_attribute	MULTISELECT_KEY;	/*key for multiselect check box*/
static short		persist=1;		/*persistance setting for open dialog*/
static unsigned short 	env_open_to;		/*environment for browser to open file into*/
static char 		usrfilter[FILE_FILTER_STRING_LEN];  /*stores file filter*/
static short		autoopen=1;		/*auto open flag*/
static short		multiselect=0;		/*multiple file select flag*/
static int		dest_itype=I_MI;	/*open to file type designator*/
static int (*xview_file_list_callback)(Panel_item item, char *dir, char* file, Xv_opaque client_data, Panel_list_op op, Event *event, int row);

static int tty_pid;	// PID for tty subwindow

/* non-static global */
Panel_item env_message_item;

/*
	os specific stuff
*/
int ttysw_output(Tty tty,char *buf,int len); /* prototype missing in linux and solaris */
/*
	arrays
*/
static unsigned short icon_bits[] = {		/* icon for xpic */
#include "xpic.icon"
};

static unsigned short icon_control_bits[] = {	/* icon for xpic's control */
#include "xpic_ctr.icon"
};

static unsigned short cursor_bits[] = {		/* cross hair */
/* Format_version=1, Width=16, Height=16, Depth=1, Valid_bits_per_item=16 */
	0x0080, 0x0080, 0x0080, 0x0080, 0x0080, 0x0080,
	0x0080, 0xFF7F, 0x0080, 0x0080, 0x0080, 0x0080,
	0x0080, 0x0080, 0x0080, 0x0080
};

static unsigned short no_cursor_bits[] = {		/* blank cursor */
/* Format_version=1, Width=16, Height=16, Depth=1, Valid_bits_per_item=16 */
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
	0x0000, 0x0000, 0x0000, 0x0000
};

static void
refresh_file_list(Panel_item item, Event *event)
{
	xv_set(open,	FILE_CHOOSER_UPDATE, 	NULL);

	//prevent dialog from closing
	xv_set(item,	PANEL_NOTIFY_STATUS,	XV_ERROR,
			NULL);
}

static void
open_file_type(Panel_item item, int value, Event *event)
{
	char msg[40];    /*message string for file chooser footer -> 40 = max length of message*/

	switch(value){
		case 0:
		//mi file
		strcpy(usrfilter,"*.mi");
		strcpy(msg, "Listing *.mi files | Type set to MI");
		dest_itype = I_MI;
		break;
#ifdef GZMI
		case 1:
		//gzmi file
		strcpy(usrfilter,"*.mi.gz");
		strcpy(msg, "Listing *.mi.gz files | Type set to GZMI");
		dest_itype = I_GZMI;
		break;
#endif
#ifdef GZMI
		case 2:
#else
	        case 1:
#endif
		//lm file
#ifdef MERLIN
		strcpy(msg, "Listing *.l3f* files | Type set to LM");
		strcpy(usrfilter,"*.l3f");
#else
		strcpy(msg, "Listing *.lm files | Type set to LM");
		strcpy(usrfilter,"*.lm");
#endif
		dest_itype = I_LM;
		break;
#ifdef GZMI
		case 3:
#else
	        case 2:
#endif
		//all files
		strcpy(usrfilter,"*");
#ifdef GZMI
		strcpy(msg, "Listing all files | Type set to GZMI");
		dest_itype = I_GZMI;
#else
		strcpy(msg, "Listing all files | Type set to MI");
		dest_itype = I_MI;
#endif
		break;
		default:
		strcpy(usrfilter,"");
		strcpy(msg, "");
	}

	xv_set(open,
		FRAME_LEFT_FOOTER, 		msg,
		FILE_CHOOSER_UPDATE,
		NULL);
}

static int
my_open_callback(File_chooser fc, char *path, char *file, Xv_opaque client_data )
{
	char temp[MAXPATHLEN];
	int i,cnt=0;
	char **ar;
	File_list open_file_list;
	struct stat dirent;

	if(!fc || !path)
		return XV_ERROR;

	xv_set(fc, FRAME_BUSY, TRUE, NULL);

	if(current_env!=env_open_to){
		env_push(current_env);
		env_pop(env_open_to,0);
		current_env=env_open_to;
		write_env_panel(current_env);
	}
	itype=dest_itype;

	if(multiselect){
		open_file_list = xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_FILE_LIST_CHILD);
		ar = malloc(sizeof(char*)*xv_get(open_file_list,PANEL_LIST_NROWS));
		if(ar==NULL){
			fputs("my_open_callback: Error: insufficient memory\n",ttyout);
			fflush(ttyout);
			return XV_ERROR;
		}
		ar[0]=strdup("");  //need empty 0th spot
		if(ar[0]==NULL){
			fputs("my_open_callback: Error: insufficient memory\n",ttyout);
			fflush(ttyout);
			free(ar);
			return XV_ERROR;
		}

		/*loop through selected entries in list, stop if we run out of entries or environments*/
		for(i=xv_get(open_file_list,PANEL_LIST_FIRST_SELECTED);
		    i>0 && cnt < ENV_N-current_env;
		    i=xv_get(open_file_list,PANEL_LIST_NEXT_SELECTED,i))
		{
			strncpy(temp,(char*)xv_get(open_file_list,FILE_LIST_DIRECTORY),MAXPATHLEN);
			strncat(temp,"/",MAXPATHLEN-strlen(temp));
			strncat(temp,(char*)xv_get(open_file_list,PANEL_LIST_STRING,i),MAXPATHLEN-strlen(temp));
			if(stat(temp,&dirent)!=0){
				/*shouldnt get here, since the file_chooser just found this file*/
				fprintf(ttyout,"Error %d: %s",errno,strerror(errno));
				return XV_ERROR;
			}
			else if(!S_ISDIR(dirent.st_mode)){	//ignore directories
				if((ar[++cnt]=strdup(temp))==NULL)
				{
					fputs("my_open_callback: Error: insufficient memory\n",ttyout);
					fflush(ttyout);
					//free allocated memory before error return
					for(i=0;i<cnt;i++)
						free(ar[i]);
					free(ar);
					return XV_ERROR;
				}
			}
		}
		expand_open(cnt+1, ar);
		for(i=0;i<=cnt;i++){
			free(ar[i]);
		}
		free(ar);
		fflush(ttyout);
		sprintf(temp,"Opened %d files",cnt);
		xv_set(fc,	FRAME_BUSY,		FALSE,
				FRAME_LEFT_FOOTER,	temp,
				NULL);

	}
	else{
		strncpy(temp,path,MAXPATHLEN);
		if( chdir(dirname(temp))!=0  ){
			fprintf(ttyout,"Error: %s\n",strerror(errno));
			return XV_ERROR;
		}
		strncpy(temp,path,MAXPATHLEN);
		strcpy(ifn, basename(temp));
		fprintf(ttyout, "Browser open: %s Env: %d\n",path,current_env+1);
		open_file();
		fflush(ttyout);

		if(persist){
			sprintf(temp,"File Opened: %s",file);
			xv_set(fc,	FRAME_BUSY,		FALSE,
					FRAME_LEFT_FOOTER,	temp,
					NULL);
		}
	}
	if(autoopen){
		images("1");
	}
		
	if (multiselect){	/*turn off multiopen if its on*/
		File_list open_file_list = xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_FILE_LIST_CHILD);
		int rows,i;
	
		xv_set((Panel_item) xv_get(fc,XV_KEY_DATA, MULTISELECT_KEY), PANEL_VALUE, 0, NULL);
			
		rows = xv_get(open_file_list, PANEL_LIST_NROWS);
		xv_set(open_file_list,XV_SHOW,FALSE,NULL);
		for(i=0;i<rows;i++)
			xv_set(open_file_list,PANEL_LIST_SELECT,i,FALSE,NULL);
		xv_set(open_file_list,XV_SHOW,TRUE,NULL);
		xv_set(open_file_list,PANEL_LIST_SELECT,	0,	TRUE,
				PANEL_LIST_DO_DBL_CLICK,	TRUE,
				NULL);
		xv_set(xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_OPEN_BUTTON_CHILD),	//make sure open button is enabled
			PANEL_BUSY,	FALSE,
			NULL);

		multiselect=0;
		xv_set(xv_get(fc, FILE_CHOOSER_CHILD, FILE_CHOOSER_FILE_LIST_CHILD), PANEL_CHOOSE_ONE, TRUE,
			PANEL_CHOOSE_NONE, FALSE, NULL);
	}
	cwd2kb = 1;
	xv_set(tty, WIN_SET_FOCUS, NULL); /*put kb focus in tty*/
	return (persist?XV_ERROR:XV_OK);
}

static int
my_exten_func(File_chooser fc, Rect *frame_rect, Rect *exten_rect, int left_edge, int right_edge, int max_height)
{
	Panel_item types = (Panel_item) xv_get(fc,XV_KEY_DATA, FILE_TYPE_KEY);
 	Panel_item refsh = (Panel_item) xv_get(fc,XV_KEY_DATA, REFRESH_BTN_KEY);
	Panel_item perst = (Panel_item) xv_get(fc,XV_KEY_DATA, PERSIST_KEY);
	Panel_item envmu = (Panel_item)	xv_get(fc,XV_KEY_DATA, ENV_MENU_KEY);
	Panel_item filtr = (Panel_item) xv_get(fc,XV_KEY_DATA, FILE_FILTER_KEY);
	Panel_item aopen = (Panel_item) xv_get(fc,XV_KEY_DATA, AUTO_OPEN_KEY);
	Panel_item fdhid = (Panel_item) xv_get(fc,XV_KEY_DATA, HIDDEN_FILE_KEY);
	Panel_item musel = (Panel_item) xv_get(fc,XV_KEY_DATA, MULTISELECT_KEY);
	int item_width;

 	item_width = MAX3((int) xv_get(types, XV_WIDTH),(int) xv_get(refsh, XV_WIDTH),(int) xv_get(perst, XV_WIDTH));
#ifndef GZMI
	item_width += 55;
#endif

//baseline row
	xv_set(types,
		XV_X, (frame_rect->r_width - item_width) / 2 - 5,
		XV_Y, exten_rect->r_top - 5,
		PANEL_PAINT, PANEL_NONE,
		NULL);
//second row
	xv_set(perst,
		XV_X, (frame_rect->r_width - item_width) / 2 - 25,
		XV_Y, exten_rect->r_top + 23,
		PANEL_PAINT, PANEL_NONE,
		NULL);
	xv_set(envmu,
		XV_X, (frame_rect->r_width - item_width) / 2 + 35,
		XV_Y, exten_rect->r_top + 23,
		PANEL_PAINT, PANEL_NONE,
		NULL);
 	xv_set(refsh, 
 		XV_X, (frame_rect->r_width - item_width) / 2 + 220,
 		XV_Y, exten_rect->r_top + 23,
 		PANEL_PAINT, PANEL_NONE, 
 		NULL); 
//third row
	xv_set(filtr,
		XV_X, (frame_rect->r_width - item_width) / 2 - 10,
		XV_Y, exten_rect->r_top + 50,
		PANEL_PAINT, PANEL_NONE,
		NULL);
//fourth row
	xv_set(aopen,
		XV_X, (frame_rect->r_width - item_width) / 2 - 10,
		XV_Y, exten_rect->r_top + 72,
		PANEL_PAINT, PANEL_NONE,
		NULL);

	xv_set(fdhid,
		XV_X, (frame_rect->r_width - item_width) / 2 + 90,
		XV_Y, exten_rect->r_top + 72,
		PANEL_PAINT, PANEL_NONE,
		NULL);

	xv_set(musel,
		XV_X, (frame_rect->r_width - item_width) / 2 + 206,
		XV_Y, exten_rect->r_top + 72,
		PANEL_PAINT, PANEL_NONE,
		NULL);


	return -1;
}

static void
open_fpersist(Panel_item item, int value, Event *event)
{
	persist=value;
	xv_set(open, FRAME_LEFT_FOOTER, (value?"Browser will remain open":""), NULL);
	xv_set(xv_get(open,FILE_CHOOSER_CHILD,FILE_CHOOSER_CANCEL_BUTTON_CHILD),
		PANEL_LABEL_STRING,	(value?"Close":"Cancel"),
		NULL);
}

static void
open_env_selected(Panel_item item, int value, Event *event)
{
	char temp[32];	//value + 1 not greater than 10
	env_open_to=value;

	sprintf(temp,"Opening files to Environment %d",value+1);
	xv_set(open, FRAME_LEFT_FOOTER, temp, NULL);
}

static int
open_filter_func(Panel_item item, Event *event)
{
	char temp[FILE_FILTER_STRING_LEN+16];// length of "Filter Applied : "+usrfilter

	strcpy(usrfilter, (char *)xv_get(item, PANEL_VALUE));
	sprintf(temp,"Filter Applied: %s",usrfilter);
	xv_set(open,			FILE_CHOOSER_UPDATE,
		FRAME_LEFT_FOOTER,	temp,
		NULL);
	return PANEL_NONE; //take no action on enter/tab

}

static File_chooser_op
open_filter_apply(File_chooser fc, char *path, struct stat *stats,
						  File_chooser_op matched, Server_image *glyph,
						  Xv_opaque client_data, Server_image *mask_glyph)
{
	char *filename;
	char *pathcopy;
	int errcode;

	pathcopy = strdup(path);
	filename = basename(pathcopy);

	//code for wildcard matching
	if((errcode=fnmatch(usrfilter,filename,FNM_PATHNAME))!=0)
	{
		if(errcode!=FNM_NOMATCH)
			xv_set(open,	FRAME_LEFT_FOOTER, "Unknown Error in fnmatch",NULL);
		free(pathcopy);
		return FILE_CHOOSER_IGNORE;
	}

	free(pathcopy);
	return FILE_CHOOSER_ACCEPT;
}

static void
open_set_auto_open(Panel_item item, int value, Event *event)
{
	autoopen = value;
	xv_set(open,	FRAME_LEFT_FOOTER, (value?"Auto Open Avtivated":"Auto Open Deactivated"),NULL);
}

static void
open_set_show_hidden(Panel_item item, int value, Event *event)
{
	xv_set(open,
		FILE_CHOOSER_SHOW_DOT_FILES,	(value?TRUE:FALSE),
		FRAME_LEFT_FOOTER, (value?"Showing hidden files":"Hiding hidden files"),
		FILE_CHOOSER_UPDATE,
		NULL);
}

static void
open_set_multiselect(Panel_item item, int value, Event *event)
{
	File_list open_file_list;
	int rows,i;

	open_file_list = xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_FILE_LIST_CHILD);
	xv_set(open_file_list, PANEL_CHOOSE_ONE, !(multiselect=value),
				PANEL_CHOOSE_NONE,	FALSE,
				PANEL_LIST_DO_DBL_CLICK,TRUE,
				NULL);
	xv_set(open, FRAME_LEFT_FOOTER, (value?"Select multiple files with mouse":"Multiple selection disabled"),NULL);

	if(!value){	//turn off
		rows = xv_get(open_file_list, PANEL_LIST_NROWS);
		xv_set(open_file_list,XV_SHOW,FALSE,NULL);
		for(i=0;i<rows;i++)
			xv_set(open_file_list,PANEL_LIST_SELECT,i,FALSE,NULL);
		xv_set(open_file_list,XV_SHOW,TRUE,NULL);
		xv_set(xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_OPEN_BUTTON_CHILD),	//make sure open button is enabled
			PANEL_BUSY,	FALSE,
			NULL);
	}
	else {		//turn on
		xv_set(open_file_list,
				PANEL_LIST_SELECT,	0,	FALSE,
				PANEL_LIST_DO_DBL_CLICK,	FALSE,
				NULL);
		i=xv_get(open_file_list, PANEL_LIST_FIRST_SELECTED);
		if(i<0) //if nothing is now selected, disable open
			xv_set(xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_OPEN_BUTTON_CHILD),
				PANEL_BUSY,	TRUE,
				NULL);
	}

}

//fix callback for handling multiopen
static int
list_callback( item, dir, file, client_data, op, event, row )
     File_list item;
     char *dir;
     char *file;
     Xv_opaque client_data;
     Panel_list_op op;
     Event *event;
     int row;
{
	if(multiselect){
		int firstrow = xv_get(item, PANEL_LIST_FIRST_SELECTED);
		Panel_item open_btn = xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_OPEN_BUTTON_CHILD);
		if( firstrow < 0){
			xv_set(open_btn,
				PANEL_BUSY,	TRUE,
				NULL);
		}
		else{
			xv_set(open_btn,
				PANEL_BUSY,	FALSE,
				NULL);
		}
		if(op==PANEL_LIST_OP_DBL_CLICK){
			op=PANEL_LIST_OP_DESELECT;
		}
	}
	return (*xview_file_list_callback)(item, dir, file, client_data, op, event, row);
}

static void
do_browser(Panel_item item, Event *event){
	Panel_item	open_ftype=0;
	Panel_item	open_refresh=0;
	Panel_item	open_persist=0;
	Panel_item	open_env_menu=0;
	Panel_item	open_fltr_txt=0;
	Panel_item	open_auto_open=0;
	Panel_item	open_show_hidden=0;
	Panel_item	open_multiselect=0;
	Panel_item	open_file_list=0;
	Panel		open_cmdpanel=0;
	int 		r1_width=0;
	int		r1_height=0;
	int		r2_width=0;
	int		r2_height=0;
	int		r3_width=0;
	int		r3_height=0;
	int		r4_width=0;
	int		r4_height=0;
	int		frame_width=0;
	int		frame_height=0;
	char		pathname[MAXPATHLEN];

	if ( !open ) {

		env_open_to=current_env;
		strcpy(usrfilter, "*");
		if(NULL == getcwd(pathname,MAXPATHLEN)) perror("xpic: display.c: do_browser: getcwd error");

		open_history
		  = xv_create(ControlFrame,				HISTORY_LIST,
				HISTORY_ROLLING_MAXIMUM,		15,
				HISTORY_ADD_FIXED_ENTRY,		"Home",		"~",
				HISTORY_ADD_FIXED_ENTRY,		"Startup Path",	pathname,
				HISTORY_ADD_FIXED_ENTRY,		"Root Directory", "/",
				HISTORY_DUPLICATE_LABELS,		FALSE,
				HISTORY_DUPLICATE_VALUES,		FALSE,
				NULL);

			strcpy(usrfilter,"*.mi");
		open
		  = xv_create(ControlFrame,				FILE_CHOOSER_OPEN_DIALOG,
				XV_LABEL,				"xpic: Open",
				FILE_CHOOSER_NOTIFY_FUNC,		my_open_callback,
				FILE_CHOOSER_ABBREV_VIEW,		TRUE,				//dont show nonmatching files
				FILE_CHOOSER_DIRECTORY,			pathname,			//set browser to cwd
				FILE_CHOOSER_HISTORY_LIST,		open_history,			//set history list
				FILE_CHOOSER_FILTER_MASK,		FC_MATCHED_FILES_MASK,
				FILE_CHOOSER_FILTER_FUNC,		open_filter_apply,
				XV_X,canvas_size_x+11,XV_Y,0,							//position of dialog on screen
			WIN_FONT, font_panel,
				NULL);
				
		open_file_list = xv_get(open, FILE_CHOOSER_CHILD, FILE_CHOOSER_FILE_LIST_CHILD);
		xv_set(xv_get(open,FILE_CHOOSER_CHILD,FILE_CHOOSER_GOTO_MESSAGE_CHILD),
				PANEL_LABEL_STRING,	"xpic File Browser",
				PANEL_LABEL_BOLD, 	TRUE,
				NULL);
				
		xview_file_list_callback = (int(*)(File_list,char*,char*,Xv_opaque,Panel_list_op,Event*,int))xv_get(open_file_list,PANEL_NOTIFY_PROC);
		
		xv_set(open_file_list,
				PANEL_NOTIFY_PROC,		list_callback,
				PANEL_LIST_DO_DBL_CLICK,	TRUE,
				NULL);

		open_cmdpanel = xv_get(open, FRAME_CMD_PANEL);


		open_ftype
		  = xv_create(open_cmdpanel,		PANEL_CHOICE,
				PANEL_LABEL_STRING,	"File Type",
#ifdef GZMI
#ifdef MERLIN
			      PANEL_CHOICE_STRINGS,	"*.mi","*.mi.gz","*.l3f*","All Files",NULL,
#else
			      PANEL_CHOICE_STRINGS,	"*.mi","*.mi.gz","*.lm","All Files",NULL,
#endif
#else
#ifdef MERLIN
			      PANEL_CHOICE_STRINGS,	"*.mi","*.l3f*","All Files",NULL,
#else
			      PANEL_CHOICE_STRINGS,	"*.mi","*.lm","All Files",NULL,
#endif

#endif
				PANEL_NOTIFY_PROC,	open_file_type,
				NULL);

 		open_refresh 
 		  = xv_create(open_cmdpanel,		PANEL_BUTTON, 
 				PANEL_LABEL_STRING,	"Refresh", 
 				PANEL_NOTIFY_PROC,	refresh_file_list, 
 				NULL);

		open_persist
		  = xv_create(open_cmdpanel,		PANEL_CHECK_BOX,
				PANEL_LABEL_STRING,	"Stay",
				PANEL_CHOICE_STRINGS,	" ", NULL,
				PANEL_NOTIFY_PROC,	open_fpersist,
				PANEL_VALUE,		TRUE,
				NULL);
		open_fpersist(0,1,0);

		open_env_menu
		  = xv_create(open_cmdpanel,		PANEL_CHOICE_STACK,
				PANEL_LAYOUT,			PANEL_HORIZONTAL,
				PANEL_LABEL_STRING,	"Environment",
				PANEL_CHOICE_STRINGS,	"Env. 1", "Env. 2", "Env. 3", "Env. 4", "Env. 5",
							"Env. 6", "Env. 7", "Env. 8", "Env. 9", "Env. 10",
							NULL,
				PANEL_VALUE,			env_open_to,
				PANEL_NOTIFY_PROC,	open_env_selected,
				NULL);

		open_fltr_txt
		  = xv_create(open_cmdpanel,		PANEL_TEXT,
				PANEL_LABEL_STRING,	"File Filter:",
				PANEL_VALUE_DISPLAY_LENGTH,	FILE_FILTER_DISPLAY_LEN,
				PANEL_VALUE_STORED_LENGTH,	FILE_FILTER_STRING_LEN,
				PANEL_NOTIFY_LEVEL,	PANEL_SPECIFIED,
				PANEL_NOTIFY_STRING,	"\n\r\t",
				PANEL_NOTIFY_PROC,	open_filter_func,
				PANEL_VALUE,		"*",			//initial value of filter is *
				NULL);

		open_auto_open
		  = xv_create(open_cmdpanel,		PANEL_CHECK_BOX,
			PANEL_LABEL_STRING,		"Auto Open",
				PANEL_NOTIFY_PROC,	open_set_auto_open,
				PANEL_VALUE,		TRUE,
				NULL);

		open_show_hidden
		  = xv_create(open_cmdpanel,		PANEL_CHECK_BOX,
				PANEL_LABEL_STRING,	"Show Hidden",
				PANEL_NOTIFY_PROC,	open_set_show_hidden,
				NULL);

		open_multiselect
		  = xv_create(open_cmdpanel,		PANEL_CHECK_BOX,
				PANEL_LABEL_STRING,	"Multiselect",
				PANEL_NOTIFY_PROC,	open_set_multiselect,
				NULL);

 		r1_width = (int) xv_get(open_ftype, XV_WIDTH);
 		r1_height= (int) xv_get(open_ftype, XV_HEIGHT);

		r2_width = (int) xv_get(open_persist, XV_WIDTH) + (int) xv_get(open_env_menu, XV_WIDTH) + (int) xv_get(open_refresh, XV_WIDTH);;
		r2_height= (int) MAX3(xv_get(open_persist, XV_HEIGHT), (int) xv_get(open_env_menu, XV_HEIGHT),(int) xv_get(open_refresh, XV_HEIGHT));

		r3_width = (int) xv_get(open_fltr_txt, XV_WIDTH);
		r3_height= (int) xv_get(open_fltr_txt, XV_HEIGHT);

		r4_width = (int) xv_get(open_auto_open, XV_WIDTH) + (int) xv_get(open_show_hidden, XV_WIDTH) +
			   (int) xv_get(open_multiselect, XV_WIDTH);
		r4_height= (int) MAX3((int) xv_get(open_auto_open, XV_HEIGHT), (int) xv_get(open_show_hidden, XV_HEIGHT),
			   (int) xv_get(open_multiselect, XV_HEIGHT));

		frame_width = (int) xv_get(open, XV_WIDTH);
		frame_height= (int) xv_get(open, XV_WIDTH);

		xv_set(open,
			XV_WIDTH,	CRC_MAX(frame_width,
						(CRC_MAX(MAX3(r1_width,r2_width,r3_width),
							r4_width) + xv_cols(open_cmdpanel,4))),
			XV_HEIGHT,	CRC_MAX(frame_height + r1_height + r2_height + r3_height + r4_height,canvas_size_y+125),
			NULL);

		xv_get(open, FRAME_MIN_SIZE, &frame_width, &frame_height);
		xv_set(open,
			FRAME_MIN_SIZE,	CRC_MAX(frame_width,
						(CRC_MAX(MAX3(r1_width,r2_width,r3_width),
							r4_width) + xv_cols(open_cmdpanel,4))),
						frame_height + r1_height + r2_height + r3_height +r4_height-10,
			NULL);

		//generate keys
		FILE_TYPE_KEY = xv_unique_key();
		REFRESH_BTN_KEY = xv_unique_key();
		PERSIST_KEY = xv_unique_key();
		ENV_MENU_KEY = xv_unique_key();
		FILE_FILTER_KEY = xv_unique_key();
		AUTO_OPEN_KEY = xv_unique_key();
		HIDDEN_FILE_KEY = xv_unique_key();
		MULTISELECT_KEY = xv_unique_key();

		xv_set(open,
				FILE_CHOOSER_EXTEN_HEIGHT,	r1_height + r2_height + r3_height +r4_height-8,
				FILE_CHOOSER_EXTEN_FUNC,	my_exten_func,
				XV_KEY_DATA, FILE_TYPE_KEY,	open_ftype,
 				XV_KEY_DATA, REFRESH_BTN_KEY, 	open_refresh,
				XV_KEY_DATA, PERSIST_KEY,	open_persist,
				XV_KEY_DATA, ENV_MENU_KEY,	open_env_menu,
				XV_KEY_DATA, FILE_FILTER_KEY,	open_fltr_txt,
				XV_KEY_DATA, AUTO_OPEN_KEY,	open_auto_open,
				XV_KEY_DATA, HIDDEN_FILE_KEY,	open_show_hidden,
				XV_KEY_DATA, MULTISELECT_KEY,	open_multiselect,
				NULL);
	}else{
		env_open_to=current_env;
		open_env_menu = (Panel_item) xv_get(open, XV_KEY_DATA, ENV_MENU_KEY);
		xv_set(open_env_menu, PANEL_VALUE, env_open_to, NULL);
		xv_set(open, FILE_CHOOSER_UPDATE, NULL);
	}

}
static void
btn_open_mi(Panel_item item, Event *event)
{

	do_browser(item, event);
	if(xv_get(open,XV_SHOW)){	//if the browser is open hide it
		xv_set(open, XV_SHOW, FALSE, NULL);
	}
	else xv_set(open, XV_SHOW, TRUE, NULL);
}


static void
browse_syntax(void)
{
	fputs("Synopsis: opens graphical file browser\n",ttyout);
	fputs("Related: open, file, reopen, images, status, header\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("browse\t\t\t* open file browser in cwd\n",ttyout);
	fputs("browse path\t\t* open file browser in path\n",ttyout);
	fputs("Browser Features:\n",ttyout);
	fputs("\tGo To\t\t* Right-click to access history menu\n",ttyout);
	fputs("\tGo To Path\t* Enter a path to browse to\n",ttyout);
	fputs("\tCurrent Folder\t* Displays current folder\n",ttyout);
	fputs("\tFile List\t* Select files to open, folders to browse\n",ttyout);
	fputs("\tFile Type\t* Select a file type filter\n",ttyout);
	fputs("\tStay\t\t* Set browser to stay/disapear on file open\n",ttyout);
	fputs("\tEnvironment\t* Right click to access environment open-to\n\t\t\t* menu\n",ttyout);
	fputs("\tFile Filter\t* Type a filter string with shell style \n\t\t\t* \
wildcard characters, \"*?[]^\" and press\n\t\t\t* enter or tab\n",ttyout);
	fputs("\tAuto Open\t* Set browser to automatically display image\n\t\t\t* on open\n",ttyout);
	fputs("\tShow Hidden\t* Set browser to show/hide hidden files\n",ttyout);
	fputs("\tMultiselect\t* Enable multiselect behavior.  Multilpe \n\t\t\t* \
files may be selected at once.  Open is\n\t\t\t* disabled if no files are selected. \n\t\t\t* \
Note: Doubleclick is disabled while in\n\t\t\t* multiselect mode\n",ttyout);
	fflush(ttyout);
}

void
cmd_browse()
{
	struct stat buf;

	if(argc < 1 || argc > 2){
		browse_syntax();
		return;
	}
	
	do_browser(0,0);
	if(argc == 2){
		if(stat(argv[1],&buf)<0){
			fprintf(ttyout,"browse: %s: %s\n",argv[1],strerror(errno));
			fflush(ttyout);
			return;
		}else if(!S_ISDIR(buf.st_mode)){
			fprintf(ttyout,"browse: %s is not a directory\n",argv[1]);
			fflush(ttyout);			
			return;
		}else{
			xv_set(open,FILE_CHOOSER_DIRECTORY,argv[1],NULL);
		}
	}	
	xv_set(open, XV_SHOW, TRUE, NULL);
}


static void
display_gethost_error(struct hostent *hostp,char *server)
{
	if(hostp != NULL) return;

	fprintf(stderr,"xpic: gethostbyname failure for host: %s\n",server);
	fprintf(stderr,"check network availability and /etc/hosts\n");
	exit(1);
}

static DPIXEL32
pixel32_swap(DPIXEL32 i)
{
	union{
		DPIXEL32 l;
		char c[4];
	} in, out;

	in.l = i;

	out.c[0] = in.c[3];
	out.c[1] = in.c[2];
	out.c[2] = in.c[1];
	out.c[3] = in.c[0];

	return(out.l);
}

static char *
classcolor(int class)
{
	switch(class){

	case StaticGray:
		return("StaticGray");
	case GrayScale:
		return("GrayScale");
	case StaticColor:
		return("StaticColor");
	case PseudoColor:
		return("PseudoColor");
	case TrueColor:
		return("TrueColor");
	case DirectColor:
		return("DirectColor");
	default:
		return("Unknown");
	}
}

static int
hostbyteorder(void)
{
	union{
		unsigned short int l;
		unsigned char c[2];
		} u;

	u.l = 0xFACE;

	if(u.c[0] == 0xCE && u.c[1] == 0xFA){
		return(LSBFirst);
	}else if(u.c[0] == 0xFA && u.c[1] == 0xCE){
		return(MSBFirst);
	}else{
		fprintf(stderr,"xpic: unknown host byte order: %x %x\n",
			u.c[0],u.c[1]);
		exit(1);
	}
}

static char *
byteorder(int order)
{
	switch(order){

	case MSBFirst:
		return("MSBFirst");
	case LSBFirst:
		return("LSBFirst");
	default:
		return("Unknown");
	}
}

static int 
color(int c)
{
	switch(c){

	case 0:
		return 0;
	case 1:
		return 171;
	case 2:
		return 213;
	case 3:
		return 255;
	default:
		parse_er("xpic/color: should not be here");
		exit(0);
	}
	return(-1);   // make compiler happier; should not be here
}

static void
btn_measure(item,event)
Panel_item item;
Event *event;
{
	func_measure();
	fflush(ttyout);
}

static void
btn_scroll_up(item,event)
Panel_item item;
Event *event;
{
	func_scroll(1,1);
	fflush(ttyout);
}

static void
btn_scroll_down(item,event)
Panel_item item;
Event *event;
{
	func_scroll(-1,1);
	fflush(ttyout);
}

static void
btn_scroll_left(item,event)
Panel_item item;
Event *event;
{
	func_scroll(1,2);
	fflush(ttyout);
}

static void
btn_scroll_right(item,event)
Panel_item item;
Event *event;
{
	func_scroll(-1,2);
	fflush(ttyout);
}

static void
btn_next(item,event)
Panel_item item;
Event *event;
{
	next_prev(1);
}

static void
btn_previous(item,event)
Panel_item item;
Event *event;
{
	next_prev(-1);
}

static Notify_value 
track_cursor(window,event,arg)
Xv_window window;
Event *event;
Notify_arg arg;
{

	char line[80];
	int x=0,y=0;		/* make ansi happier w/ assignment */
	int value;
	static int adjust;


#ifdef DEBUG_EVENT
	printf("%d %d\n",event_id(event),event_action(event));
#endif
				/* support xpic_button */
	if(cursor_in_win && event_is_down(event) &&
	  (event_action(event) == ACTION_ADJUST ||
	  event_action(event) == ACTION_MENU ||
	  event_action(event) == ACTION_SELECT)){
		switch(event_action(event)){
			case ACTION_SELECT:	x=0; break;
			case ACTION_ADJUST:	x=1; break;
			case ACTION_MENU:	x=2; break;
		}
		if(port_button_proc(x)) return(NOTIFY_DONE);
	}

				/* pick up adjust button's state for drag */
	if(event_action(event) == WIN_REPAINT){
		return(NOTIFY_DONE);
	}else if(event_action(event) == ACTION_ADJUST){
		adjust = event_is_down(event);

				/* annotate or print cross info in tty */
	}else if(!cursor_stat && box.status == 0 &&
	 event_action(event) == ACTION_SELECT && event_is_down(event)){
		x = event_x (event);
		if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
		y = event_y (event);
		if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);
		if(note == 1){
			annotate_add(x,y,note_text);
			note = 0;
		}else{
			value = image16[y*canvas_size_x + x] - zero_offset;
			box.lx = x;
			box.ly = y;
			sprintf(line,"(%4d,%4d) %5d\n",x,y,value);
			display_tty_out(line);
		}
				/* track cursor outside of window */
	}else if((!box.status|| !cursor_stat)&& event_id(event) ==LOC_WINENTER){ 
		cursor_in_win = 1;
	}else if((!box.status|| !cursor_stat) && event_id(event) ==LOC_WINEXIT){ 
		cursor_in_win = 0;
	}else if(!cursor_stat){			/* need cursor for more proc */
		return(NOTIFY_IGNORED);
	}else if(box.status == 0){		/* handle cross hair & text*/
		x = event_x (event);
		if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
		y = event_y (event);
		if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);

						/* display annotation */
		if(event_action(event) == ACTION_SELECT && note && event_is_down(event)){
			annotate_add(x,y,note_text);
			note = 0;
		}else{				/* print location and value */
			value = image16[y*canvas_size_x + x] - zero_offset;
			sprintf(line,"(%4d,%4d) %5d",x,y,value);
			ctext(line);
			if((event_action(event) == ACTION_SELECT || event_action(event) == ACTION_MENU)
			 && event_is_down(event)){
				push_cursor(x,y);
				box.lx = x;
				box.ly = y;
				strcat(line,"\n");
				display_tty_out(line);
				if(event_action(event) == ACTION_MENU){
					deposit_cursor();
				}
			}
		}
	}else{					/* handle box cursor */
		x = event_x (event);
		y = event_y (event);

		if(event_id(event) == LOC_WINEXIT){
			cursor_in_win = 0;
			if(box.vis){
				draw_box(LINE_ERASE);
				box.vis = 0;
			}
		} else if(event_id(event) == LOC_WINENTER){
			cursor_in_win = 1;
			if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
			if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);
			box.x = x;
			box.y = y;
			if(!box.vis){
				draw_box(LINE_SOFT);
				box.vis = 1;
			}
		} else if(event_id(event) == LOC_MOVE){
			x = event_x (event);
			y = event_y (event);
			if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
			if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);
			if(box.vis){
				draw_box(LINE_ERASE);
				box.vis = 0;
			}
			box.x = x;
			box.y = y;
			if(!box.vis){
				draw_box(LINE_SOFT);
				box.vis = 1;
			}
			sprintf(line,"(%4d,%4d)",x,y);
			ctext(line);
		}else if(event_id(event) == LOC_DRAG && adjust){
			x = event_x (event);
			if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
			y = event_y (event);
			if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);
			if(box.vis){
				// jeep - editor tag
				draw_box(LINE_ERASE);
				box.vis = 0;
			}
			box.xw = 2 * abs(x - box.x);
			if(box.xw < 2)box.xw = 2;
			box.yw = 2 * abs(y - box.y);
			if(box.yw < 2)box.yw = 2;
			if(!box.vis){
				draw_box(LINE_SOFT);
				box.vis = 1;
			}
		}else if(event_action(event) == ACTION_MENU && event_is_down(event)){
			deposit_cursor();
			push_cursor(x,y);
			roi(1);
		}else if(event_action(event) == ACTION_SELECT && event_is_down(event)){
			if(x < 0 || x >= canvas_size_x) return(NOTIFY_DONE);
			if(y < 0 || y >= canvas_size_y) return(NOTIFY_DONE);
			if(note == 1){
				annotate_add(x,y,note_text);
				note = 0;
			}else{
				push_cursor(x,y);
				if(box.vis){
					draw_box(LINE_ERASE);
					box.vis = 0;
				}
				box.x = x;
				box.y = y;
				if(!box.vis){
					draw_box(LINE_SOFT);
					box.vis = 1;
				}
				roi(1);
				box.lx = x;
				box.ly = y;
			}
		}
	}
	return(NOTIFY_DONE);
}


void
display_status(void)
{
	fputs("Window memory: ",ttyout);
	switch(mitshm){
		case 0: fputs("normal (not shared)\n",ttyout); break;
		case 1: fputs("shared images",ttyout); break;
		case 2: fputs("shared pixmaps",ttyout); break;
	}

#ifdef MITSHM
	if(mitshm)fprintf(ttyout,", shared memory id: %d\n",shminfo.shmid);
#endif

	fputs("Canvas memory: ",ttyout);
	if(canvas_shm){
		fprintf(ttyout,"shared, id: %d\n",canvas_id);
	}else{
		fprintf(ttyout,"normal (not shared)\n");
	}

	fprintf(ttyout,"Depth: %d",dsp_depth);
	fprintf(ttyout," Bits/Pixel: %d",dsp_bits_pixel);
	fprintf(ttyout," Class: %s\n",classcolor(dsp_class));
	fprintf(ttyout,"Server: %s",byteorder(dsp_byte_order));
	fprintf(ttyout," Host: %s",byteorder(host_byte_order));
	fprintf(ttyout," Visual: 0x%lx\n",visid);
	fprintf(ttyout,"Masks: Red: %08X Green: %08X Blue: %08X\n",red_mask,green_mask,blue_mask);
}

#ifdef MITSHM
static void
canvas_shm_error(char *s)
{
	fprintf(stderr,"%s: %s\n",s,strerror(errno));
	fputs("some port operations will be degraded\n",stderr);
	canvas_shm = 0;
}
#endif


int 
display_canvas_id(void)
{
	return(canvas_id);
}

void
display_clear_annotate_clip(void)
{
	XSetClipMask(dpy,gc_l_xor,None);
	XSetClipMask(dpy,gc_a,None);

	XDestroyRegion(region);
}

void
display_set_annotate_clip(int left,int top,int width,int height)
{
	XRectangle rec;

	region = XCreateRegion();

	rec.x = left;
	rec.y = top;
	rec.width = width;
	rec.height = height;

	XUnionRectWithRegion(&rec,region,region);

	XSetRegion(dpy,gc_l_xor,region);
	XSetRegion(dpy,gc_a,region);
}

#ifdef TIME

static void
time_init(void)
{
	gettimeofday(&timep,NULL);
	etimeu = timep.tv_usec;
	etimes = timep.tv_sec;
	getrusage(RUSAGE_SELF,&usage);
	utimeu = usage.ru_utime.tv_usec;
	utimes = usage.ru_utime.tv_sec;
	stimeu = usage.ru_stime.tv_usec;
	stimes = usage.ru_stime.tv_sec;
}

static void
time_print(void)
{
	gettimeofday(&timep,NULL);
	etimeu = timep.tv_usec - etimeu;
	etimes = timep.tv_sec - etimes;
	if(etimeu < 0){
		etimeu += 1000000;
		etimes -= 1;
	}
	etimeu /= 1000;
	getrusage(RUSAGE_SELF,&usage);
	utimeu = usage.ru_utime.tv_usec - utimeu;
	utimes = usage.ru_utime.tv_sec - utimes;
	stimeu = usage.ru_stime.tv_usec - stimeu;
	stimes = usage.ru_stime.tv_sec - stimes;
	if(stimeu < 0){
		stimeu += 1000000;
		stimes -= 1;
	}
	if(utimeu < 0){
		utimeu += 1000000;
		utimes -= 1;
	}
	utimeu /= 1000;
	stimeu /= 1000;
	fprintf(ttyout,"%d.%03.du %d.%03.ds %d.%03.d\n",
		utimes,utimeu,stimes,stimeu,etimes,etimeu);
}
#endif

#ifdef MITSHM
static void
shm_error(char *s)
{
	fprintf(stderr,"%s: %s\n",s,strerror(errno));
	fputs("window/level operations will be degraded\n",stderr);
	mitshm = 0;
}
#endif

static void
control_done_proc(void)
{
	/*
	puts("in control done proc");
	*/
}


/*
static Notify_value
frame_destroy(client,status)
Notify_client client;
Destroy_status status;
{
	if(client == ControlFrame){
		puts("control frame ...");
	}else if(client == BaseFrame){
		puts("base frame ...");
	}else{
		puts("unknown frame ...");
	}

	if(status == DESTROY_CHECKING){
		puts("destroy_checking");
		notify_veto_destroy(client);
	}else if(status == DESTROY_CLEANUP){
		puts("destroy_cleanup");
		return(notify_next_destroy_func(client,status));
	}else if(status == DESTROY_SAVE_YOURSELF){
		puts("destroy_save_yourself");
		notify_next_destroy_func(client,status);
	}else{
		puts("destroy_process_death");
	}

	return NOTIFY_DONE;
}
*/

void
display_clean_up(void)
{

#ifdef MITSHM
	if(mitshm){
		XShmDetach(dpy,&shminfo);
		XDestroyImage(xi);
		shmdt(shminfo.shmaddr);
		shmctl(shminfo.shmid,IPC_RMID,0);
	}

	if(canvas_shm){
		shmdt((char *)image16);
		shmctl(canvas_id,IPC_RMID,0);
	}
#endif

#ifdef NEED_KILL
		// could be dangerous is process died and another process
		// has the same PID
	kill(tty_pid,SIGKILL);
#endif
}

void
display_destroy(void)
{
	xv_set(BaseFrame,FRAME_NO_CONFIRM,TRUE,0);
	xv_destroy_safe(BaseFrame);
}

static Notify_value
signal_catch(client,sig,when)
Notify_client client;
int sig;
Notify_signal_mode when;
{
	printf("\n\nSIGNAL\n\n");

	display_destroy();
	display_clean_up();
	exit(1);
	return(-1);	/* make compiler happier */
}

void
display_get_mouse(int *x,int *y)
{
	Rect *mp;

	mp = (Rect *) xv_get(canvas_paint,WIN_MOUSE_XY);

	*x = mp->r_left;
	*y = mp->r_top;
}

void
display_set_mouse(int x,int y)
{
	xv_set(canvas_paint,WIN_MOUSE_XY,x,y,NULL);
}

static void
set_sliders_messages(void)
{
	char s[10];

	if(opt_panel){
		sprintf(s,"W:%5d",window);
		xv_set(message_win,PANEL_LABEL_STRING,s,NULL);
		sprintf(s,"L:%5d ",level);
		xv_set(message_lev,PANEL_LABEL_STRING,s,NULL);
	}
}

void
display_set_sliders(void)
{
	int r,wmin,wmax,lmin,lmax;

	if(!opt_panel) return;

	set_sliders_messages();
	if(cf_mode){		/* coarse */
		lmin = -zero_offset;
		lmax = mi_max - zero_offset;
		wmin = 0;
		wmax = mi_max;
	}else{			/* fine */
		r = SLIDER_FINE_RANGE / 2;

		wmin = window - r;
		if(wmin < 0) wmin = 0;
		wmax = wmin + SLIDER_FINE_RANGE;
		if(wmax > mi_max){
			wmax = mi_max;
			wmin = wmax - SLIDER_FINE_RANGE;
		}

		lmin = level - r;
		if(lmin < -zero_offset) lmin = -zero_offset;
		lmax = lmin + SLIDER_FINE_RANGE;
		if(lmax > mi_max-zero_offset){
			lmax = mi_max - zero_offset;
			lmin = lmax - SLIDER_FINE_RANGE;
		}

	}

	xv_set(lev_slider,
		PANEL_VALUE,level,
		PANEL_MIN_VALUE, lmin,
		PANEL_MAX_VALUE, lmax,
		NULL);

	xv_set(win_slider,
		PANEL_VALUE,window,
		PANEL_MIN_VALUE, wmin,
		PANEL_MAX_VALUE, wmax,
		NULL);
#ifdef DEBUG_WL
	printf("w/l (min/max): %d %d %d %d\n",wmin,wmax,lmin,lmax);
#endif
}

static void
slider_action(item,value,event)
Panel_item item;
int value;
Event *event;
{

#ifdef DEBUG_EVENT
	printf("%d %d\n",event_id(event),event_action(event));
	printf("%d %d %d\n",event_x(event),event_y(event),event_is_up(event));
#endif

	if(item == win_slider){
		window = value;
	}else{
		level = value;
	}
	set_sliders_messages();

	if((event_action(event) == ACTION_SELECT && event_is_up(event)) || wltrack){
		window_level_lut(window,level);
		if(event_is_up(event)){
			load_pix(0);
		}else{
			load_pix(-1);
		}
	}
}

static void 
choice_proc(item,event)
Panel_item item;
Event *event;
{
	if(cf_mode != xv_get(cf_select,PANEL_VALUE)){
		cf_mode ^= 1;
		display_set_sliders();
	}
}


void
display_tty_out(char *line)
{
	char l;

	l = strlen(line);
	if(!l) return;
	ttysw_output(tty,line,l);

#ifdef __alpha
	if(line[l-1] == '\n')ttysw_output(tty,"\r",1);
#endif
}

void
display_panel_message(char *s1)
{
	xv_set(message_item,PANEL_LABEL_STRING,s1,NULL);
}

void
display_cross_off(void)
{
	xv_set(canvas_paint_window(DrawingCanvas),
		WIN_CURSOR, no_cursor,
		NULL);
}

void
display_cross_on(void)
{
	xv_set(canvas_paint_window(DrawingCanvas),
		WIN_CURSOR, cross_hair,
		NULL);
}

void
display_put_image(int rx,int ry,int rdx,int rdy)
{
	if(mitshm == 1){
#ifdef MITSHM
		XShmPutImage(dpy,pixmap,canvas_gc,xi,rx,ry,rx,ry,rdx,rdy,False);
#endif
	}else if(mitshm == 0){
		XPutImage(dpy,pixmap,canvas_gc,xi,rx,ry,rx,ry,rdx,rdy);
	}
}

void
display_copy_area(int rx,int ry,int rdx,int rdy)
{
	XCopyArea(dpy,pixmap,canvas_win,canvas_gc,rx,ry,rdx,rdy,rx,ry);
}

static void 
repaint_canvas(c,p,r)
Canvas c;
Xv_Window *p;
Rectlist *r;
{
	refresh(
		(int) r->rl_bound.r_left,
		(int) r->rl_bound.r_top,
		(int) r->rl_bound.r_width,
		(int) r->rl_bound.r_height);
}

void
display_drawline(int x1,int y1,int x2,int y2,int type)
{
	int x1p,y1p,dx,dy;

	if(annotate_loc){
		XDrawLine(dpy,pixmap,gc_L,x1,y1,x2,y2);
	}else{
		if(boxred == 0){
			if(type == LINE_HARD){
				XDrawLine(dpy,canvas_win,gc_l_src,x1,y1,x2,y2);
			}else if(type == LINE_ERASE || type == LINE_SOFT){
				XDrawLine(dpy,canvas_win,gc_l_xor,x1,y1,x2,y2);
			}else{
				parse_er("xpic: display_draw_line: internal error: bad type");
			}
		}else{
			if(type == LINE_HARD || type == LINE_SOFT){
				XDrawLine(dpy,canvas_win,gc_l_src,x1,y1,x2,y2);
			}else if(type == LINE_ERASE){
				if(y1 == y2){
					dx = abs(x1-x2) + 1;
					dy = 1;
					x1p = (x1 < x2)? x1 : x2;
					y1p = y1;
				}else{
					dx = 1;
					dy = abs(y1-y2) + 1;
					x1p = x1;
					y1p = (y1 < y2)? y1 : y2;
				}
				display_copy_area(x1p,y1p,dx,dy);
			}else{
				parse_er("xpic: display_draw_line: internal error: bad type");
			}
		}
	}

}


		/* no processing of backslash */
void
display_text_raw(int x,int y,char *s)
{
	if(annotate_loc){
		XDrawImageString(dpy,pixmap,gc_A,x,y,s,strlen(s));
	}else{
		XDrawString(dpy,canvas_win,gc_a,x,y,s,strlen(s));
	}
}

void
display_text(int x,int y,char *s)
{
	static char buf[KB_BUF];
	char *p = buf;

	while(*s){
		if(*s != '\\'){
			*p++ = *s++;
		}else switch(*++s){

		case 0:
			*p++ = '\\';
			break;
		case '_':
		case ' ':
			*p++ = ' ';
			s++;
			break;
		case 'n':
			*p = 0;
			display_text_raw(x,y,buf);
			y += font_height;
			p = buf;
			s++;
			break;
		default:
			*p++ = *s++;
		}
	}
	if(p != buf){
		*p = 0;
		display_text_raw(x,y,buf);
	}
}

void
display_get_image(void)
{
		/* always get to allow shared memory to catch up*/

	txi = XGetImage(dpy,pixmap,0,0,canvas_size_x,
		canvas_size_y,AllPlanes,ZPixmap);
	pr2_data = (DPIXEL8 *) txi->data;
	pr2_pixels = txi->bytes_per_line;

#ifdef DEBUG
	printf("display_get_image: width, height, bytes_per_line, %d %d %d\n",
		txi->width,
		txi->height,
		txi->bytes_per_line);
#endif
}

void
display_release_image(void)
{
	XDestroyImage(txi);
}

static Notify_value read_input(client,in_fd)
Notify_client client;         
int in_fd;          
{
	char buf[KB_BUF];
	int ptr;
	char *c;
	char pathname[MAXPATHLEN];

	ptr = read(in_fd,buf,KB_BUF);
	if(ptr == 0){
		printf("xpic's keyboard processor died.\n");
		xv_set(BaseFrame, FRAME_NO_CONFIRM, TRUE, 0);
                window_done(tty);
		exit(0);
	}

	buf[ptr] = 0;
	if((c = strchr(buf,'\n')) != NULL) *c = 0;
	if((c = strchr(buf,13)) != NULL) *c = 0;

#ifdef TIME
	time_init();
#endif
	if(!strcmp(buf,SPECIALMODE))
		cwd2kb=1;
	else if (*buf!='\0')
		command(buf,0);
#ifdef TIME
	time_print();
#endif

	buf[0] = EOC;
	buf[1] = cwd2kb;
	fflush(ttyout);
	if(-1 == write(tofd[1],buf,2)) perror("xpic: display.c: read_input: write error");

	if(cwd2kb){
		if(NULL == getcwd(pathname,MAXPATHLEN)) perror("xpic: display.c: read_input: getcwd error");
		if(-1 == write(tofd[1],pathname,MAXPATHLEN)) perror("xpic: display.c: read_input: write error");
	}
	cwd2kb = 0;		// reset send cwd to kb
	return(NOTIFY_DONE);
}

void
display_attach(int *gargc,char **gargv)
{
	xv_init(XV_INIT_ARGC_PTR_ARGV,gargc,gargv,NULL);
}

static void
display_canvas_init(void)
{
	canvas_shm = 0;

#ifdef MITSHM
	if(no_attach == 0){
		canvas_id = shmget(IPC_PRIVATE, 
			sizeof(CPIXEL) * canvas_size_x * canvas_size_y, 
			IPC_CREAT|0777);
		if(canvas_id == -1){
			canvas_shm_error("can't create canvas shared memory");
			if(errno == EINVAL){
				fprintf(stdout,
				 "canvas shared memory too large for operating system\n");
			}
			goto canvas_shm_out;
		}
		image16 = (CPIXEL *) shmat(canvas_id,0,0);
		if((int ) image16 == -1){
			canvas_shm_error("can't get shared canvas address");
			goto canvas_shm_out;
		}
		canvas_shm = 1;
canvas_shm_out:
	;
	}
#endif

	if(!canvas_shm){
		image16 = (CPIXEL *) calloc(canvas_size_x*canvas_size_y,sizeof(CPIXEL));
		if(image16 == NULL){
			fprintf(stderr,"xpic: can't malloc canvas memory\n");
			exit(1);
		}
		canvas_id = -1;
	}
}

static Xv_opaque
quit_notify_proc(Panel_item item, Event *event)
{
	int result=0;

	result = notice_prompt(panel,	NULL,
		NOTICE_FOCUS_XY,	event_x(event),	event_y(event),
		NOTICE_MESSAGE_STRINGS,	"Do you really want to quit xpic?",	NULL,
		NOTICE_BUTTON_YES,	"Yes",
		NOTICE_BUTTON_NO,	"No",
		NULL);
	if(result == NOTICE_YES)
	{
		quit_guts();
		exit(0);
	}
	return result;
}

void
write_env_panel(int env)
{
	char temp[4];
	sprintf(temp,"e%d",current_env + 1);
	xv_set(env_message_item,PANEL_LABEL_STRING,temp,NULL);
}

static void
btn_env_pls(Panel_item item, Event *event)
{

	env_inc(1);
	write_env_panel(current_env);
}

static void
btn_env_mns(Panel_item item, Event *event)
{
	env_dec(1);
	write_env_panel(current_env);
}

void
display_init(char *prog_name,int init,int pcolor,char *title)
{
	int w,i,n;
#ifdef MITSHM
	int try_mitshm;
#endif
	int ttyfd;
	char header[50];
	char *c;
	char *nc_argv[11];
	char inits[3];
	char pn0[4],pn1[4];
	char pp0[4],pp1[4];
	Icon icon;
	Icon control_icon;
	Server_image frame_image;
	Server_image control_image;
	Server_image cursor_image,no_cursor_image;
	Cms cms;
	Xv_singlecolor *colors;
	Xv_Font     _font;
	XFontStruct *font;
	XGCValues gcvalues;
	DPIXEL32 background,foreground;
	static Notify_client  my_client;
	static Notify_client  sig_me;
	static char outbuf[BUFSIZ];
	char server_name[100];
	struct hostent *hostp;
#ifdef MITSHM
	char host1[MAXHOSTNAMELEN],host2[MAXHOSTNAMELEN];
#endif
	int tlut;			/* total lut size */
	XPixmapFormatValues *lpf;
	DPIXEL32 red,white,black;
	XVisualInfo to_vinfo,*frm_vinfo;
	// editor tag: veep
/* 	struct termios termptr; */
/*
	check machine-dependent variables
*/
	if(sizeof(char) != 1) parse_pri("*xpic: sizeof(char): expected: 1 found:",sizeof(char));
	if(sizeof(short) != 2) parse_pri("*xpic: sizeof(short): expected: 2 found:",sizeof(short));
	if(sizeof(int) != 4) parse_pri("*xpic: sizeof(int): expected: 4 found:",sizeof(int));
/*
	set up base frame
*/
	WTEXT("debug 1");
	BaseFrame = xv_create(XV_NULL,FRAME,
		WIN_X, xp,
		WIN_Y, yp,
		FRAME_INHERIT_COLORS,TRUE,
		FRAME_SHOW_RESIZE_CORNER, FALSE,
		FRAME_X_ACCELERATOR,"o+Ctrl" ,btn_open_mi,0,
		//  FRAME_NO_CONFIRM, FALSE,
		NULL);

	dpy = (Display *)xv_get(BaseFrame, XV_DISPLAY);

	vis = (Visual *)xv_get(BaseFrame, XV_VISUAL);

	visid = XVisualIDFromVisual(vis);

	to_vinfo.visualid = visid;
	frm_vinfo = XGetVisualInfo(dpy,VisualIDMask,&to_vinfo,&n);

	if(frm_vinfo == NULL) parse_er("xpic: get get default visual - XGetVisualInfo");

	red_mask = frm_vinfo[0].red_mask;
	green_mask = frm_vinfo[0].green_mask;
	blue_mask = frm_vinfo[0].blue_mask;

	dsp_byte_order = XImageByteOrder(dpy);
	if(dsp_byte_order != LSBFirst && dsp_byte_order != MSBFirst){
		fprintf(stderr,"xpic: unknown byte order; found %d order\n",
			dsp_byte_order);
		exit(1);
	}

	host_byte_order = hostbyteorder();

	dsp_depth = xv_get(BaseFrame, XV_DEPTH);

	if(dsp_depth == 24){
		dsp_bits_pixel = 32;
		dsp_bytes_pixel = 4;
	}else if(dsp_depth == 32){
		dsp_bits_pixel = 32;
		dsp_bytes_pixel = 4;
	}else if(dsp_depth == 8){
		dsp_bits_pixel = 8;
		dsp_bytes_pixel = 1;
	}else{
		fprintf(stderr,"xpic: requires 8, 24 or32 bit buffer; found %d bits\n",
			dsp_depth);
		exit(1);
	}

	dsp_class = xv_get(BaseFrame, XV_VISUAL_CLASS);


	lpf = XListPixmapFormats(dpy,&n);
	if(lpf == NULL){
		fprintf(stderr,"xpic: get can't pixmap formats\n");
		exit(1);
	}
	for(i=0;i<n;i++){
		if(lpf[i].depth == dsp_depth){
			if(lpf[i].bits_per_pixel != dsp_bits_pixel){
				fprintf(stderr,"xpic: bit/pixel error. expected %d. found: %d\n",
					dsp_bits_pixel,lpf[i].bits_per_pixel);
				exit(1);
			}else{
				i = -1;
				break;
			}
		}
	}
	if(i != -1){
		parse_er("xpic: could not find pixmap matching depth");
	}

#ifdef DEBUG
	{
	int n,i;
	XVisualInfo t,*l;
	int id,c;
	c = xv_get(BaseFrame, XV_VISUAL_CLASS);
	printf("used class %d\n",c);
	id = xv_get(BaseFrame, XV_XID);
	printf("used id %d\n",id);
	t.depth = dsp_depth;
	printf("dpy %d vis %d\n",(int)dpy,(int)vis);
	l = XGetVisualInfo(dpy,VisualDepthMask,&t,&n);
	if(l == NULL)
		printf("no match\n");
	else{
		for(i=0;i<n;i++){
			printf("address: %x id: %x, depth %d, class %s",
				(int)l[i].visual,
				(int)l[i].visualid,
				l[i].depth,
				classcolor(l[i].class));
		}
	}

	}
#endif

	WTEXT("debug 2");
	frame_image  = (Server_image) xv_create((int)NULL, 
		SERVER_IMAGE,
		XV_WIDTH, 64,
		XV_HEIGHT, 64,
		SERVER_IMAGE_BITS, icon_bits,
		NULL);

	WTEXT("debug 3");
	icon = (Icon) xv_create(XV_NULL,ICON,
		ICON_IMAGE, frame_image,
		ICON_TRANSPARENT, TRUE,
		NULL);

	xv_set(BaseFrame, FRAME_ICON,icon,NULL);
/*
	setup colormap
*/
	tlut = lut_size + 3; 	/* 3 for red/blue/green highlights */
	tlut += OUTLINE_LENGTH;
	lut_red = lut_size;
	lut_blue = lut_size + 1;
	lut_green = lut_size + 2;
	lut_outline = lut_size +3;

	lut_scale = 256 / lut_size;

	colors =(Xv_singlecolor *) parse_malloc(sizeof(Xv_singlecolor)*(tlut));

	for(i=0;i<lut_size;i++){
		if(pcolor == 0){
			{
			int v;
				v =  i * lut_scale;
				colors[i].blue = v;
				colors[i].green = v;
				colors[i].red = v;
			}
		}else{
			int c1,c2,c3;

			c1 = color((i >> 0) & 3);
			c2 = color((i >> 2) & 3); 
			c3 = color((i >> 4) & 3);

			switch(i%3){
			case 0:
				colors[i].blue = c1;
				colors[i].green = c2;
				colors[i].red = c3;
				break;
			case 1:
				colors[i].blue = c2;
				colors[i].green = c3;
				colors[i].red = c1;
				break;
			case 2:
				colors[i].blue = c3;
				colors[i].green = c1;
				colors[i].red = c2;
				break;
			}
		}
	}
	colors[lut_red].blue = 0;
	colors[lut_red].green = 0;
	colors[lut_red].red = 255;

	colors[lut_blue].blue = 255;
	colors[lut_blue].green = 0;
	colors[lut_blue].red = 0;

	colors[lut_green].blue = 0;
	colors[lut_green].green = 255;
	colors[lut_green].red = 0;

	// editor tag - peep

	{
	int oml[OUTLINE_LENGTH*3] = { // Colors for outline mode
		255,0,0,		// Provided by Matt Hirsch
		0,255,0,		// Red/Green/Blue 3-tuples
		0,0,255,
		255,255,0,
		255,0,255,
		0,255,255,
		255,128,128,
		128,255,128,
		128,128,255,
		0,64,128,
		128,0,64,
		64,128,0,
		64,0,0,
		0,64,0,
		0,0,64
	};

	for(i=0;i<OUTLINE_LENGTH;i++){
		colors[lut_outline+i].red = oml[3*i];
		colors[lut_outline+i].green = oml[3*i+1];
		colors[lut_outline+i].blue = oml[3*i+2];
	}

	}


	if(dsp_depth == 8){
		cms = (Cms) xv_create((int)NULL,CMS,
			XV_VISUAL, vis,
			CMS_SIZE, tlut,
			CMS_COLORS, colors,
			NULL);
		if(cms == XV_NULL) parse_er("xpic: can't create real cms");
		cms_map = (DPIXEL32 *) xv_get(cms, CMS_INDEX_TABLE);
		if((int)cms_map == 0)parse_er("xpic: cms cms_map error");

#ifdef DEBUG
	{
		Xv_singlecolor *colorsp =(Xv_singlecolor *)
			parse_malloc(sizeof(Xv_singlecolor)*(tlut));

		xv_get(cms,CMS_COLORS, colorsp);
		for(i=0;i<tlut;i++){
			printf("cms (%2d): %6X: %3d %3d %3d %3d %3d %3d %d\n",
				i,
				cms_map[i],
				colorsp[i].blue,
				colorsp[i].green,
				colorsp[i].red,
				colors[i].blue,
				colors[i].green,
				colors[i].red,
				colorsp[i].blue != colorsp[i].blue
				|| colorsp[i].red != colorsp[i].red
				|| colorsp[i].green != colorsp[i].green);
		}
	}
#endif

	}else{
		cms_map = (DPIXEL32 *) malloc(sizeof(DPIXEL32)*tlut);
		if(cms_map == NULL) parse_er("xpic: can't allocate cms_map memory");

		for(i=0;i<tlut;i++){
			cms_map[i] =
				(((colors[i].red << 16) | (colors[i].red << 8) | (colors[i].red)) & red_mask) |
				(((colors[i].blue << 16) | (colors[i].blue << 8) | (colors[i].blue)) & blue_mask) |
				(((colors[i].green << 16) | (colors[i].green << 8) | (colors[i].green)) & green_mask);
			if(dsp_byte_order != host_byte_order) cms_map[i] = pixel32_swap(cms_map[i]);
#ifdef DEBUG
			printf("cms_map[%d]: %8X\n",i,cms_map[i]);
#endif
		}

		cms = (Cms) NULL;		// make compiler happier
	}
/*
	set up canvas
*/
	WTEXT("debug 5");
	DrawingCanvas = xv_create(BaseFrame, CANVAS,
		CANVAS_AUTO_SHRINK , FALSE,
		CANVAS_AUTO_EXPAND , FALSE,
		CANVAS_WIDTH,canvas_size_x,
		CANVAS_HEIGHT,canvas_size_y,
		XV_WIDTH,canvas_size_x,
		XV_HEIGHT,canvas_size_y,
		CANVAS_RETAINED, FALSE,
		CANVAS_REPAINT_PROC, repaint_canvas,
		OPENWIN_NO_MARGIN, TRUE,
		OPENWIN_SHOW_BORDERS,FALSE,
		NULL);


	if(dsp_depth == 8){
		xv_set(
			DrawingCanvas,
			WIN_CMS, cms,
			WIN_BACKGROUND_COLOR, xv_get(cms,CMS_BACKGROUND_PIXEL),
			WIN_FOREGROUND_COLOR, xv_get(cms,CMS_FOREGROUND_PIXEL),
			NULL);
	}

	canvas_paint = canvas_paint_window(DrawingCanvas);
	canvas_win = (Window)xv_get(canvas_paint, XV_XID);

	xv_set(canvas_paint_window(DrawingCanvas),
		WIN_EVENT_PROC, track_cursor,
		WIN_CONSUME_EVENTS,
			WIN_NO_EVENTS,
			WIN_MOUSE_BUTTONS,
			LOC_MOVE,
			LOC_WINENTER,
			LOC_WINEXIT,
			LOC_DRAG,
			WIN_REPAINT,
			NULL,
		NULL);
/*
	set up graphics contexts
*/
	{
		int fs = 14;

		while(1){
			_font = (Xv_Font)xv_find(BaseFrame, FONT, 
				FONT_FAMILY,FONT_FAMILY_DEFAULT_FIXEDWIDTH,
				FONT_STYLE, FONT_STYLE_BOLD,
				FONT_SIZE, fs,
				NULL);

			font_width = xv_get(_font,FONT_DEFAULT_CHAR_WIDTH);
#ifdef DEBUG
			printf("fs: %d, gc width: %d\n",fs,font_width);
#endif
			if(font_width <= 9) break;
			fs--;
		}
	}


	font_height = xv_get(_font,FONT_DEFAULT_CHAR_HEIGHT);
	font = (XFontStruct *)xv_get(_font, FONT_INFO);

	if(dsp_depth == 8){
		background = xv_get(BaseFrame,WIN_FOREGROUND_COLOR);
		foreground = xv_get(BaseFrame,WIN_BACKGROUND_COLOR);

		black = background;
		white = foreground;
	}else{
		white = 0xffffff;
		black = 0;
		if(dsp_byte_order != host_byte_order) white = pixel32_swap(white);
		if(dsp_byte_order != host_byte_order) black = pixel32_swap(black);
	}
	red = cms_map[lut_size];

	gcvalues.foreground = white;
	gcvalues.background = black;
	gcvalues.function = GXcopy;
	canvas_gc = XCreateGC(dpy,canvas_win,
		GCFunction |  GCForeground | GCBackground, &gcvalues);

	gcvalues.foreground = red;
	gcvalues.function = GXcopy;
	gc_l_src = XCreateGC(dpy,canvas_win,
		GCFunction |  GCForeground, &gcvalues);

	if(dsp_depth == 8){
		gcvalues.foreground = cms_map[0] ^ cms_map[lut_size-1];
		gcvalues.background = cms_map[0];
	}else{
		gcvalues.background = black;
		gcvalues.foreground = white;
	}
	gcvalues.function = GXxor;
	gc_l_xor = XCreateGC(dpy,canvas_win,
		GCFunction |  GCForeground | GCBackground, &gcvalues);

	gcvalues.font = font->fid;
	gcvalues.foreground = red;
	gcvalues.function = GXcopy;
	gc_a = XCreateGC(dpy,canvas_win,
		GCFunction | GCFont | GCForeground, &gcvalues);

	white = 255;
	black = 0;

	gcvalues.foreground = white;
	gcvalues.background = black;
	gcvalues.function = GXxor;
	gc_L = XCreateGC(dpy,canvas_win,
		GCFunction |  GCForeground | GCBackground, &gcvalues);

	gcvalues.font = font->fid;
	gcvalues.foreground = white;
	gcvalues.background = black;
	gcvalues.function = GXcopy;

	gc_A = XCreateGC(dpy,canvas_win,
		GCFunction | GCFont | GCForeground | GCBackground, &gcvalues);
/*
	set up cursors
*/
	WTEXT("debug 6");
	cursor_image  = (Server_image) xv_create(XV_NULL, SERVER_IMAGE,
		XV_WIDTH, 16,
		XV_HEIGHT, 16,
		SERVER_IMAGE_BITS, cursor_bits,
		NULL);


	{

	Xv_singlecolor bg,fg;

	fg.red = 255;
	fg.blue = fg.green = 0;
	bg.red = bg.blue = bg.green = 0;

	WTEXT("debug 7");
	cross_hair = (Xv_Cursor) xv_create(XV_NULL,CURSOR,
		CURSOR_IMAGE, cursor_image,
		CURSOR_XHOT,8,
		CURSOR_YHOT,7,
		CURSOR_FOREGROUND_COLOR, &fg,
		NULL);
	}

	WTEXT("debug 8");
	no_cursor_image  = (Server_image) xv_create(XV_NULL, SERVER_IMAGE,
		XV_WIDTH, 16,
		XV_HEIGHT, 16,
		SERVER_IMAGE_BITS, no_cursor_bits,
		NULL);

	WTEXT("debug 9");
	no_cursor = (Xv_Cursor) xv_create(XV_NULL,CURSOR,
		CURSOR_IMAGE, no_cursor_image,
		CURSOR_XHOT,8,
		CURSOR_YHOT,7,
		NULL);

	display_cross_on();
/*
	set up frame for controls (panel and tty)
*/
	if(sep_ctr){
		ControlFrame = xv_create(BaseFrame,FRAME,
			FRAME_LABEL, "xpic controls",
			FRAME_DONE_PROC, control_done_proc,
			NULL);

		control_image  = (Server_image) xv_create((int)NULL, 
			SERVER_IMAGE,
			XV_WIDTH, 64,
			XV_HEIGHT, 64,
			SERVER_IMAGE_BITS, icon_control_bits,
			NULL);

		control_icon = (Icon) xv_create(XV_NULL,ICON,
			ICON_IMAGE, control_image,
			ICON_TRANSPARENT, TRUE,
			NULL);

		xv_set(ControlFrame, FRAME_ICON,control_icon,NULL);
	}else{
		ControlFrame = BaseFrame;
	}
/*
	set up options panel
*/
	if(opt_panel){
		int pwidth;
		int fs = 10;

		while(1){
			font_panel = (Xv_Font)xv_find(BaseFrame, FONT, 
				FONT_SIZE, fs,
				NULL);

			w = xv_get(font_panel,FONT_DEFAULT_CHAR_WIDTH);
#ifdef DEBUG
			printf("fs: %d opt width: %d\n",fs,w);
#endif
			if(w <= 6) break;
			fs--;
		}
		w *= 7;

		panel = xv_create(ControlFrame,PANEL,
			XV_X, 0,
			WIN_ROW_GAP, 0,
			XV_BOTTOM_MARGIN, 0,
			PANEL_ITEM_Y_GAP, 8,
			PANEL_ITEM_X_GAP, 8,
			WIN_FONT, font_panel,
			NULL);

		if(!sep_ctr){
			pwidth = canvas_size_x;
			xv_set(panel,
				WIN_BELOW, DrawingCanvas,
				XV_WIDTH, pwidth,
				NULL
			);
		}else{
			pwidth = 512;
			xv_set(panel,
				XV_WIDTH, pwidth,
				NULL
			);
		}

		message_win = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_X, xv_col(panel,0),
			PANEL_LABEL_Y, xv_rows(panel,0),
			PANEL_LABEL_STRING, "W:      ",
			PANEL_LABEL_WIDTH, w,
			NULL);

		win_slider = xv_create(panel,PANEL_SLIDER,
			WIN_RIGHT_OF, message_win,
			PANEL_LABEL_Y, xv_rows(panel,0),
			PANEL_VALUE, window,
			PANEL_MIN_VALUE, 0,
			PANEL_MAX_VALUE, 2047,
			PANEL_SHOW_RANGE, FALSE,
			PANEL_SHOW_VALUE, FALSE,
			PANEL_SLIDER_WIDTH, pwidth - SLIDER_SLOP,
			PANEL_NOTIFY_LEVEL, PANEL_ALL,
			PANEL_NOTIFY_PROC, slider_action, 
			NULL);

		message_lev = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_X, xv_col(panel,0),
			PANEL_LABEL_Y, xv_rows(panel,1),
			PANEL_LABEL_STRING, "L:      ",
			PANEL_LABEL_WIDTH, w,
			NULL);

		lev_slider = xv_create(panel,PANEL_SLIDER,
			WIN_RIGHT_OF, message_lev,
			PANEL_LABEL_Y, xv_rows(panel,1),
			PANEL_VALUE, level,
			PANEL_MIN_VALUE, -zero_offset,
			PANEL_MAX_VALUE, 2047-zero_offset,
			PANEL_SHOW_RANGE, FALSE,
			PANEL_SLIDER_WIDTH, pwidth - SLIDER_SLOP,
			PANEL_SHOW_VALUE, FALSE,
			PANEL_NOTIFY_LEVEL, PANEL_ALL,
			PANEL_NOTIFY_PROC, slider_action, 
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_X, xv_col(panel,0),
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "-",
			PANEL_NOTIFY_PROC, btn_previous, 
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "+",
			PANEL_NOTIFY_PROC, btn_next, 
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "u",
			PANEL_NOTIFY_PROC, btn_scroll_up, 
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "d",
			PANEL_NOTIFY_PROC, btn_scroll_down, 
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "l",
			PANEL_NOTIFY_PROC, btn_scroll_left,
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "r",
			PANEL_NOTIFY_PROC, btn_scroll_right,
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_Y, xv_rows(panel,2),
			PANEL_LABEL_STRING, "m",
			PANEL_NOTIFY_PROC, btn_measure,
			NULL);

		cf_select = xv_create(panel, PANEL_CHOICE,
			PANEL_LABEL_Y, xv_rows(panel,2)-2,
			PANEL_CHOICE_STRINGS, "F","C",NULL,
			PANEL_VALUE, 0,
			PANEL_NOTIFY_PROC, choice_proc,
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_X, xv_col(panel, (canvas_size_x<304?34:40)),
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x<304?3:2))+(canvas_size_x<304?4:0),
			PANEL_LABEL_STRING, "Browse",
			PANEL_NOTIFY_PROC, btn_open_mi,
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_X, xv_col(panel,(canvas_size_x>393?50:0)),
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x>393?2:3)),
			PANEL_LABEL_STRING, "e-",
			PANEL_NOTIFY_PROC, btn_env_mns,
			NULL);


		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_X, xv_col(panel,(canvas_size_x>393?55:5)),
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x>393?2:3)),
			PANEL_LABEL_STRING, "e+",
			PANEL_NOTIFY_PROC, btn_env_pls,
			NULL);

		env_message_item = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_X, xv_col(panel,(canvas_size_x>393?60:10))+2,
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x>393?2:3))+3,
			PANEL_LABEL_BOLD, FALSE,
			PANEL_LABEL_STRING, "e1 ",
			XV_MARGIN, 0,
			NULL);

		xv_create(panel, PANEL_BUTTON,
			PANEL_LABEL_X, xv_col(panel,(canvas_size_x>426?64:(canvas_size_x>393?0:14))),
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x>426?2:3)),
			PANEL_LABEL_STRING, "Quit",
			PANEL_NOTIFY_PROC, quit_notify_proc,
			NULL);

		message_item = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_X, xv_col(panel,(canvas_size_x>511?70:(canvas_size_x>426?1:(canvas_size_x>393?10:20))))+(canvas_size_x<304?0:6),
			PANEL_LABEL_Y, xv_rows(panel,(canvas_size_x>511?2:3))+5,
			PANEL_LABEL_BOLD, FALSE,
			PANEL_LABEL_STRING, "                         ",
			XV_MARGIN, 0,
			NULL);

		if(sep_ctr){
			window_fit_height(panel);
		}else{
			window_fit_height(panel);
		}
	}

/*
	set up tty/term window
*/
	if(-1 == pipe(frfd)) perror("xpic: display_init: from pipe create error");
	sprintf(pn0,"%d",frfd[0]);
	sprintf(pn1,"%d",frfd[1]);

	if(-1 == pipe(tofd)) perror("xpic: display_init: to pipe create error");
	sprintf(pp0,"%d",tofd[0]);
	sprintf(pp1,"%d",tofd[1]);
	sprintf(inits,"%d",init);

	nc_argv[0] = prog_name;
	nc_argv[1] = MAIN_KB;
	nc_argv[2] = pn0;
	nc_argv[3] = pn1;
	nc_argv[4] = pp0;
	nc_argv[5] = pp1;
	nc_argv[6] = inits;
	nc_argv[7] = 0;

	WTEXT("debug 10");
	tty = xv_create(ControlFrame,
#ifndef __alpha
		TERMSW,
#else
 		TTY,
#endif
		WIN_IS_CLIENT_PANE,
		WIN_ROWS,rows,
		TTY_ARGV, nc_argv,
		TTY_QUIT_ON_CHILD_DEATH, TRUE,
			WIN_FONT, font_panel,
		NULL);

	if(!sep_ctr){
		xv_set(tty,
			WIN_BELOW, (opt_panel)? panel : DrawingCanvas,
			WIN_X, 0,
			XV_WIDTH, canvas_size_x,
			NULL
		);
	}else{
		xv_set(tty,
			XV_WIDTH, 512,
			NULL
		);

		if(opt_panel){
			xv_set(tty,
				WIN_BELOW, (opt_panel)? panel : DrawingCanvas,
				WIN_X, 0,
				NULL
			);
		}
	}

	close(frfd[1]);
	close(tofd[0]);
	ttyfd = (int)xv_get( tty, TTY_TTY_FD );
	tty_pid = (int)xv_get( tty, TTY_PID);

/* 	tcgetattr(ttyfd,&termptr); */
/* 	termptr.c_lflag &= ~ECHO|ICANON; */
/* 	tcsetattr(ttyfd,TCSANOW,&termptr); */

	ttyin = fdopen( ttyfd, "r+" );
	ttyout = fdopen( ttyfd, "w+" );
	setbuf(ttyout,outbuf);

	window_fit(BaseFrame);


	if(sep_ctr){
		window_fit(ControlFrame);
	}
/*
	set gamma of display
*/
	strcpy(server_name,(char *)xv_get(xv_default_server,XV_NAME));
	if(NULL != (c = strchr(server_name,':'))) *c = 0;
	if(!*server_name) strcpy(server_name,host_name);

	hostp = gethostbyname(server_name);
	display_gethost_error(hostp,server_name);

	gamma_match(
		strcpy(
			parse_malloc_char(strlen(hostp->h_name)+1),
			hostp->h_name
		)
	);
/*
	set up for shared memory
*/
#ifdef MITSHM
	if(*server_name){
		hostp = gethostbyname(server_name);
		display_gethost_error(hostp,server_name);
		strcpy(host1,hostp->h_name);
		hostp = gethostbyname(host_name);
		display_gethost_error(hostp,host_name);
		strcpy(host2,hostp->h_name);
		try_mitshm = !strcmp(host1,host2);
	}else{		/* if server_name empty, assume client=server */
		try_mitshm = 1;
	}
	if(mitshm_override == 1) try_mitshm = 0;

	if(try_mitshm){
		mitshm = XShmQueryVersion(dpy,&rmajor,&rminor,&shm_pixmap);
	}else{
		mitshm = 0;
	}

#ifdef DEBUG
	printf("mitshm info: %d %d %d %d %d\n",
		try_mitshm,mitshm,rmajor,rminor,shm_pixmap);
#endif

	if(mitshm){
		mitshm += shm_pixmap;
		xi = XShmCreateImage(dpy,vis,dsp_bits_pixel,ZPixmap,NULL,
			&shminfo,canvas_size_x,canvas_size_y);
		if(xi->bytes_per_line != canvas_size_x * dsp_bytes_pixel || 
	  	  xi->height != canvas_size_y){
			mitshm = 0;
		}else{
			shminfo.shmid = shmget(IPC_PRIVATE, 
				dsp_bytes_pixel * canvas_size_x * canvas_size_y, 
				IPC_CREAT|0777);
			if(shminfo.shmid == -1){
				shm_error("can't create image shared memory");
				goto mitshm_exit;
			}
     			shminfo.shmaddr = xi->data = (char *)
				shmat(shminfo.shmid,0,0);
			pr = (DPIXEL8 *) xi->data;
			if((int ) shminfo.shmaddr == -1){
				shm_error("can't get shared memory address");
				goto mitshm_exit;
			}
			shminfo.readOnly = False;
     			if(False == XShmAttach(dpy,&shminfo)){
				shm_error("can't attach shared memory");
				goto mitshm_exit;
			}
			if(mitshm == 2){
				pixmap = XShmCreatePixmap(
					dpy,
					canvas_win,
					(char *) pr,
					&shminfo,
					canvas_size_x,
					canvas_size_y,
					dsp_depth);
			}
		}
		wltrack = 1;
	}
mitshm_exit:
#else
	mitshm = 0;
#endif
/*
	trap signals to [primarily] release shared memory
*/
	sig_me = (Notify_client)MAIN_NOTIFY_CLIENT;

	notify_set_signal_func(sig_me,signal_catch,SIGQUIT,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGBUS,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGHUP,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGINT,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGTERM,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGUSR1,NOTIFY_ASYNC);
	notify_set_signal_func(sig_me,signal_catch,SIGUSR2,NOTIFY_ASYNC);
/*
	initialize other stuff
*/
	if(title == NULL){
		strcpy(header,"xpic");
		xv_set(BaseFrame, XV_LABEL,header,NULL);
	}else{
		xv_set(BaseFrame, XV_LABEL,title,NULL);
	}

	if(!mitshm){
		pr = (DPIXEL8 *) calloc(canvas_size_x*canvas_size_y ,
			dsp_bytes_pixel);
		xi = XCreateImage(dpy,vis,dsp_depth,ZPixmap,0,(char *)pr,
			canvas_size_x,canvas_size_y,dsp_bits_pixel,0);
	}
	if(mitshm != 2){
		pixmap = XCreatePixmap(dpy,canvas_win,canvas_size_x,
			canvas_size_y,dsp_depth);
	}

	if(pr == NULL){
		fprintf(stderr,"xpic: can't grab enough memory for backing image\n");
		exit(1);
	}
/*
	start up other items
*/
	display_canvas_init();
/*
	start up connection
*/
	my_client =  (Notify_client) &tty;
	notify_set_input_func(my_client, read_input, frfd[0]);
/*
	make popups visible
*/
	if(sep_ctr){
		xv_set(ControlFrame,WIN_SHOW,TRUE,NULL);
	}
}

void
display_run(void)
{
	xv_main_loop(BaseFrame);
}

void
display_wait3(int me,int pid)
{
	notify_set_wait3_func(me,notify_default_wait3,pid);
}

int
display_tty_width(void)
{
	return((int) xv_get(tty,WIN_COLUMNS));
}

