/*
gzmi - handle gzmi files 
*/

#ident "$Id: gzmi.c,v 1.3 2007/05/22 13:22:45 ccrawford Exp $"

#include <xpic.h>

/*
	static buffers
*/
static PARSE_MIHEAD *gzmihead = NULL;

static
struct {
	PARSE_MIHEAD *gzmihead;
} envs[ENV_N];

static int 
gzmi_fread(void *ptr,int size,int n,gzFile *unit,int err)
{
  const char *zlib_error_message;
  int zlib_error_number;

  zlib_error_number = (int) gzread(unit,ptr,size*n);
  if(n*size != zlib_error_number){
    zlib_error_message = gzerror(unit,&zlib_error_number);
    if(zlib_error_number == Z_ERRNO) {
      zlib_error_message = strerror(err);
    }
    fprintf(stderr,"gzmi_fread: %s\n",zlib_error_message);
    return(-1);
  }
  return(0);
}

static int 
gzmi_fseek(gzFile *unit,long byte,char *type,int err)
{
  int m;
  const char *zlib_error_message;
  int zlib_error_number;
  
  if(!strcmp(type,"b")){
    m = SEEK_SET;
  }else if(!strcmp(type,"c")){
    m = SEEK_CUR;
  }else{
    fprintf(stderr,"gzmi_fseek: illegal mode: %s",
	    type);
    exit(1);
  }
  
  if(-1 == gzseek(unit,byte,m)){
    zlib_error_message = gzerror(unit,&zlib_error_number);
    if(zlib_error_number == Z_ERRNO) {
      zlib_error_message = strerror(err);
    }
    fprintf(stderr,"gzmi_fseek: %s\n",zlib_error_message);
    return(-1);
  }
  return(0);
}

#define checklimit (-1)

static int 
gzmi_rblock(gzFile *unit,int start,void *data,int count,int err)
{
  if(-1 == gzmi_fseek(unit,(long) (start * 512L),"b",err)){
    return(-1);
  }
  if(-1 == gzmi_fread(data,1,count*512,unit,err)){
    return(-1);
  }
  return(0);
}

PARSE_MIHEAD *parse_ogzmi_fe(gzFile *iu,int writeonly,FILE *ou,int err) {
  PARSE_MIHEAD *mihead_ogzmi=NULL;	/* return header */
  int i;				/* looper */
  int n;				/* number of images in file */
  unsigned short head[1024];	/* mi-file header */
  unsigned short ihead[1024];	/* mi-image header */
  int contig=-1;			/* 1=images contiguous */
  int same=-1;			/* 1=images same size */
  int order=-1;			/* 1=images in order in file */
  int ncol=-1;			/* # col in specific image */
  int nrow=-1;			/* # rows in specific image */
  int min_ncol=-1,max_ncol=-1;	/* min/max ncol for all images */
  int min_nrow=-1,max_nrow=-1;	/* min/max ncol for all images */
  int nblk;			/* nblks for specific file */
  int nblkp=-1;			/* nblks for previous file */
  char lbl[PARSE_MI_LEN_LABEL+1];	/* temp array for manipulating labels */
  int tblk=-1;			/* total blocks in file */
  int first=-1;			/* first image block */
  int iblk;			/* block pointer */
  int type=-1;			/* type of mi-file - see parse.h */
  //struct stat sbuf;		/* for stat'ing file */
  int needcheck;			// 1=need to check all images
  
  /* open file and get initial header */
  if(-1 == gzmi_rblock(iu,0,(char *)head,1,err)){
    fputs("Can't read gzmi-file header\n",ou);
    return(NULL);
  }
  parse_from_can_shortv((short *)head,256);
  
  
  /* get rest of header and get header information */
  if(head[PARSE_MI_LOC_MAGIC] == PARSE_MI_MAGIC_65K){
    type = PARSE_MI_65K;
    n = head[PARSE_MI_LOC_N];
    first = head[PARSE_MI_LOC_FIRST];
  }else{
    if(head[255]){
      if(-1 == gzmi_rblock(iu,1,head+256,3,err)){
	fputs("Can't read gzmi-file header\n",ou);
	return(NULL);
      }
      parse_from_can_shortv((short *)head+256,768);
      if(head[255] == 1) head[255] = 0;
      type = PARSE_MI_1024;
      first = 4;
    }else{
      type = PARSE_MI_256;
      for(i=256;i<1024;i++) head[i] = 0;
      first = 1;
    }
    
    for(i=0;i<1024;i++){
      if(head[i] == 0) break;
    }
    n = i;
  }
  
  /* start filling up header */
  mihead_ogzmi = (PARSE_MIHEAD *) parse_malloc(sizeof(PARSE_MIHEAD));
  mihead_ogzmi->first = first;
  mihead_ogzmi->type = type;
  mihead_ogzmi->n = n;
  
  /* make sure file is block aligned */
  /* this doesn't work for gzipped files, for obvious reasons, so for now just hope for the best */

/*   if(-1 == fstat(fileno(iu),&sbuf)){ */
/*     fputs("parse_ogzmi_fe: can't stat gzmi-file\n",ou); */
/*     if(err){  */
/*       exit(1); */
/*     }else{ */
/*       free(mihead_ogzmi); */
/*       return(NULL); */
/*     } */
/*   } */
/*   tblk = sbuf.st_size / 512; */
/*   if(tblk * 512 != sbuf.st_size){ */
/*     if(type == PARSE_MI_65K){ */
/*       fputs("parse_omi: mi-file is not block aligned\n",ou); */
/*       if(err){  */
/* 	exit(1); */
/*       }else{ */
/* 	free(mihead_ogzmi); */
/* 	return(NULL); */
/*       } */
/*     } */
/*   } */
  
/*   mihead_ogzmi->tblk = tblk; */
  mihead_ogzmi->tblk = tblk = 0;
  /* allocate arrays if necessary */
  if(n == 0 || writeonly == 1){
    mihead_ogzmi->ncol = NULL;
    mihead_ogzmi->nrow = NULL;
    mihead_ogzmi->bptr = NULL;
    mihead_ogzmi->nblk = NULL;
    mihead_ogzmi->label = NULL;
    return(mihead_ogzmi);
  }else{
    mihead_ogzmi->ncol = (int *) parse_malloc(sizeof(int) * n);
    mihead_ogzmi->nrow = (int *) parse_malloc(sizeof(int) * n);
    mihead_ogzmi->bptr = (long unsigned int *) parse_malloc(sizeof(long) * n);
    mihead_ogzmi->nblk = (long unsigned int *) parse_malloc(sizeof(long) * n);
    mihead_ogzmi->label = (char **) parse_malloc(sizeof(char *) * n);
    for(i=0;i<n;i++){
      mihead_ogzmi->label[i] = NULL; 
    }
  }
  
  // editor tag: peep
  if(type == PARSE_MI_65K){
    iblk = first;
    contig = 1;
    order = 1;
    
    needcheck = 1;
    if(checklimit > 0 && n > checklimit){
      if(-1 == gzmi_rblock(iu,iblk,ihead,1,err)){
	fputs("Can't read gzmi-file image headers\n",ou);
	parse_frmi(mihead_ogzmi);
	return(NULL);
      }
      ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
      nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
      nblk = PARSE_BLOCK_SHORT(ncol*nrow);
      if(n*(nblk+1)+1 == tblk){
	needcheck = 0;
	max_ncol = min_ncol = ncol;
	max_nrow = min_nrow = nrow;
	for(i=0;i<n;i++){
	  mihead_ogzmi->nrow[i] = nrow;
	  mihead_ogzmi->ncol[i] = ncol;
	  mihead_ogzmi->bptr[i] = iblk;
	  mihead_ogzmi->nblk[i] = nblk;
	  
	  iblk += 1 + nblk;
	}
      }
    }
    
    if(needcheck){
      for(i=0;i<n;i++){
	if(-1 == gzmi_rblock(iu,iblk,ihead,1,err)){
	  fputs("Can't read gzmi-file image headers\n",ou);
	  parse_frmi(mihead_ogzmi);
	  return(NULL);
	}
	ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
	nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
	nblk = PARSE_BLOCK_SHORT(ncol*nrow);
	
	if(i==0){
	  max_ncol = min_ncol = ncol;
	  max_nrow = min_nrow = nrow;
	}else{
	  if(ncol > max_ncol) max_ncol = ncol;
	  if(ncol < min_ncol) min_ncol = ncol;
	  if(nrow > max_nrow) max_nrow = nrow;
	  if(nrow < min_nrow) min_nrow = nrow;
	}
	mihead_ogzmi->nrow[i] = nrow;
	mihead_ogzmi->ncol[i] = ncol;
	mihead_ogzmi->bptr[i] = iblk;
	mihead_ogzmi->nblk[i] = nblk;
	
	memcpy(lbl,(char *)&ihead[PARSE_MI_LOC_LABEL],
	       PARSE_MI_LEN_LABEL);
	lbl[PARSE_MI_LEN_LABEL] = 0;
	mihead_ogzmi->label[i] = parse_malloc_char(strlen(lbl)+1);
	strcpy(mihead_ogzmi->label[i],lbl);
	
	iblk += 1 + nblk;
      }
    }
  }else{
    for(i=0;i<n;i++){
      if(-1 == gzmi_rblock(iu,head[i],ihead,1,err)){
	fputs("Can't read gzmi-file image headers\n",ou);
	parse_frmi(mihead_ogzmi);
	return(NULL);
      }
      ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
      nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
      nblk = PARSE_BLOCK_SHORT(ncol*nrow);
      
      if(i==0){
	contig = 1;
	order = 1;
	max_ncol = min_ncol = ncol;
	max_nrow = min_nrow = nrow;
      }else{
	if(ncol > max_ncol) max_ncol = ncol;
	if(ncol < min_ncol) min_ncol = ncol;
	if(nrow > max_nrow) max_nrow = nrow;
	if(nrow < min_nrow) min_nrow = nrow;
	if(head[i] <= head[i-1]) order = 0;
	if(head[i-1]+nblkp+1 != head[i]) contig = 0;
      }
      mihead_ogzmi->nrow[i] = nrow;
      mihead_ogzmi->ncol[i] = ncol;
      mihead_ogzmi->bptr[i] = head[i];
      mihead_ogzmi->nblk[i] = nblk;
      
      memcpy(lbl,(char *)&ihead[PARSE_MI_LOC_LABEL],
	     PARSE_MI_LEN_LABEL);
      lbl[PARSE_MI_LEN_LABEL] = 0;
      mihead_ogzmi->label[i] = parse_malloc_char(strlen(lbl)+1);
      strcpy(mihead_ogzmi->label[i],lbl);
      
      nblkp  = nblk;
    }
  }
  
  same = 1;
  nrow = mihead_ogzmi->nrow[0];
  ncol = mihead_ogzmi->ncol[0];
  for(i=1;i<n;i++){
    if(nrow != mihead_ogzmi->nrow[i]){
      same = 0;
      break;
    }
    if(ncol != mihead_ogzmi->ncol[i]){
      same = 0;
      break;
    }
  }
  
  mihead_ogzmi->same = same; 
  mihead_ogzmi->contig = contig; 
  mihead_ogzmi->order = order; 
  mihead_ogzmi->max_ncol = max_ncol;
  mihead_ogzmi->min_ncol = min_ncol;
  mihead_ogzmi->max_nrow = max_nrow;
  mihead_ogzmi->min_nrow = min_nrow;
  
  return(mihead_ogzmi);
}

void
gzmi_push(int e)
{
	envs[e].gzmihead = gzmihead;
}

void
gzmi_pop(int e)
{
	gzmihead = envs[e].gzmihead;
}

int 
gzmi_image(int *nx,int *ny,int image_num,char *label)
{
	int i;

	if(gzmihead == NULL){
		fputs("mi_image (internal): no gzmihead\n",ttyout);
		return(-1);
	}

	if(image_num < 1 || image_num > gzmihead->n){
		fputs("mi_image (internal): image number out of range\n",ttyout);
		return(-1);
	}

	image_num--;
	i = gzmihead->bptr[image_num] + 1;

	if(-1 == gzseek(gz_image_fd,(long) i * 512L,0)){
		fputs("Error in seeking to image\n",ttyout);
		return(-1);
	}

	strncpy(label,gzmihead->label[image_num],PARSE_MI_LEN_LABEL+1);
	*nx = gzmihead->ncol[image_num];
	*ny = gzmihead->nrow[image_num];
	return(0);
}

int
gzmi_total(void)
{ 
  /* reset error flag */
  errno = 0;
  gzmihead = parse_ogzmi_fe(gz_image_fd,0,ttyout,errno);
  if(gzmihead == NULL){
    return(-1);
  }else{
    return(gzmihead->n);
  }
}

void
gzmi_close(void)
{ 
  if(gzmihead == NULL){
    fputs("mi_close (internal): no gzmihead\n",ttyout);
  }else{
    parse_frmi(gzmihead);
  }
}

