/*
handle_port - handle port (via socket) to other programs
*/

#ident "$Id: handle_port.c,v 1.10 2011/09/14 11:08:54 carl Exp $"

#define OWTOOLKIT_WARNING_DISABLED
#include "xpic.h"
#include "xpic_port.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <xview/xview.h>

static int socket_handle;
static int socket_fd;
static int debug=0;
static int id1,id2;
static int *me_fd = &id1;
static int *me_sock = &id2;
static int packet_size;
static char packet[PACKET_LENGTH];
static int mode;
static int transpose;
static int mouse_reply=-1;	/* >=0 need to send reply in port_button_proc */
static int port_num = PORT_SOCKET;

static int get_connection(int s);
static void process_packet(void);
static void tell_user(char *s);
static void close_socket(void);

int
port_button_proc(int b)
{
	int stat;

	if(port_status == 2 && (mouse_reply == b || mouse_reply == 3)){
		stat = b;
		if(-1 == write(socket_fd,&stat,sizeof(stat))) perror("xpic: port_button_proc: write error");
		mouse_reply = -1;
		return(1);
	}else{
		return(0);
	}
}

static void port_test(void)
{
	int i;
	TEST *t;
	int stat;

	t = (TEST *)packet;
	stat = 0;
	for(i=0;i<P_TEST_WORDS;i++)if(t->buf[i] != i){
		stat = 1;
		break;
	}
	if(-1 == write(socket_fd,&stat,sizeof(stat))) perror("xpic: port_test");
}

static void port_message(void)
{
	MESSAGE *t;

	t = (MESSAGE *)packet;
	fputs(t->message,ttyout);
	fflush(ttyout);
}

static void port_annotate(void)
{
	ANNOTATE *t;

	t = (ANNOTATE *)packet;
	annotate_add(t->x,t->y,t->message);
}

static void port_stat(void)
{
	STATUS_INFO s;

	s.x = canvas_size_x;
	s.y = canvas_size_y;
	s.window = window;
	s.level = level;
	s.mi_max = mi_max;
	s.canvas_id = display_canvas_id();
	if(-1 == write(socket_fd,&s,sizeof(s))) perror("xpic: port_stat");
}

static void port_load(void)
{
	load_pix(0);
}

static void port_wl(void)
{
	WL *w;

	w = (WL *)packet;
	window = w->window;
	level = w-> level;
	window_level_lut(w->window,w->level);
	display_set_sliders();
	load_pix(0);
}

static void port_debug(void)
{
	DEBUG_PORT *t;

	t = (DEBUG_PORT *)packet;
	debug = t->debug;
}

static void port_clear(void)
{
	reset();
	do_clear(1);
}

static void port_write(void)
{
	WRITE *w;
	CPIXEL *ip;
	int inc,i;

	w = (WRITE *)packet;
	if(w->y < 0 || w->y >= canvas_size_y) return;
	if(w->x < 0 || w->x >= canvas_size_x) return;
	if(w->n <= 0)return;
	if(transpose){
		if(w->y + w->n > canvas_size_y) w->n = canvas_size_y - w->y;
		inc = canvas_size_x;
	}else{
		if(w->x + w->n > canvas_size_x) w->n = canvas_size_x - w->x;
		inc = 1;
	}
	ip = image16 + w->y*canvas_size_x + w->x;
	for(i=0;i<w->n;){
		*ip = w->data[i++];
		if((int)*ip > mi_max)*ip = mi_max;
		ip += inc;
	}
}

static void port_read(void)
{
	READ *w;
	READ_INFO rd;
	CPIXEL *ip;
	int inc=0,i,np;		 /* make ansi happier w/ initialization */

	w = (READ *)packet;
	if(w->y < 0 || w->y >= canvas_size_y
	 || w->x < 0 || w->x >= canvas_size_x){
		np = 0;
	}else{
		if(transpose){
			if(w->y + w->n > canvas_size_y){
				np = canvas_size_y - w->y;
			}else{
				np = w->n;
			}
			inc = canvas_size_x;
		}else{
			if(w->x + w->n > canvas_size_x){
				np = canvas_size_x - w->x;
			}else{
				np = w->n;
			}
			inc = 1;
		}
	}
	rd.n = np;
	if(np){
		ip = image16 + w->y*canvas_size_x + w->x;
		for(i=0;i<np;){
			rd.data[i++] = *ip;
			ip += inc;
		}
	}

	if(-1 == write(socket_fd,&rd,sizeof(rd) - (XP_LINE_MAX - w->n)*sizeof(short)))
		perror("xpic: port_read");
}

static void port_transpose(void)
{
	TRANSPOSE *t;

	t = (TRANSPOSE *)packet;
	transpose = t->transpose;
}

static void port_cursor(void)
{
	CURSOR_INFO s;

	if(cursor_stat){
		s.type = box.status + 1;
	}else{
		s.type = 0;
	}
	s.x = box.lx;
	s.y = box.ly;
	s.xw = box.xw;
	s.yw = box.yw;
	
	if(-1 == write(socket_fd,&s,sizeof(s))) perror("xpic: port_cursor");
}

static void port_command(void)
{
	COMMAND *t;

	t = (COMMAND *)packet;
	fprintf(ttyout,"port> %s\n",t->cmd);
	fflush(ttyout);
	command(t->cmd,1);
}

static 
Notify_value socket_read(me,fd)
int *me;
int fd;
{
	int n;
	char c[512];

	if(0 < (n = read(socket_fd,c,512))){
		if(debug>2){
			fprintf(ttyout,"received packet of length %d\n",n);
			fflush(ttyout);
		}
		ring_put(c,n);
		process_packet();
	}else{
		if(debug)tell_user("Port access completed\n");
		close_socket();
	}

	return(NOTIFY_DONE);
}

static Notify_value socket_connect(me,fd)
int *me;
int fd;
{
	socket_fd = get_connection(socket_handle);
	if(socket_fd == -1){
		tell_user("socket accept error\n");
		return(NOTIFY_DONE);
	}
	if(debug)tell_user("Port access initiated\n");
	notify_set_input_func((int)me_fd,socket_read,socket_fd);
	notify_set_input_func((int)me_sock,NOTIFY_FUNC_NULL,socket_handle);
	mouse_reply = -1;
	port_status = 2;
	mode = 0;
	debug = 0;
	transpose = 0;
	ring_init();
	return(NOTIFY_DONE);
}

static void
close_socket(void)
{
	notify_set_input_func((int)me_fd,NOTIFY_FUNC_NULL,socket_fd);
	close(socket_fd);

	notify_set_input_func((int)me_sock,socket_connect,socket_handle);
	port_status = 1;
	mouse_reply = -1;
}

static void 
port_close(void)
{
	close_socket();
}

static void port_move(void)
{
	MOVE *w;
	int x,y;

	w = (MOVE *)packet;
	x = w->x;
	y = w->y;
	line_add(x,y,0);
}

static void port_line(void)
{
	LINE *w;
	int x,y;

	w = (LINE *)packet;
	x = w->x;
	y = w->y;
	line_add(x,y,1);
}

static void port_exit(void)
{
	close_socket();
	quit_guts();
	exit(0);
}

static void port_mouse(void)
{
	MOUSE *w;
	int b;

	w = (MOUSE *)packet;
	b = w->b;
	mouse_reply = b;
}

void 
port_set_port(int p)
{
	port_num = p;
}

int
port_get_port(void)
{
	return(port_num);
}

static int establish(int port)
{
	int    s;
	struct sockaddr_in sa;
	struct hostent *hp;
	u_short portnum;
	int on=1;

	portnum = port;
	memset(&sa,0,sizeof(struct sockaddr_in));   /* clear our address */
	hp= gethostbyname(host_name);               /* get our address info */
	if (hp == NULL)                             /* we don't exist !? */
		return(-1);
	sa.sin_family= hp->h_addrtype;              /* this is our host addr */
	sa.sin_port= htons(portnum);                /* this is our port # */
	if ((s= socket(AF_INET,SOCK_STREAM,0)) < 0) /* create socket */
		return(-1);

					/* allow reconnection */
	if(setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on)) < 0)
		return(-1);

					/* bind address to socket */
	if (bind(s,(struct sockaddr *)&sa,sizeof sa) < 0) 
		return(-1);                               
	return(s);
}

void
handle_port(void)
{
	if(!port_access){
		notify_set_input_func((int)me_sock,NOTIFY_FUNC_NULL,socket_handle);
		port_status = 0;
		shutdown(socket_handle,2);
		close(socket_handle);
	}else{
		socket_handle = establish(port_num);
		if(socket_handle == -1){
			pperror("handle_port: establish socket");
			port_access = 0;
			return;
		}
		if(listen(socket_handle,1) == -1){
			port_access = 0;
			pperror("handle_port: rec/listen");
			return;
		}
		notify_set_input_func((int)me_sock,socket_connect,socket_handle);
		port_status = 1;
	}
}

static void
port_syntax(void)
{
	fputs("Synopsis: manage port access via xpic_open\n",ttyout);
	min_abbreviation();
	fputs("Related: status\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("port on            * enable access\n",ttyout);
	fputs("port off           * disable access\n",ttyout);
	fputs("port               * print status of port access\n",ttyout);
}

void
cmd_port(void)
{
	if(argc < 1 || argc > 2){
		port_syntax();
		return;
	}
	if(argc == 1){
		if(port_status == 0){
			fputs("Status: closed\n",ttyout);
		}else if(port_status == 1){
			fputs("Status: waiting\n",ttyout);
		}else{
			fputs("Status: active\n",ttyout);
		}
		fprintf(ttyout,"Socket: %d\n",port_num);
	}else{
		if(!strcmp("on",argv[1])){
			if(!port_access){
				port_access = 1;
				handle_port();
			}
		}else if(!strcmp("off",argv[1])){
			if(port_access){
				if(port_status == 2){
				       fputs("port active. try again\n",ttyout);
				}else{
					port_access = 0;
					handle_port();
				}
			}
		}else{
			fputs("On or off please!\n",ttyout);
		}
	}
}

static void
tell_user(char *s)
{
	fputs(s,ttyout);
	fflush(ttyout);
}

static void
process_packet(void)
{
	int type;

	while(ring_length()){
		if(mode == 0){
			if(ring_length() < sizeof(int)) return;
			ring_get((char *)&packet_size,sizeof(int));
			if(debug>2){
				fprintf(ttyout,"packet size: %d\n",packet_size);
				fflush(ttyout);
			}
			if(packet_size < 4 || packet_size > PACKET_LENGTH){
				tell_user("packet size error\n");
				close_socket();
				return;
			}
			mode = packet_size -= sizeof(packet_size);
		}

		if(ring_length() < mode) return;

		memcpy( packet, &packet_size, sizeof(int));
		ring_get(packet+sizeof(int),packet_size);
		type = *((int *)(packet + sizeof(int)));
		if(debug>1){
			fprintf(ttyout,"packet type: %d\n",type);
			fflush(ttyout);
		}

		switch(type){

		case P_TEST:
			port_test();
			break;
		case P_MESSAGE:
			port_message();
			break;
		case P_STATUS:
			port_stat();
			break;
		case P_CURSOR:
			port_cursor();
			break;
		case P_ANNOTATE:
			port_annotate();
			break;
		case P_COMMAND:
			port_command();
			break;
		case P_CLOSE:
			port_close();
			break;
		case P_WL:
			port_wl();
			break;
		case P_LOAD:
			port_load();
			break;
		case P_DEBUG_PORT:
			port_debug();
			break;
		case P_CLEAR:
			port_clear();
			break;
		case P_WRITE:
			port_write();
			break;
		case P_READ:
			port_read();
			break;
		case P_TRANSPOSE:
			port_transpose();
			break;
		case P_MOVE:
			port_move();
			break;
		case P_LINE:
			port_line();
			break;
		case P_EXIT:
			port_exit();
			break;
		case P_MOUSE:
			port_mouse();
			break;
		default:
			tell_user("unknown packet type\n");
			close_socket();
			return;
		}
		mode = 0;
		fflush(ttyout);
	}
}


static int get_connection(int s)
{
	struct sockaddr_in isa; /* address of socket */
	socklen_t i;                  /* size of address */
	int sock=-1;

	errno = EINTR;
	while (errno == EINTR) {
		i = sizeof(isa);
		if ((sock = accept(s,(struct sockaddr *)&isa,&i)) < 0) {
			if(errno != EINTR) return(-1);
		} else errno = 0;
	}

	if(sock == -1) parse_er("xpic internal error: get_connection: sock");
	return(sock);
}



