/*
kb - handle keyboard i/o
*/

#ident "$Id: kb.c,v 1.18 2011/09/14 12:11:41 carl Exp $"

#include 	"xpic.h"
#include	<setjmp.h>
#include 	<pwd.h>
#include	<signal.h>

static sigjmp_buf env;
static int nargc;
static char *nargv[MAX_CMD_WORDS];
static void interrupt(int);
static void cleanup(int);
static int fi,fo;
static FILE *log_fd = NULL;

static void handle_file(FILE *fd,int echo,char *name);

static int cine_first=1;	/* cine first image */
static int cine_last=1;		/* cine last image */
static int cine_pause=0;	/* cine pause (ms) */
static int cine_inc=1;		/* cine increment, by frames */

static void
sendbuf(char *obuf,int doexit)
{
	char rchar[2];
	char pathname[MAXPATHLEN];

	if(-1 == write(fo,obuf,strlen(obuf))) perror("xpic: kb: sendbuf: write error");
	if(doexit){
		sleep(1);
		exit(0);
	}
	if(2 != read(fi,rchar,2) || rchar[0] != EOC){
		fputs("xpic message error\n",stderr);
		exit(1);
	}

	if(rchar[1]){
		if(MAXPATHLEN != read(fi,pathname,MAXPATHLEN)){
			fputs("xpic message error - reading directory\n",stderr);
			exit(1);
		}
		if(-1 == chdir(pathname)) perror("xpic: kb: sendbuf: chdir error");
#ifdef DEBUG
		parse_prs("kb changing path to",pathname);
#endif
	}
}

static void
get_xpic_dir(void)
{
	sendbuf(SPECIALMODE,0);
}

#ifdef KBPWD
static void
kb_pwd(void)
{
	char pathname[MAXPATHLEN];

	fprintf(ttyout,"pwd: %s\n",getcwd(pathname,MAXPATHLEN));
}
#endif

void
cine_syntax(void)
{
	fputs("Synopsis: loop through images\n",ttyout);
	min_abbreviation();
	fputs("Related: +, -\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("cine first last [-s skip]   * loop from first to last images\n",ttyout);
	fputs("cine first last p [-s skip] * pause p msec between images\n",ttyout);
	fputs("cine                        * use previous values of first, last, p\n\n",ttyout);
	fputs("*options\n",ttyout);
	fputs("\t-s\t\t    *specify frame increment (skip) count\n",ttyout);
}

static void 
cine()
{
	char num[16];
	int i;

	if(nargc >= 3 && nargc <= 6){
		i = parse_atof(nargv[1]);
		if(parse_atof_error()) return;
		if(i < 1){
			puts("first image out of range");
			return;
		}
		cine_first = i;

		i = parse_atof(nargv[2]);
		if(parse_atof_error()) return;
		cine_last = i;
		if(i < 1){
			puts("last image out of range");
			return;
		}
		/*this is messy, no?*/
		/*the user specified more than first/last*/
		if(nargc > 3){
			/* check if the next option is a -s or a pause */
			if(!strcmp(nargv[3],"-s")){
				if(nargc == 5){
					i = parse_atof(nargv[4]);
					if(parse_atof_error()) return;
					if(i < 0){
						puts("skip out of range");
						return;
					}
					cine_inc = i;
					/*no pause arg*/
					cine_pause=0;
				}else{
					cine_inc=1;
					puts("bad syntax: -s without explicit argument");
					return;
				}
			/* no -s, assume it was a pause */
			}else if(nargc > 3){
				i = parse_atof(nargv[3]);
				if(parse_atof_error()) return;
				if(i < 0){
					puts("pause out of range");
					return;
				}
				cine_pause = i*1000;

				/*there was a pause, is there a -s now?*/				
				if(nargc > 4){
				 	if (!strcmp(nargv[4],"-s")){
						if(nargc==6){
							i = parse_atof(nargv[5]);
							if(parse_atof_error()) return;
							if(i < 0){
								puts("skip out of range");
								return;
							}
							cine_inc = i;
						}else{ /*no s arg specified*/
							cine_inc=1;
							puts("bad syntax: -s without explicit argument");
							return;
						}
					}else{ /*no s arg specified*/
						cine_inc=1;
						puts("warning: extra parameters ignored");
					}
				}else{ /*no s arg specified*/
					cine_inc=1;
				}
			}else{ /*no s arg or pause arg */
				cine_inc = 1;
				cine_pause = 0;
			}
		}
	}else if(nargc != 1){
		cine_syntax();
		return;
	}
	
	/* adjust the last slice index */
	
	if(cine_inc != 1){
		i = cine_last - cine_first;
		i /= cine_inc;
		cine_last = cine_first+i*cine_inc;
	}

	i = cine_first;
	while(1){
		sprintf(num,"%d",i);
		sendbuf(num,0);
		if(cine_last >= cine_first){
			if(i == cine_last) break;
			i+=cine_inc;
		}else{
			if(i == cine_last) break;
			i-=cine_inc;
		}
		usleep(cine_pause);
	}
}

static void
prompt(n)
char *n;
{
	fputs("xpic",stdout);
	if(*n)printf("[%s]",n);
	fputs("> ",stdout);
}

static char *
tilde(char *home)
{
	static char path[BUFL];
	struct passwd *pw;
	char *s,*sp;
	int uid;

	s = home;
	if(*home == '~'){
		if(isalpha(s[1])){
			sp = ++s;
			while(isalnum(*s)) s++;
			strncpy(path,sp,s-sp);
			path[s-sp] = 0;
			if(NULL == (pw = getpwnam(path))) return(NULL);
		}else{
			uid = geteuid();
			if(NULL == (pw = getpwuid(uid))) return(NULL);
			s++;
		}
		strcpy(path,pw->pw_dir);
		strncat(path,s,BUFL-strlen(path));
	}else{
		strncpy(path,s,BUFL);
	}
	return(path);
}


static void 
interrupt(int a)
{
  fprintf(stderr,"interrupt\n");
  a = 0;
  signal(SIGINT,interrupt); 
  siglongjmp(env,0);
}

static void 
cleanup(int a)
{
	a = 0;
	fprintf(stderr,"in cleanup\n");
	exit(1);
}

static void
source(void)
{
	char name[BUFL];
	char *np;
	FILE *sf;

	if(nargc != 2){
		fputs("Syntax: source file\n",stdout);
		return;
	}

	if(NULL == (np = tilde(nargv[1]))) return;

	if(NULL == (sf = fopen(np,"r"))){
		fprintf(stdout,"Can't open: %s\n",np);
		return;
	}
	strcpy(name,np);

	handle_file(sf,1,name);
	fclose(sf);
}

static void
kb_pause(void)
{
	char r[2];

	fputs("hit RETURN to continue",stdout);
	if(NULL == fgets(r,2,stdin)) parse_er("xpic: kb: kb_pause: fputs error");
}

static void
log_file(void)
{
	char ofn[BUFL];
	int over;
	struct stat sbuf;

	if(nargc != 2){
		fputs("Syntax: log [!]file\n",stdout);
		return;
	}

	strcpy(ofn,nargv[1]);
	if(ofn[0] == '!'){
		over = 1;
		strcpy(ofn,ofn+1);
	}else{
		over = 0;
	}

	if(!stat(ofn,&sbuf) && !over){
		fputs("File exists. Use 'log !file' to overwrite\n",stdout);
		return;
	}
	unlink(ofn);
	if((log_fd=fopen(ofn,"w")) == NULL){
		fprintf(stdout,"Can't open: %s\n",ofn);
		return;
	}
}

static void
start_up(void)
{
	FILE *fd;
	char buf[BUFL];
	char fn[BUFL];
	char *s;
	char pn[MAXPATHLEN];

	if((s = getenv("HOME")) == NULL)return;
	strcpy(fn,s);
	strcat(fn,"/");
	strcat(fn,START_UP_FILE);
	if((fd = fopen(fn,"r")) != NULL){
		strcpy(fn,"~/");
		strcat(fn,START_UP_FILE);
		while(fgets(buf,BUFL,fd) != NULL){
			if(buf[0] == '/' && buf[1] == '/'){ 
				handle_file(fd,1,fn);
				break;
			}
		}
		fclose(fd);
	}

	if(0 == getcwd(pn,MAXPATHLEN)) return;
	if(!strcmp(pn,s)) return;

	if((fd = fopen(START_UP_FILE,"r")) != NULL){
		while(fgets(buf,BUFL,fd) != NULL){
			if(buf[0] == '/' && buf[1] == '/'){ 
				handle_file(fd,1,START_UP_FILE);
				break;
			}
		}
		fclose(fd);
	}
}

static void
handle_file(FILE *fd,int echo,char *name)
{
	char buf1[KB_BUF],buf[KB_BUF],*b;
	int doexit=0,m=0,send=0;

	signal(SIGBUS,cleanup); 
	signal(SIGINT,interrupt); 
	signal(SIGTSTP,SIG_IGN);
	sigsetjmp(env,1);

	prompt(name);
	while(NULL != fgets(buf,KB_BUF,fd)){

#ifdef TESTIT
		if(strlen(buf) > 1 && echo){
			fprintf(stdout,"carl: %s",buf);
		}
		printf("%d %d\n",strlen(buf),echo);
#else
#endif
		if(echo)fputs(buf,stdout);

		if(log_fd != NULL){
			if(echo)fprintf(log_fd,"#[%s]: ",name);
			if(EOF == fputs(buf,log_fd)){
				puts("EOF on fputs");
			}
		}

		b = buf;
		while(*b == ' ' || *b == '\t')b++;
		if(*b == '#'){
			;
		}else if(*b == '!'){
			b++;
			while(*b == ' ' || *b == '\t')b++;
			if(*b == '\n' || *b == 0){
				if(-1 == system("/bin/csh"))perror("xpic: handle_file: system error");
			}else{
				if(-1 == system(b))perror("xpic: handle_file: system error");
			}
		}else{
			send = 1;
			doexit = 0;
			strcpy(buf1,buf);
			words(buf,&nargc,nargv,MAX_CMD_WORDS);
			if(nargc && 0 <= (m = lookup(nargv[0],0))){
				if(!strcmp(clist[m].name,"source")){
					send = 0;
					get_xpic_dir();
					source();
				}else if((nargc && 0 <= (lookup(nargv[0],0))) && (!strcmp(clist[m].name,"ls"))){
					get_xpic_dir();
					if(-1 == system(b)) perror("xpic: handle_file: system error");
				}else if(!strcmp(clist[m].name,"pause")){
					send = 0;
					kb_pause();	
				}else if(!strcmp(clist[m].name,"cine")){
					send = 0;
					cine();	
				}else if(!strcmp(clist[m].name,"exit")||!strcmp(clist[m].name,"quit")){
				    if(!echo){
					fputs("Really exit? [y/n] ",stdout);
					if(NULL == fgets(buf,KB_BUF,fd)) parse_er("xpic: kb: handle_file: write error");
					if(*buf == 'y' || *buf == 'Y'){
						doexit = 1;
					}else{
						doexit = 0;
					}
				    }else{
					doexit = 1;
				    }
				    if(doexit){
					if(log_fd != NULL)fclose(log_fd);
				    }else{
					send = 0;
				    }
				}else if(!strcmp(clist[m].name,"log")){
					send = 0;
					get_xpic_dir();
					log_file();	
#ifdef KBPWD
				}else if(!strcmp(clist[m].name,"pwd")){
					get_xpic_dir();
					kb_pwd();
#endif
			 	}
			}
		}

		if(send){
			sendbuf(buf1,doexit);
		}
			

		
		prompt(name);
	}
	if(echo)puts("End-Of-File");
}

void
kb(int c_argc,char **c_argv)
{
	char c;
	int init;

	close(parse_atof(c_argv[2]));
	fo = parse_atof(c_argv[3]);
	fi = parse_atof(c_argv[4]);
	close(parse_atof(c_argv[5]));
	init = atoi(c_argv[6]);
	parse_atof_args(stdout,1);	
	ttyout = stdout;

	if(init) start_up();

	handle_file(stdin,0,"");
	if(log_fd != NULL)fclose(log_fd);
	if(-1 == write(fo,"exit",strlen("exit"))) perror("xpic: kb: write error");
	sleep(1);
	exit(0);
	if(-1 == read(fi,&c,1)) perror("xpic: kb: read error");
}
