/*
lm - handle lm files
*/

#ident "$Id: lm.c,v 1.8 2007/05/22 13:22:45 ccrawford Exp $"

#include <xpic.h>
#include "lm.h"

static int
lm_depth(void)
{ 
	int d;

	if(-1 == fseek(image_fd,LM_DEPTH_OFFSET,SEEK_SET)){
		fputs("Error in seeking to LM depth\n",ttyout);
		return(-1);
	}
	if(1 != fread(&d,sizeof(int),1,image_fd)){
		fputs("Can't read LM depth\n",ttyout);
		return(-1);
	}
	d = ican_int(d);
	if(d < 1 || d > LM_MAX_SLICES){
		fprintf(ttyout,"Bad number of LM slices (%d)\n",d);
		return(-1);
	}
	return(d);
}

int 
lm_image(int *w,int *h,int i,char *label)
{
	char bag_type;
	long offset;
	int delta;
	short phase;			// phase of first image
	char id[LM_BAG_ID_LENGTH];	// bag id
	char s[PARSE_MI_LEN_LABEL];	// for building the label

	if(-1 == fseek(image_fd,LM_BAG_ID_OFFSET,SEEK_SET)){
		fputs("lmmi: error in seeking to LM id\n",stderr);
		exit(1);
	}

	if(LM_BAG_ID_LENGTH != fread(id,sizeof(char),LM_BAG_ID_LENGTH,image_fd)){
		fputs("lmmi: can't read LM id\n",stderr);
		exit(1);
	}

	if(-1 == fseek(image_fd,LM_BAG_PHASE_OFFSET,SEEK_SET)){
		fputs("lmmi: error in seeking to LM phase\n",stderr);
		exit(1);
	}
	if(1 != fread(&phase,sizeof(short),1,image_fd)){
		fputs("lmmi: can't read LM phase\n",stderr);
		exit(1);
	}
	phase = ican_short(phase);

	if(isub == I_LM_HIGH || isub == I_LM_LOW){

#ifdef MERLIN
		bag_type = LM_BAG_TYPE_A;
#else
		if(-1 == fseek(image_fd,LM_BAG_TYPE_OFFSET,SEEK_SET)){
			fputs("Error in seeking to LM type\n",ttyout);
			return(-1);
		}
		if(1 != fread(&bag_type,sizeof(char),1,image_fd)){
			fputs("Can't read LM type\n",ttyout);
			return(-1);
		}
		if(bag_type != LM_BAG_TYPE_A && bag_type != LM_BAG_TYPE_B){
			fputs("Illegal LM bag type\n",ttyout);
			return(-1);
		}
#endif

		if(isub == I_LM_HIGH){
			if(bag_type == LM_BAG_TYPE_A){
				*w = LM_HE_A_WIDTH;
				*h = LM_HE_A_HEIGHT;
			}else{
				*w = LM_HE_B_WIDTH;
				*h = LM_HE_B_HEIGHT;
			}
			offset = LM_HE_FIRST_SLICE;
			delta = LM_HE_NEXT_SLICE;
		}else{
			if(bag_type == LM_BAG_TYPE_A){
				*w = LM_LE_A_WIDTH;
				*h = LM_LE_A_HEIGHT;
			}else{
				*w = LM_LE_B_WIDTH;
				*h = LM_LE_B_HEIGHT;
			}
			offset = LM_LE_FIRST_SLICE;
			delta = LM_LE_NEXT_SLICE;
		}
		i--;
		if(-1 == fseek(image_fd,offset+i*delta,SEEK_SET)){
			fputs("Error in seeking to LM image\n",ttyout);
			return(-1);
		}
	}else{
		*w = LM_PROJ_WIDTH;
		*h = lm_depth();
		if(*h == -1) return(-1);

		switch(i){
			case 1: offset= LM_PROJ_00_OFFSET; break;
			case 2: offset= LM_PROJ_45_OFFSET; break;
			case 3: offset= LM_PROJ_90_OFFSET; break;
			default:
				fputs("Bad LM projection number\n",ttyout);
				return(-1);
		}
		if(-1 == fseek(image_fd,offset,SEEK_SET)){
			fputs("Error in seeking to LM image\n",ttyout);
			return(-1);
		}
	}

	strncpy(label,"ID: ",PARSE_MI_LEN_LABEL);
	strncat(label,id,PARSE_MI_LEN_LABEL);
	switch(isub){
	case I_LM_HIGH:
		strncat(label," HE ",PARSE_MI_LEN_LABEL);
		break;
	case I_LM_LOW:
		strncat(label," LE ",PARSE_MI_LEN_LABEL);
		break;
	case I_LM_PROJ:
		strncat(label," PROJ-",PARSE_MI_LEN_LABEL);
		sprintf(s,"%d ",(i-1)*45);
		strncat(label,s,PARSE_MI_LEN_LABEL);
		break;
	}

	sprintf(s,"phase=%d",phase);
	strncat(label,s,PARSE_MI_LEN_LABEL);

	// strncat(label," file: ",PARSE_MI_LEN_LABEL);
	// strncat(label,ifn,PARSE_MI_LEN_LABEL);

      	return(0);
}

int
lm_total(void)
{ 
	switch(isub){

	case I_LM_HIGH: 
	case I_LM_LOW:
		return(lm_depth());
	case I_LM_PROJ:
		return(3);
	default:
		fputs("Bad LM type\n",ttyout);
		return(-1);
	}
}
