/*
lut - look-up-table routines
*/

#ident "$Id: lut.c,v 1.12 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static int gamma_tab[LUT_SIZE_P];
static int lut_size_save, lut_scale_save;
static int fall = 0;
static int range_red=0;
static int red_low=8000,red_high=8100;
static int range_blue=0;
static int blue_low=8200,blue_high=8300;
static int range_green=0;
static int green_low=8400,green_high=8500;
static int range_outline=0;

void
window_level_lut(int window,int level)
{
	float delta;
	int i,j;
	int lowp,highp,low,high,width;
	int lowv,highv;
	unsigned int v;

	low = level + zero_offset - window/2;
	high = level + zero_offset + window/2;
	if(low > mi_max) low = mi_max;
	if(high < 0) high = 0;
	lowp = (low < 0)? 0 : low;
	highp = (high > mi_max)? mi_max : high;
	width = high - low;
	if(width == 0) width = 1;

	if(!rvideo){
		lowv = 0;
		highv = LUT_SIZE_P - 1;
	}else{
		highv = 0;
		lowv = LUT_SIZE_P - 1;
	}

	v = gamma_tab[lowv];
	for(i=0;i<lowp;i++) mi_lut[i] = v;

	v = gamma_tab[highv];
	if(fall){
		v = gamma_tab[lowv];
	}
	for(i=highp;i<=mi_max;i++) mi_lut[i] = v;

	delta = (float) LUT_SIZE_P / width; if(rvideo) delta = -delta;
	for(i=lowp;i<highp;i++){
		v = gamma_tab[(int)((i-low) * delta + lowv)];  
		mi_lut[i] = v;
	}

	if(range_red){
		for(i=0;i<=mi_max;i++){
			if(i >= red_low+zero_offset && i <= red_high+zero_offset){
				mi_lut[i] = cms_map[lut_red];
			}
		}
	}

	if(range_blue){
		for(i=0;i<=mi_max;i++){
			if(i >= blue_low+zero_offset && i <= blue_high+zero_offset){
				mi_lut[i] = cms_map[lut_blue];
			}
		}
	}

	if(range_green){
		for(i=0;i<=mi_max;i++){
			if(i >= green_low+zero_offset && i <= green_high+zero_offset){
				mi_lut[i] = cms_map[lut_green];
			}
		}
	}

	if(range_outline){
		for(i=0;i<OUTLINE_LENGTH;i++){
			j = i + outline_start;
			if(j >= 0 && j < mi_max){
				mi_lut[j] = cms_map[lut_outline+i];
			}
		}
	}

#ifdef DEBUG
	{
	static float *tmpf;

	tmpf = parse_malloc_float(mi_max);

	if(dsp_depth == 8){
		parse_wtext("lut saved in lut.da");
		for(i=0;i<mi_max;i++) tmpf[i] = mi_lut[i];
		parse_wda(tmpf,mi_max,1,"lut");
	}else{
		parse_wtext("lut saved in lut.{red,green,blue}.da");
		for(i=0;i<mi_max;i++) tmpf[i] = (0xFF0000 & mi_lut[i]) >> 16;
		parse_wda(tmpf,mi_max,1,"lut_red");

		for(i=0;i<mi_max;i++) tmpf[i] = (0x00FF00 & mi_lut[i]) >> 8;
		parse_wda(tmpf,mi_max,1,"lut_green");

		for(i=0;i<mi_max;i++) tmpf[i] = 0x0000FF & mi_lut[i];
		parse_wda(tmpf,mi_max,1,"lut_blue");
	}

	parse_free((char *)tmpf);

	}
#endif
}

void
set_gamma_tab(double value,int map)
{
	int i;
	float delta,gammap;
	float t[LUT_SIZE_P];

	t[0] = gamma_tab[0] = 0;
	gammap = 1.0 / value;
	delta = 1.0 / LUT_SIZE_P;
	for(i=1;i<LUT_SIZE_P;i++){
		t[i] = gamma_tab[i] = lut_size * pow(i*delta,gammap);
	}
	if(map){
		for(i=0;i<LUT_SIZE_P;i++){
			gamma_tab[i] = cms_map[gamma_tab[i]];
		}
	}

#ifdef DEBUG
	{
	float *tmpf;

	parse_wtext("gamma saved in gamma.da");
	tmpf = parse_malloc_float(LUT_SIZE_P);
	for(i=0;i<LUT_SIZE_P;i++) tmpf[i] = 0xFF & gamma_tab[i];
	parse_wda(tmpf,LUT_SIZE_P,1,"gamma");
	parse_free((char *)tmpf);
	}
#endif

	window_level_lut(window,level);
}

void
gamma_set_device(double gamma_device)
{
	lut_scale_save = lut_scale;
	lut_size_save = lut_size;
	lut_scale = 1;
	lut_size = 256;
	set_gamma_tab(gamma_device,0);
	load_pix(3);
	annotate_loc = 1;
	annotate_refresh();
	display_get_image();
	annotate_loc = 0;
}

void
gamma_restore(void)
{
	display_release_image();
	lut_scale = lut_scale_save;
	lut_size = lut_size_save;
	set_gamma_tab(gamma_display,1);
	load_pix(0);
}

void
do_awl(void)
{
	CPIXEL *mip,pmin,pmax;
	int i,j;

	mip = image16 + da_y*canvas_size_x + da_x;
	pmin = pmax = *mip;
	for(i=0;i<da_yw;i++){
		for(j=0;j<da_xw;j++){
			if(*mip < pmin) pmin = *mip;
			if(*mip > pmax) pmax = *mip;
			mip++;
		}
		mip -= da_xw - canvas_size_x;
	}
	fprintf(ttyout,"min: %d, max: %d\n",
		pmin-zero_offset,pmax-zero_offset);
	window = pmax - pmin;
	level = -zero_offset + (int) (pmax + pmin) / 2;

	window_level_lut(window,level);
	display_set_sliders();
}

static void
wl_syntax(void)
{

	fputs("Synopsis: sets the window and level\n",ttyout);
	min_abbreviation();
	fputs("Related: window, level, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("wl win lev        * sets window to win and level to lev\n",ttyout);
	fputs("wl -r [min max]   * pixels between min & max displayed in red\n",ttyout);
	fputs("wl -b [min max]   * pixels between min & max displayed in blue\n",ttyout);
	fputs("wl -g [min max]   * pixels between min & max displayed in green\n",ttyout);
	fputs("wl -o             * outline pixels displayed in color\n",ttyout);
	fputs("wl -a             * automatically set window/level\n",ttyout);
	fputs("wl -f             * max pixels are set to black (fallback mode)\n",ttyout);
	fputs("wl -n             * normal display; no color and no black\n",ttyout);
}

void
cmd_wl(void)
{
	int synerr=0,awl=0;
	int i;

	if(argc == 2){
		synerr = 0;
		if(strcmp(argv[1],"-a") == 0){
			awl = 1;
		}else if(strcmp(argv[1],"-f") == 0){
			fall = 1;
		}else if(strcmp(argv[1],"-n") == 0){
			range_red = 0;
			range_blue = 0;
			range_green = 0;
			range_outline = 0;
			fall = 0;
			rvideo = 0;
		}else if(strcmp(argv[1],"-r") == 0){
			if(range_red == 0){
				range_red = 1;
				fprintf(ttyout,"Red=(%d,%d)\n",red_low,red_high);
			}else{
				range_red = 0;
				fprintf(ttyout,"Red=Off\n");
			}
		}else if(strcmp(argv[1],"-b") == 0){
			if(range_blue == 0){
				range_blue = 1;
				fprintf(ttyout,"Blue=(%d,%d)\n",blue_low,blue_high);
			}else{
				range_blue = 0;
				fprintf(ttyout,"Blue=Off\n");
			}
		}else if(strcmp(argv[1],"-g") == 0){
			if(range_green == 0){
				range_green = 1;
				fprintf(ttyout,"Green=(%d,%d)\n",green_low,green_high);
			}else{
				range_green = 0;
				fprintf(ttyout,"Green=Off\n");
			}
		}else if(strcmp(argv[1],"-o") == 0){
			if(range_outline == 0){
				range_outline = 1;
				fprintf(ttyout,"Ouline=(%d,%d)\n",outline_start,outline_start+OUTLINE_LENGTH-1);
			}else{
				range_outline = 0;
				fprintf(ttyout,"Outline=Off\n");
			}
		}else{
			synerr = 1;
		}
	}else if(argc == 3){
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		window = i;

		i = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		level = i;
	}else if(argc == 4){
		if(strcmp(argv[1],"-r") == 0){
			i = parse_atof(argv[2]);
			if(parse_atof_error()) return;
			red_low = i;

			i = parse_atof(argv[3]);
			if(parse_atof_error()) return;
			red_high = i;
			synerr = 0;
			range_red= 1;
		}else if(strcmp(argv[1],"-b") == 0){
			i = parse_atof(argv[2]);
			if(parse_atof_error()) return;
			blue_low = i;

			i = parse_atof(argv[3]);
			if(parse_atof_error()) return;
			blue_high = i;
			synerr = 0;
			range_blue= 1;
		}else if(strcmp(argv[1],"-g") == 0){
			i = parse_atof(argv[2]);
			if(parse_atof_error()) return;
			green_low = i;

			i = parse_atof(argv[3]);
			if(parse_atof_error()) return;
			green_high = i;
			synerr = 0;
			range_green= 1;
		}else{
			synerr = 1;
		}
	}else{
		synerr = 1;
	}

	if(synerr){
		wl_syntax();
		return;
	}

	if(awl){
		do_awl();
	}else{
		window_level_lut(window,level);
		display_set_sliders();
	}
	load_pix(0);
}

static void
refresh_syntax(void)
{
	fputs("Synopsis: refreshes display\n",ttyout);
	min_abbreviation();
	fputs("Related: clear, reset\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("refresh          * refreshes display\n",ttyout);
}


void
cmd_refresh(void)
{
	if(argc <= 0){
		refresh_syntax();
		return;
	}

	window_level_lut(window,level);
	display_set_sliders();
	load_pix(0);
}

static void
level_syntax(void)
{

	fputs("Synopsis: sets the level as in the window/level\n",ttyout);
	min_abbreviation();
	fputs("Related: wl, level, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("level #              * sets level to #\n",ttyout);
	fputs("level +#             * increases level by #\n",ttyout);
	fputs("level --#            * decreases level by #\n",ttyout);
}

void
cmd_level(void)
{
	int i;
	if(argc != 2){
		level_syntax();
		return;
	}

	if(argv[1][0] == '+'){
		i = parse_atof(argv[1]+1);
		if(parse_atof_error()) return;
		level += i;
	}else if(argv[1][0] == '-' && argv[1][1] == '-'){
		i = parse_atof(argv[1]+2);
		if(parse_atof_error()) return;
		level -= i;
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		level = i;
	}

	window_level_lut(window,level);
	display_set_sliders();
	load_pix(0);
}

static void
window_syntax(void)
{

	fputs("Synopsis: sets the window as in the window/level\n",ttyout);
	min_abbreviation();
	fputs("Related: wl, level, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("window #              * sets window to #\n",ttyout);
	fputs("window +#             * increases window by #\n",ttyout);
	fputs("window -#             * decreases window by #\n",ttyout);
}

void
cmd_window(void)
{
	int i;

	if(argc != 2){
		window_syntax();
		return;
	}

	if(argv[1][0] == '+'){
		i =  parse_atof(argv[1]+1);
		if(parse_atof_error()) return;
		window += i;
	}else if(argv[1][0] == '-'){
		i =  parse_atof(argv[1]+1);
		if(parse_atof_error()) return;
		window -= i;
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		window = i;
	}

	window_level_lut(window,level);
	display_set_sliders();
	load_pix(0);
}

static void
gamma_syntax(void)
{
	fputs("Synopsis: set gamma (nonlinear) mapping\n",ttyout);
	min_abbreviation();
	fputs("Related: wl, window, level, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("gamma #        * set gamma to # \n",ttyout);
	fputs("gamma          * print the current value of gamma\n",ttyout);
}

void
cmd_gamma(void)
{
	float g;

	if(argc < 1){
		gamma_syntax();
		return;
	}
	if(argc == 1){
		fprintf(ttyout,"%f\n",gamma_display);
	}else{
		g = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		if(g <=0){
			fputs("gamma must be positive\n",ttyout);
			return;
		}
		gamma_display = g;
		set_gamma_tab(gamma_display,1);
		load_pix(0);
	}
}

static void
reverse_syntax(void)
{
	fputs("Synopsis: manages reverse video mode\n",ttyout);
	min_abbreviation();
	fputs("Related: wl, window, level, inverse, scale\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("reverse on            * enable reverse video\n",ttyout);
	fputs("reverse off           * disables reverse video\n",ttyout);
	fputs("reverse               * change from on/off to off/on\n",ttyout);
}

void
cmd_reverse(void)
{
	if(argc < 1 || argc > 2){
		reverse_syntax();
		return;
	}

	if(argc == 1){
		rvideo = !rvideo;
	}else{
		if(!strcmp("on",argv[1])){
			rvideo = 1;
		}else if(!strcmp("off",argv[1])){
			rvideo = 0;
		}else{
			fputs("On or off please!\n",ttyout);
			return;
		}
	}
	fprintf(ttyout,"reverse video: %s\nn",rvideo? "on" : "off");
	window_level_lut(window,level);
	load_pix(0);
}

void
lut_init(void)
{
	mi_max = (1<<image_bits) - 1;
	mi_lut = (unsigned int *) malloc((mi_max+1)*sizeof(DPIXEL32));

	if(mi_lut == NULL){
		fprintf(stderr,"xpic: can't malloc lut memory\n");
		exit(1);
	}
}

void
wl_status(void)
{
	fprintf(ttyout,"W/L: %d/%d",window,level);
	if(fall) fprintf(ttyout,"/F");
	if(rvideo) fprintf(ttyout,"/R");
	if(range_red) fprintf(ttyout,"/Red=(%d,%d)",red_low,red_high);
	if(range_blue) fprintf(ttyout,"/Blue=(%d,%d)",blue_low,blue_high);
	if(range_green) fprintf(ttyout,"/Green=(%d,%d)",green_low,green_high);
	if(range_outline) fprintf(ttyout,"/Outline");
}
