/*
mag - magnify 2x using bi-cubic iterpolation or pixel replication
*/

#ident "$Id: mag.c,v 1.8 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

#define SMALL 10	/* minimun image size; need intrinsic padding */

			/* inter - cubic interpolator */
static int
inter(a,b,c,d)
CPIXEL a,b,c,d;
{
	int p;

	p = .5625*(a + b)-.0625*(c + d)+ .5;
	return((p<0)? 0 : p);
}

static void
magnify_syntax(void)
{
	fputs("Synopsis: magnify images; default=bi-cubic interpolation\n",ttyout);
	min_abbreviation();
	fputs("Related: set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("magnify     * magnify about image center\n",ttyout);
	fputs("magnify -c  * magnify about deposited cursor location\n",ttyout);
	fputs("magnify -x  * magnify in x only\n",ttyout);
	fputs("magnify -y  * magnify in y only\n",ttyout);
	fputs("            * -c can be used with -x or -y\n",ttyout);
	fputs("            * use 'set replicate' for pixel replication\n",ttyout);
}

void
do_mag(int cursor,int x_only,int y_only)
{
	int	x_stride, y_stride;
	int row,nrow,ncol,nrowp,ncolp,ncolpp;
	int loop,lin,col;
	CPIXEL *ip,*in,*op,*out,*outm1,*outm3,*outp1,*out0,*out2,*out1,*outp3;
	int cenx,ceny,ix,iy;
/*
	check for small images
*/
	x_stride = 2 - y_only;
	y_stride = 2 - x_only;

	ncol = da_xw/x_stride;
	nrow = da_yw/y_stride;

	if(ncol <= SMALL){
		parse_upri(ttyout,"insufficient columns in subcanvas:",da_xw);
		return;
	}
	if(nrow <= SMALL){
		parse_upri(ttyout,"insufficient rows in subcanvas:",da_yw);
		return;
	}


/*
	set up output picture
*/
	ncolp = ncol * x_stride;
	nrowp = nrow * y_stride;

	ncolpp = canvas_size_x;
	op = image16 + da_y*ncolpp + da_x;
/*
	extract subcanvas 
*/
	if(cursor == 0){
		cenx =  da_x + (da_xw - ncol)/2;
		ceny =  da_y + (da_yw - nrow)/2;

	}else{
		cenx = box.lx;
		ceny = box.ly;
		if(cenx < da_x || cenx > da_x + da_xw ||
		  ceny < da_y || ceny > da_y + da_yw){
			fputs("cursor outside of subcanvas\n",ttyout);
			return;
		}
		cenx -= ncol/2;
		ceny -= nrow/2; 
	}


	ip = (CPIXEL *) parse_malloc(sizeof(CPIXEL)*ncol*nrow);
	out = ip;

	for(row=0;row<nrow;row++){
		iy = ceny + row;
		ix = cenx;
		in = image16 + (da_y+row+(da_yw-nrow)/2)*ncolpp + da_x +(da_xw-ncol)/2;
		if(iy >= 0 && iy < canvas_size_y){
			for(col=0;col<ncol;col++){
				if(ix >= 0 && ix < canvas_size_x){
					*out++ = image16[iy*ncolpp + ix];
				}else{
					*out++ = 0;
				}
				ix++;
			}
		}else{
			for(col=0;col<ncol;col++) *out++ = 0;
		}
	}


#ifdef DEBUG
	parse_wtext("xpic: mag: subcanvas written to mag.mi");
	parse_wmi("mag",ip,ncol,nrow);
#endif
/*
	do it
*/
	if(mag_replicate == 0){			/* bi-cubic interpolation */

				/* magnify columns */
		for(lin=0;lin<nrow;lin++){
			in = ip + ncol*lin;
			out = lin*ncolpp*y_stride + op;

						/* replicate even columns */
			for(col=0;col<ncol;col ++) out[col*x_stride]=in[col];

			if(!y_only){
							/* set up boundaries */
				out[1] = out[0];
				out[ncolp-3] = out[ncolp-1] = out[ncolp-2];

							/* interpolate odd columns */
				for(col=3,loop=ncolp-3;col<loop;col += 2){
					out[col] = inter(out[col-1],out[col+1],
						out[col-3],out[col+3]);
				}
			}
		}

		if(!x_only){
					/* magnify rows */
			out0 = 0*ncolpp + op;
			out1 = 1*ncolpp + op;
							/* set up top boundary */
			for(col=0;col<ncolp;col++){
				out1[col] = out0[col];
			}

							/* set up bottom boundary */
			out0 = (nrowp-2)*ncolpp + op;
			out1 = (nrowp-1)*ncolpp + op;
			out2 = (nrowp-3)*ncolpp + op;
			for(col=0;col<ncolp;col++){
				out2[col] = out1[col] = out0[col];
			}

							/* interpolate odd rows */
			for(lin=3,loop=nrowp-3;lin<loop;lin += 2){
				out = lin*ncolpp + op;
				outm1 = (lin-1)*ncolpp + op;
				outm3 = (lin-3)*ncolpp + op;
				outp1 = (lin+1)*ncolpp + op;
				outp3 = (lin+3)*ncolpp + op;
				for(col=0;col<ncolp;col++){
					out[col]= inter(outm1[col],outp1[col],
						outm3[col],outp3[col]);
				}
			}
		}
	}else{				/* pixel replication */
		for(row=0;row<nrow;row++){
			for(loop=0;loop<y_stride;loop++){
				out = op + (row*y_stride + loop)*ncolpp;
				in = ip + row*ncol;
				for(col=0;col<ncol;col++){
					if(!y_only) *out++ = *in;
					*out++ = *in++;
				}
			}
		}
	}
/*
	clean up
*/

	parse_free(ip);

}

void
cmd_magnify(void)
{
	int cursor=0;
	int	x_only = 0;
	int	y_only = 0;
	int	i;
/*
	check command syntax
*/
	if(argc < 1 || argc > 3 ){
		magnify_syntax();
		return;
	}
	for(i=1;i<argc;i++) {
		if(!strcmp(argv[i],"-c")){
			cursor = 1;
		} else if(!strcmp(argv[i],"-x")){
			x_only = 1;
		} else if(!strcmp(argv[i],"-y")){
			y_only = 1;
		}else{
			magnify_syntax();
			return;
		}
	}
	if(x_only && y_only) {
		fputs("mag: can't set both -x and -y\n",ttyout);
		return;
	}
	do_mag(cursor,x_only,y_only);
	load_pix(1);
}
