/*
main - handle command line and create windows
*/

#ident "$Id: main.c,v 1.23 2008/02/10 05:37:40 carl Exp $"

#include "xpic.h"

#ifndef TEST_HELP

#define MAX_GARGC 100		/* max # of args in file, env, command line */
static int colors=0;		/* 1=colors in display */
static int gargc;
static char *gargv[MAX_GARGC];
static int port_num;		/* number for port communication */
static int zero_offset_change;	/*1= change (to zero) zero_offset */
/* static char *first_image; */
static char *title=NULL;	/* title of xpic frame */

static void gamma_arg(char *op,char *arg);
static void size_arg(char *op,char *arg);
void parse(int cargc,char **cargv);
static void eq_proc(char *option,char *arg);

static char** cmd_line_open_argv = NULL;
static long   cmd_line_open_size = 0;
static int    cmd_line_open_argc = 0;

struct PARSE_STRUCT parse_table[] = {
	{"gamma",PARSE_PROC,(PARSE_ARG)gamma_arg,"display gamma correction"},
	{"size",PARSE_INT,(PARSE_ARG)&canvas_size_x,"x- & y-size of canvas"},
	{"size",PARSE_PROC,(PARSE_ARG)size_arg},
	{"xw",PARSE_INT,(PARSE_ARG)&canvas_size_x,"x-size of canvas"},
	{"yw",PARSE_INT,(PARSE_ARG)&canvas_size_y,"y-size of canvas"},
	{"xp",PARSE_INT,(PARSE_ARG)&xp,"display location of x-upper left"},
	{"yp",PARSE_INT,(PARSE_ARG)&yp,"display location of y-upper left"},
	{"rows",PARSE_INT,(PARSE_ARG)&rows,"rows in tty window"},
	{"window",PARSE_INT,(PARSE_ARG)&window,"initial window setting"},
	{"level",PARSE_INT,(PARSE_ARG)&level,"initial level setting"},
	{"bits",PARSE_INT,(PARSE_ARG)&image_bits,"valid bits in pixel"},
	{"port",PARSE_INT,(PARSE_ARG)&port_num,"IPC port address"},
	{"port",PARSE_SET,(PARSE_ARG)&port_access},
	{"printer",PARSE_STRING,(PARSE_ARG)half_printer,"name of printer"},
	{"title",PARSE_CHAR,(PARSE_ARG)&title,"title for xpic frame"},
	{"=",PARSE_PROC,(PARSE_ARG)eq_proc,"set files",},
	{"zero",PARSE_INT,(PARSE_ARG)&zero_offset,"pixel value set to zero"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'a',&no_attach,PARSE_FLAG_CLR,"allow xpic_attach"},
	{'c',&cursor_stat,PARSE_FLAG_CLR,"start w/o cursor"},
	{'C',&colors,PARSE_FLAG_SET,"pseudo color"},
	{'F',&do_fork,PARSE_FLAG_CLR,"don't fork from terminal"},
	{'l',&itype,PARSE_FLAG_VALUE(I_LM),"LM file format"},
#ifdef GZMI
	{'Z',&itype,PARSE_FLAG_VALUE(I_GZMI),"GZMI file format"},
#endif
	{'k',&kludge,PARSE_FLAG_SET,"kludge mode"},
	{'o',&opt_panel,PARSE_FLAG_CLR,"no options panel"},
	{'p',&port_access,PARSE_FLAG_SET,"allow port access"},
	{'s',&sep_ctr,PARSE_FLAG_SET,"separate controls from baseframe"},
	{'z',&zero_offset_change,PARSE_FLAG_SET,"zero image offset"},
	{'S',&mitshm_override,PARSE_FLAG_SET,"do not use MIT shared memory"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

void
parse(int cargc,char **cargv)
{
	int i;

	for(i=1;i<cargc;i++){
		if(gargc == MAX_GARGC) 
			parse_err("too many command line arguments",NULL);

		gargv[gargc++] = cargv[i];
	}
}

static void
size_arg(char *op,char *arg)
{
	canvas_size_y = canvas_size_x;
}

static void
gamma_arg(char *op,char *arg)
{
	char *cp;

	if(NULL != (cp = strchr(arg,'/'))){
		*cp = 0;
		gamma_insert(cp+1,parse_atof(arg));
	}else{
		gamma_insert(NULL,parse_atof(arg));
	}
}

int
main(cargc,cargv)
int	cargc;
char	**cargv;
{
	char *s,*t;
	int init,i,pid;
/*
	set up global variables
*/
	if((s = getenv("PRINTER")) != NULL) strcpy(half_printer,s); 
	strncpy(host_name,parse_host(),MAXHOSTNAME);
/*
	check if in keyboard mode
*/
	if(cargc > 1 && !strcmp(cargv[1],MAIN_KB))kb(cargc,cargv);
/*
	do some other setup before parsing command line 
*/
	port_num = port_get_port();
/*
	parse command line, files and environement
*/
	gargc = 1;
	gargv[0] = cargv[0];

					/* _file and _args stuff their args 
					into gargv because of parse() */
	for(init=i=1;i<cargc;i++){
		if(cargv[i][0] == '-' && (s = strchr(cargv[i],'@')) != NULL){
			init = 0;
			*s = ' ';
		}
	}

	parse_file_default(START_UP_FILE);
	parse_arg_default("XPICARGS");

	if(init){
		parse_file(START_UP_FILE);
		parse_args("XPICARGS");
	}

					/* copy command line into gargv */
	for(i=1;i<cargc;i++){
		if(gargc == MAX_GARGC) 
			parse_err("too many command line arguments",NULL);

		gargv[gargc++] = cargv[i];
	}

					/* xv_init takes what it needs */
#ifndef NOXVINIT
	display_attach(&gargc,gargv);
#endif

					/* now parse what is left */
	parse_basic(gargc,gargv);
/*
	undo some parsing
*/
	if(zero_offset_change) zero_offset = 0;
/*
	check for strange operation
*/
	if(canvas_size_x < 256 || canvas_size_y < 256){
	   fprintf(stderr,"Canvas can't be smaller than 256x256\n");
	   exit(1);
	}

	if(image_bits < 1 || image_bits > 16){
	   fprintf(stderr,"Only from 2 to 16 bits can be used per pixel\n");
	   exit(1);
	}
/*
	detach process
*/
	if(do_fork){
		if((pid = fork()) == -1){
			perror("Can't fork copy of xpic");
			exit(1);
		}
		if(pid != 0){
			_exit(0);
		}
	}
/*
	set up windowing system
*/
	display_init(cargv[0],init,colors,title);
/*
	initialize other "packages"
*/
	parse_atof_args(ttyout,1);	/* set error control for parse_atof */

	lut_init();
	scroll_init();
	env_init();

	annotate_loc = 0;	/* annotate for display */

	reset();		/* paramaters of display */

	set_gamma_tab(gamma_display,1);	
	load_pix(0);
	display_set_sliders();
	if(cursor_stat)cur_on_off();

	file_open = 0;

	box_init();

	port_set_port(port_num);
	if(port_access){
		handle_port();	
		fflush(ttyout);
	}
/*
	handle special configurations
*/
	s = strrchr(cargv[0],'/');
	if(s == NULL){
		t = cargv[0];
	}else{
		t = s+1;
	}

	if(!strcmp(t,"dpic")){
		itype = I_LM;
		isub = I_LM_PROJ;

		box.x = canvas_size_x/4;
		box.y = canvas_size_y/4;
		box.xw = 10;
		box.yw = 10;
		box.lx = 0;
		box.ly = 0;
		cursor_stat = 1;
		box.status = 1;
		cur_on_off();

		cor_set = 1;
		cor_x = 0;
		cor_y = 0;

		zero_offset = 0;
		window = 200;
		level = 0;
		window_level_lut(window,level);
		display_set_sliders();

		autoreset = 0;
	}
/*
	load image from command line
*/
	if(cmd_line_open_argc){ /*if we found open arguments*/
		/* open the files */
		parse_expand_open(cmd_line_open_argc, cmd_line_open_argv);
		/* go to env 1*/
		env_push(current_env);
		env_pop(current_env=0,0); /* dont be verbose */
		write_env_panel(current_env);
		images("1");
		
		/*delete this arglist*/
		for(cmd_line_open_size=0;cmd_line_open_size<cmd_line_open_argc;cmd_line_open_size++){
			free(cmd_line_open_argv[cmd_line_open_size]);
		}
		free(cmd_line_open_argv);
	}

/*
	start up window system
*/
	display_run();
/*
	clean up and exit
*/
	display_clean_up();
	exit(0);
}

static void eq_proc(option,arg)
char *option,*arg;
{
	int i;

	if(!cmd_line_open_argv){ /*this is the first call*/
		/*make a 0th entry*/
		if((cmd_line_open_argv = malloc(cmd_line_open_size=sizeof(char*)))==NULL){
			fprintf(stderr,"Memory Allocation Error\n");
			exit(1);
		}
		if((cmd_line_open_argv[0]=strdup(""))==NULL){
			free(cmd_line_open_argv);
			fprintf(stderr,"Memory Allocation Error\n");
			exit(1);		
		}
		cmd_line_open_argc=1;
	}
	/* add this entry to the list */
	if((cmd_line_open_argv=realloc(cmd_line_open_argv,(cmd_line_open_size+=sizeof(char*))))==NULL){
		for(i=0;i<cmd_line_open_argc;i++)
			free(cmd_line_open_argv[i]);
		free(cmd_line_open_argv);
		fprintf(stderr,"Memory Allocation Error\n");
		exit(1);
	}
	if((cmd_line_open_argv[++cmd_line_open_argc-1]=strdup(arg))==NULL){
		for(i=0;i<cmd_line_open_argc-2;i++)
			free(cmd_line_open_argv[i]);
		free(cmd_line_open_argv);
		fprintf(stderr,"Memory Allocation Error\n");
		exit(1);
	}
}


#endif
