/*
measure - ... distance and angle
*/

#ident "$Id: measure.c,v 1.7 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static int erase=1;		/* 1=erase annotation */
static int cursors=1;		/* 1 = deposit cursors */
static int line=0;		/* 1 = draw line in canvas window */
static int restore=0;		/* 1 = restore measure's parameters */

				/* cursor stack */
static int cx1=-1,cx2=-1,cx3=-1; 		
static int cy1=-1,cy2=-1,cy3=-1;

static void measure_syntax(void);

static struct PARSE_OPTION_TABLE options[] = {
	{"dx",PARSE_FLOAT,(PARSE_ARG)&pixel_dx},
	{"dydx",PARSE_FLOAT,(PARSE_ARG)&pixel_dydx},
	{"?",PARSE_PROC,(PARSE_ARG)measure_syntax},
	{0,0,0,0}
};

static struct PARSE_FLAG_TABLE flags[] = {
	{'c',&cursors,PARSE_FLAG_CLR},
	{'l',&line,PARSE_FLAG_SET},
	{'e',&erase,PARSE_FLAG_CLR},
	{'r',&restore,PARSE_FLAG_SET},
	{'?',(int *)measure_syntax,0},
	{0,0,0,0}
};

void 
push_cursor(int x,int y)
{

#ifdef DEBUG
	printf("cursor: %d %d\n",x,y);
#endif

	cx1 = cx2;
	cx2 = cx3;
	cx3 = x;
	cy1 = cy2;
	cy2 = cy3;
	cy3 = y;
}

static void
measure_syntax(void)
{

	fputs("Synopsis: sets attributes for measure panel button\n",ttyout);
	min_abbreviation();
	fputs("Related: panel, set\n",ttyout);
	fputs("Syntax: measure [options]\n",ttyout);
	fprintf(ttyout,"dx=[%.2f]: x-pixel size\n",pixel_dx);
	fprintf(ttyout,"dydx=[%.2f]: y/x-pixel size ratio\n",pixel_dydx);
	fputs("-l: draw lines\n",ttyout);
	fputs("-c: don't draw cursors\n",ttyout);
	fputs("-e: don't erase previous annotation/lines\n",ttyout);
	fputs("-r: reset command parameters\n",ttyout);
}

void
cmd_measure(void)
{
	if(argc < 0){
		measure_syntax();
		return;
	}

	if(-1 == parse_general(options,flags,ttyout,argc,argv)) return;

	if(restore == 1){
		restore = 0;
		cursors = 0;
		erase = 0;
		line = 0;
		pixel_dx = 1.0;
		pixel_dydx = 1.0;
		fputs("measure default parameters restored\n",ttyout);
	}

	parse_uprt(ttyout,"draw lines",line);
	parse_uprt(ttyout,"draw cursors",cursors);
	parse_uprt(ttyout,"erase annotation",erase);
	parse_uprf(ttyout,"x-pixel size, dx=",pixel_dx);
	parse_uprf(ttyout,"y/x-pixel size ratio, dydx=",pixel_dydx);
}

void 
func_measure(void)
{
	float d;
	float a1,a2;
	int mode; 	// 0=distance 1=angle

	if(cx2 == -1 || cx3 == -1){
		fprintf(ttyout,"Need two deposited cursors\n");
		return;
	}

	if(cx1 == -1){
		mode = 0;
	}else{
		mode = 1;
	}

	if(erase) annotate_clear();

	if(mode == 0){
		fprintf(ttyout,"p1=(%d,%d)",cx2,cy2);
		fprintf(ttyout," p2=(%d,%d)",cx3,cy3);
		if(line){
			line_add(cx2,cy2,0);
			line_add(cx3,cy3,1);
		}
		if(cursors){
			box_add(cx2,cy2,0,0,1);
			box_add(cx3,cy3,0,0,1);
		}

	}else{
		fprintf(ttyout,"p1=(%d,%d)",cx1,cy1);
		fprintf(ttyout," p2=(%d,%d)",cx2,cy2);
		fprintf(ttyout," p3=(%d,%d)",cx3,cy3);
		if(line){
			line_add(cx1,cy1,0);
			line_add(cx2,cy2,1);
			line_add(cx3,cy3,1);
		}
		if(cursors){
			box_add(cx1,cy1,0,0,1);
			box_add(cx2,cy2,0,0,1);
			box_add(cx3,cy3,0,0,1);
		}
	}

	fprintf(ttyout," dx=%.2f dydx=%.2f\n",pixel_dx,pixel_dydx);


	if(mode == 0){
		d = sqrt(pow((cx3-cx2)*pixel_dx,2.0) + pow((cy3-cy2)*pixel_dx*pixel_dydx,2.0));
		fprintf(ttyout,"d12=%.2f\n",d);
	}else{
		d = sqrt(pow((cx1-cx2)*pixel_dx,2.0) + pow((cy1-cy2)*pixel_dx*pixel_dydx,2.0));
		fprintf(ttyout,"d12=%.2f",d);
		d = sqrt(pow((cx3-cx2)*pixel_dx,2.0) + pow((cy3-cy2)*pixel_dx*pixel_dydx,2.0));
		fprintf(ttyout," d23=%.2f",d);

		a1 = atan2((cy2-cy1)*pixel_dydx,cx1-cx2);
		a2 = atan2((cy2-cy3)*pixel_dydx,cx3-cx2);

		a1 = a1 * 180 / M_PI;
		a2 = a2 * 180 / M_PI;
		fprintf(ttyout," angle=%.2f (degrees)\n",a1-a2);
	}

	fflush(ttyout);

	if(erase || line || cursors) load_pix(0);

	cx1=-1;
	cx2=-1;
	cx3=-1; 		
	cy1=-1;
	cy2=-1;
	cy3=-1;
}
