/*
region - set up a subregion 
*/

#ident "$Id: region.c,v 1.7 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static void
hexant_syntax()
{
	fputs("Synopsis: set hexant (one sixteenth) region of sub-canvas\n",ttyout);
	min_abbreviation();
	fputs("Related: quadrant, nanant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("hexant #     * set hexant (1-16)\n",ttyout);
	fputs("hexant +     * set to next (+1) hexant\n",ttyout);
	fputs("hexant -     * set to previous (-1) hexant\n",ttyout);
}

static void
quadrant_syntax()
{
	fputs("Synopsis: set quadrant (one quarter) region of sub-canvas\n",ttyout);
	min_abbreviation();
	fputs("Related: hexant, nanant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("quadrant #     * set quadrant (1-4)\n",ttyout);
	fputs("quadrant +     * set to next (+1) quadrant\n",ttyout);
	fputs("quadrant -     * set to previous (-1) quadrant\n",ttyout);
}

static void
nanant_syntax()
{
	fputs("Synopsis: set nanant (one ninth) region of sub-canvas\n",ttyout);
	min_abbreviation();
	fputs("Related: hexant, quadrant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("nanant #     * set nanant (1-9)\n",ttyout);
	fputs("nanant +     * set to next (+1) nanant\n",ttyout);
	fputs("nanant -     * set to previous (-1) nanant\n",ttyout);
}

void
cmd_hexant(void)
{
	int i;

	if(argc != 2){
		hexant_syntax();
		return;
	}

	if(argv[1][0] == '+'){
		if(last_hex == 16){
			fprintf(ttyout,"Can't bump past hexant 16\n");
			return;
		}
		last_hex++;
	}else if(argv[1][0] == '-'){
		if(last_hex == 1){
			fprintf(ttyout,"Can't bump before hexant 1\n");
			return;
		}
		last_hex--;
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		last_hex = i;

		if(last_hex < 1 || last_hex > 16){
			fputs("Bad Hexant\n",ttyout);
			return;
		}
	}

	da_xw = canvas_size_x / 4 ;
	da_yw = canvas_size_y / 4 ;

	if(last_hex < 5)
		da_y = 0;
	else if(last_hex < 9)
		da_y = da_yw;
	else if(last_hex < 13)
		da_y = 2 * da_yw;
	else
		da_y = 3 * da_yw;

	switch(last_hex & 03){
	case 1:
		da_x = 0;
		break;
	case 2:
		da_x = da_xw;
		break;
	case 3:
		da_x = 2 * da_xw;
		break;
	case 0:
		da_x = 3 * da_xw;
		break;
	}
}

static void
horizontal_syntax(void)
{
	fputs("Synopsis: set horizontal attributes of subregion\n",ttyout);
	min_abbreviation();
	fputs("Related: vertical, quadrant, hexant, nanant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("horizontal offset       * set horizontal offset\n",ttyout);
	fputs("horizontal offset size  * set horizontal offset and size\n",ttyout);
}

void
cmd_horizontal(void)
{
	int o,s;

	if(argc < 2 || argc > 3){
		horizontal_syntax();
		return;
	}

	o = parse_atof(argv[1]);
	if(parse_atof_error()) return;

	if(o < 0 || o >= canvas_size_x){
		fputs("Bad offset\n",ttyout);
		return;
	}
	da_x = o;

	if(argc == 3){
		s = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		if(s <= 0){
			fputs("Bad size\n",ttyout);
			return;
		}
		da_xw = s;
	}

	if(da_x + da_xw > canvas_size_x)da_xw = canvas_size_x - da_x;
}

void
cmd_quadrant(void)
{
	int i;

	if(argc != 2){
		quadrant_syntax();
		return;
	}

	if(argv[1][0] == '+'){
		if(last_quad == 4){
			fprintf(ttyout,"Can't bump past quadrant 4\n");
			return;
		}
		last_quad++;
	}else if(argv[1][0] == '-'){
		if(last_quad == 1){
			fprintf(ttyout,"Can't bump before quadrant 1\n");
			return;
		}
		last_quad--;
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		last_quad = i;

		if(last_quad < 1 || last_quad > 4){
			fputs("Bad Quadrant\n",ttyout);
			return;
		}
	}

	da_xw = canvas_size_x / 2 ;
	da_yw = canvas_size_y / 2 ;

	if( last_quad & 01)
		da_x = 0;
	else
		da_x = da_xw;

	if(last_quad < 3)
		da_y = 0;
	else
		da_y = da_yw;
}

static void
vertical_syntax(void)
{
	fputs("Synopsis: set vertical attributes of subregion\n",ttyout);
	min_abbreviation();
	fputs("Related: horizontal, quadrant, hexant, nanant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("vertical offset       * set vertical offset\n",ttyout);
	fputs("vertical offset size  * set vertical offset and size\n",ttyout);
}

void
cmd_vertical(void)
{
	int o,s;

	if(argc < 2 || argc > 3){
		vertical_syntax();
		return;
	}

	o = parse_atof(argv[1]);
	if(parse_atof_error()) return;

	if(o < 0 || o >= canvas_size_y){
		fputs("Bad offset\n",ttyout);
		return;
	}
	da_y = o;

	if(argc == 3){
		s = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		if(s <= 0){
			fputs("Bad size\n",ttyout);
			return;
		}
		da_yw = s;
	}

	if(da_y + da_yw > canvas_size_y)da_yw = canvas_size_y - da_y;
}

void
cmd_nanant(void)
{
	static int xoff[] = {0,1,2,0,1,2,0,1,2};
	static int yoff[] = {0,0,0,1,1,1,2,2,2};
	int i;

	if(argc != 2){
		nanant_syntax();
		return;
	}

	if(argv[1][0] == '+'){
		if(last_nan == 9){
			fprintf(ttyout,"Can't bump past nanant 9\n");
			return;
		}
		last_nan++;
	}else if(argv[1][0] == '-'){
		if(last_nan == 1){
			fprintf(ttyout,"Can't bump before nanant 1\n");
			return;
		}
		last_nan--;
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		last_nan = i;

		if(last_nan < 1 || last_nan > 9){
			fputs("Bad nanant\n",ttyout);
			return;
		}
	}

	da_xw = canvas_size_x / 3 ;
	da_yw = canvas_size_y / 3 ;

	da_x = xoff[last_nan-1] * da_xw;
	da_y = yoff[last_nan-1] * da_yw;
}

static void
copyquad_syntax(void)
{
	fputs("Synopsis: copy quadrants\n",ttyout);
	min_abbreviation();
	fputs("Related: quadrant\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("copyquad s d         * copy from s to d. (s,d)=1-4\n",ttyout);
}

void
cmd_copyquad(void)
{
	int i,j,xs,ys,xd,yd,xw,yw,s,d;
	CPIXEL *sp,*dp;

	if(argc != 3){
		copyquad_syntax();
		return;
	}

	s = parse_atof(argv[1]);
	if(parse_atof_error()) return;
	d = parse_atof(argv[2]);
	if(parse_atof_error()) return;

	if(s < 1 || s > 4 || d < 1 || d > 4){
		fputs("Bad Quadrant\n",ttyout);
		return;
	}

	if(s == d) return;

	xw = canvas_size_x / 2 ;
	yw = canvas_size_y / 2 ;

	if(s & 01)
		xs = 0;
	else
		xs = xw;

	if(s < 3)
		ys = 0;
	else
		ys = yw;

	if(d & 01)
		xd = 0;
	else
		xd = xw;

	if(d < 3)
		yd = 0;
	else
		yd = yw;

	for(i=0;i<yw;i++){
		sp = (i+ys)*canvas_size_x + xs + image16;
		dp = (i+yd)*canvas_size_x + xd + image16;
		for(j=0;j<xw;j++){
			*dp++ = *sp++;
		}
	}

	da_xw = xw;
	da_yw = yw;
	da_x = xd;
	da_y = yd;

	load_pix(1);
}

