/*
tiff - write header of tiff file
*/

#ident "$Id: tiff.c,v 1.5 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"


/* Data type */
typedef unsigned char     UChar8;
typedef unsigned short    UShort16;
typedef unsigned int      UInt32;

#define assert(x) if(!(x)) return(-1);

/* Tiff tag-type partial list, name-value pairs */
/* Ref. Graphics File Formats by J. Murry and W. vanRyper */
#define NewSubfileType          254
#define SubfileType             255
#define ImageWidth              256
#define ImageLength             257
#define BitsPerSample           258
#define Compression             259
#define PhotometricInter        262
#define Thresholding            263
#define CellWidth               264
#define CellLength              265
#define ImageDescription        270
#define Make                    271
#define Model                   272
#define StripOffsets            273
#define Orientation             274
#define SamplesPerPixel         277
#define RowsPerStrip            278
#define StripByteCounts         279
#define XResolution             282
#define YResolution             283
#define ResolutionUnit          296

/* Tiff tag scalar data types */
#define BYTE		1
#define ASCII		2
#define SHORT		3
#define LONG		4
#define RATIONAL	5

/* Tiff tag structure */
struct  {
	UShort16		TagId;
	UShort16		DataType;
	UInt32			DataCount;
	union  {
		UShort16	sval;
		UInt32		ival;
		UInt32		DataOffset;
	} val;
} TIFTAG;


int 
write_tiff_head(int height, int width, FILE *fp,char *ofn) 
{
    int			n;
    short		x;
    char		*p;
    UChar8		ui8;
    UShort16 	        ui16;
    UInt32		ui32;

    /* determine the byte order */
    x = 1;
    p = (char *) &x;
    if (p[0] == 1 && p[1] == 0){
		ui8 = 'I'; /* I - 49h, little-endian */	  
    }else if (p[0] == 0 && p[1] == 1){
		ui8 = 'M'; /* M - 4Dh, big-endian */	  
    }else{                              
		return(-1);
    }

    /* Write Tiff Identifier, 49h 49h or 4Dh 4Dh */
    n = fwrite(&ui8, sizeof ui8, 1, fp);                   
	assert(n == 1);
    n = fwrite(&ui8, sizeof ui8, 1, fp);                   
	assert(n == 1);
 
    /* Write Tiff Version number (42), 2Ah 00h or 00h 2Ah */
    ui16 = 42;
    n = fwrite(&ui16, sizeof ui16, 1, fp);
	assert(n == 1);

    /* Offset to the first Image File Directory */
    ui32 = sizeof ui16 +
           sizeof ui16 +
           sizeof ui32;  
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);

  /** Write image file directory entries */

    /* Total number of directory entries */
    ui16 = 11;  
    n = fwrite(&ui16, sizeof ui16, 1, fp);
	assert(n == 1);
 
    /* Chech size of TIFTAG, 16+16+32+32 bits */
    assert(sizeof(TIFTAG) == 
	sizeof(UShort16) + sizeof(UShort16) + sizeof(UInt32) + sizeof(UInt32));
 
    /* Format and write each tiff tag */
    TIFTAG.TagId	= ImageWidth;
    TIFTAG.DataType     = LONG;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= width;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= ImageLength;
    TIFTAG.DataType     = LONG;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= height;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= BitsPerSample;
    TIFTAG.DataType     = SHORT;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= 0;  /* set the whole union to zero */
    TIFTAG.val.sval	= 8;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= Compression;
    TIFTAG.DataType     = SHORT;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= 0;  
    TIFTAG.val.sval	= 1;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= PhotometricInter;
    TIFTAG.DataType     = SHORT;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= 0;
    TIFTAG.val.sval	= 1;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= StripOffsets;
    TIFTAG.DataType     = LONG;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= sizeof ui16 + sizeof ui16 +
			sizeof ui32 + sizeof ui16 + 11 * sizeof TIFTAG +
			sizeof ui32 + 2 * sizeof ui32 + 2 * sizeof ui32;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= RowsPerStrip;
    TIFTAG.DataType     = LONG;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= height;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= StripByteCounts;
    TIFTAG.DataType     = LONG;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.ival	= width * height;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= XResolution;
    TIFTAG.DataType     = RATIONAL;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.DataOffset = sizeof ui16 + sizeof ui16 +
			sizeof ui32 + sizeof ui16 +
			11 * sizeof TIFTAG + sizeof ui32;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
 
    TIFTAG.TagId	= YResolution;
    TIFTAG.DataType     = RATIONAL;
    TIFTAG.DataCount    = 1;
    TIFTAG.val.DataOffset = sizeof ui16 + sizeof ui16 +
			sizeof ui32 + sizeof ui16 + 11 * sizeof TIFTAG +
			sizeof ui32 + 2 * sizeof ui32;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp); 
	assert(n == 1);
 
    TIFTAG.TagId	= ResolutionUnit;
    TIFTAG.DataType	= SHORT;
    TIFTAG.DataCount	= 1;
    TIFTAG.val.ival	= 0; 
    TIFTAG.val.sval	= 1;
    n = fwrite(&TIFTAG, sizeof TIFTAG, 1, fp);
	assert(n == 1);
    
    /* Set offset to the next image file directory */
    ui32 = 0;
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);

    /* x res numerator */
    ui32 = 1;
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);
    /* x res denominator */
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);
    /* y res numerator */
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);
    /* y res denominator */
    n = fwrite(&ui32, sizeof ui32, 1, fp);
	assert(n == 1);
 
    /* write image raw pixel data */
    /*
    n = fwrite(buff, height * width * sizeof(*buff), 1, fp);
    assert(n == 1);
    */

    return(0);
}
