/*
wmi - write mi-image
*/

#ident "$Id: wmi.c,v 1.9 2011/09/14 11:23:38 carl Exp $"

#include "xpic.h"

static void 
werr(char *ofn)
{
	fprintf(ttyout,"Error writing to: %s\n",ofn);
}

static void 
write_syntax(void)
{
	fputs("Synopsis: writes display to output file\n",ttyout);
	min_abbreviation();
	fputs("Related: halftone\n",ttyout);
	fputs("Syntax: write [!]file[.mi] [-{s8t}]\n",ttyout);
	fputs("  file[.mi]  * write 16-bits (no graphics) to file\n",ttyout);
	fputs("  !file      * allow overwrite of file\n",ttyout);
	fputs("  -s         * include graphics; write mi-file\n",ttyout);
	fputs("  -8         * include graphics; write 8-bit raw file\n",ttyout);
	fputs("  -t         * include graphics; write tiff image\n",ttyout);
}

void
cmd_write()
{
	int n;			/* counter for block alignment */
	int over;		/* allow overwrite of files */
	int i,j;		/* loopers */
	FILE *fd;		/* for output file */
	char ofn[BUFL];		/* output file name */
	short int head[256];	/* headers for mi-file */
	struct stat sbuf;	/* for stat() call */
	DPIXEL8 *image;		/* pointer to 16-bit buffer */
	CPIXEL *line;		/* pointer to line of 16-bit buffer */
	CPIXEL *ip;			/* pointer into image */
	char *cline;		/* line of 8-bit output */

	enum {			/* content of output */
		BUF16,		/* 16-bit canvas */
		BUF8		/* 8-bit display buffer */
	} content = BUF16;

	enum {			/* output file type */
		MIFILE,		/* mi-file */
		RAW8,		/* raw, 8-bit, flat, no-header */
		TIFF		/* TIFF file */
	} otype = MIFILE;

	if(argc <= 1 || argc > 3){
		write_syntax();
		return;
	}

	strcpy(ofn,argv[1]);
	if(ofn[0] == '!'){
		over = 1;
		strcpy(ofn,ofn+1);
	}else{
		over = 0;
	}

	if(argc == 3){
		if(!strcmp(argv[2],"-s")){
			content = BUF8;
			otype = MIFILE;
		}else if(!strcmp(argv[2],"-8")){
			content = BUF8;
			otype = RAW8;
		}else if(!strcmp(argv[2],"-t")){
			content = BUF8;
			otype = TIFF;
		}else{
			write_syntax();
			return;
		}
	}

	if(otype == MIFILE) parse_suffix(ofn,"mi");
	if(otype == TIFF) parse_suffix(ofn,"tif");

	if(!stat(ofn,&sbuf)){
		if(!over){
			fputs("File exists. Use 'write !file' to overwrite\n",ttyout);
			return;
		}
		if(-1 == unlink(ofn)){
			fprintf(ttyout,"Can't unlink: %s\n",ofn);
			return;
		}
	}

	if((fd=fopen(ofn,"w")) == NULL){
		fprintf(ttyout,"Can't open: %s\n",ofn);
		return;
	}

	line = (CPIXEL *) malloc(canvas_size_x * sizeof(CPIXEL));
	cline = malloc(canvas_size_x);
	if(line == NULL || cline == NULL){
		fprintf(ttyout,"Can't malloc lines\n");
		return;
	}

	if(otype == MIFILE){
		parse_zbuf(head,256);
		head[PARSE_MI_LOC_MAGIC] = PARSE_MI_MAGIC_65K;
		head[PARSE_MI_LOC_N] = 1;
		head[PARSE_MI_LOC_FIRST] = 1;
		parse_to_can_shortv(head,256);
		if(256 != fwrite(head,sizeof(short int),256,fd)){
			werr(ofn);
			goto out;
		}  

		parse_zbuf(head,256);
		head[PARSE_MI_LOC_NCOL] = canvas_size_x;
		head[PARSE_MI_LOC_NROW] = canvas_size_y;
		parse_to_can_shortv(head,256);
		if(256 != fwrite(head,sizeof(short int),256,fd)){
			werr(ofn);
			goto out;
		}  
	}
	if(otype == TIFF){
		if(-1 == write_tiff_head(canvas_size_y,canvas_size_x,fd,ofn)){
			goto out;
		}
	}

	if(content == BUF16){		/* write from 16-bit buffer */
		ip = image16;
		for(j=0;j<canvas_size_y;j++){
			for(i=0;i<canvas_size_x;i++) line[i] = *ip++;
			parse_to_can_shortv((short int *) line,canvas_size_x);
			if(canvas_size_x != fwrite(line,sizeof(CPIXEL),canvas_size_x,fd)){
				werr(ofn);
				goto out;
			}  
		}
	}else if(content == BUF8){		/* write from 8-bit buffer */
		gamma_set_device(1.0);
		image = pr2_data;
		for(j=0;j<canvas_size_y;j++){
			if(otype == MIFILE){		
				for(i=0;i<canvas_size_x;i++){
					line[i] = image[dsp_bytes_pixel*i] + 896;
				}
				parse_to_can_shortv((short int *) line,canvas_size_x);
				if(canvas_size_x != fwrite(line,sizeof(CPIXEL),canvas_size_x,fd)){
					werr(ofn);
					goto out;
				}  
			}else if(otype == RAW8 || otype == TIFF){		
				for(i=0;i<canvas_size_x;i++){
					cline[i] = image[dsp_bytes_pixel*i];
				}
				if(canvas_size_x != fwrite(cline,1,canvas_size_x,fd)){
					werr(ofn);
					goto out;
				}  
			}
			image += pr2_pixels;
		}
		gamma_restore();
	}else{
		fprintf(ttyout,"Internal error: writing illegal content\n");
		goto out;
	}

	if(otype == MIFILE){		/* pad mi-file to block boundary */
		n = (canvas_size_x * canvas_size_y * sizeof(CPIXEL)) % 512;
		if(n){
			n = 512 - n;
			if(n != fwrite(head,1,n,fd)){
				werr(ofn);
				goto out;
			}  
		}
	}

	switch (content){

	case BUF16:
		fprintf(ttyout,"Canvas");
		break;
	case BUF8:
		fprintf(ttyout,"Display");
		break;
	}

	fprintf(ttyout," written to ");

	switch (otype){

	case MIFILE:
		fprintf(ttyout,"mi");
		break;
	case RAW8:
		fprintf(ttyout,"8-bit-raw");
		break;
	case TIFF:
		fprintf(ttyout,"tiff");
		break;
	}

	fprintf(ttyout,"-file: %s\n",ofn);

			/* clean up */
out:
	free(line);
	free(cline);
	if(EOF == fclose(fd)){
		fprintf(ttyout,"Error closing: %s\n",ofn);
	}
}
