/*
notify - handle socket and file communication
*/

#ident "$Id: notify.c,v 1.6 2011/09/14 10:09:37 carl Exp $"

#define OWTOOLKIT_WARNING_DISABLED
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <errno.h>
#include <xview/xview.h>
#include <xview/canvas.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/utsname.h>
#include "crcplot_config.h"
#include "xplot.h"
#include "parse.h"

#ifdef DEBUG
#define MESSAGE(x) puts(x)
#else
#define MESSAGE(x)
#endif	

FILE *plot_file;
extern int errno;

static int socket_fd,socket_handle;

static int id1,id2;
static int *me_fd = &id1;
static int *me_sock = &id2;

	/* need these because of recursive calling */
static Notify_value pipe_open(int *me,int fd);
static Notify_value read_pipe(int *me,int fd);

static int 
get_connection(s)
  int s;                    /* socket created with establish() */
  { 
    struct sockaddr_in isa; /* address of socket */
    socklen_t i;                  /* size of address */
    int sock=-1;

    errno = EINTR;
    while (errno == EINTR) {
        i = sizeof(isa);                
    	if ((sock = accept(s,(struct sockaddr *)&isa,&i)) < 0) { 
		if(errno != EINTR) return(-1);
		MESSAGE("EINTR error");
	}else errno = 0;
      }

    if(sock == -1)parse_er("xplot internal error: get_connection: sock");
    return(sock);
}

int 
establish()
{ 
    int    s;
    int n;		/* for setting socket options */
    struct sockaddr_in sa;
    struct hostent *hp;
	static struct utsname buf;

    memset((char *)&sa,0,sizeof(struct sockaddr_in));   /* clear our address */
	uname(&buf);
    hp= gethostbyname(&buf.nodename[0]);                 /* get our address info */
    if (hp == NULL)                             /* we don't exist !? */
      return(-1);
    sa.sin_family= hp->h_addrtype;              /* this is our host address */
    sa.sin_port= htons((u_short)port_num);      /* this is our port number */
    if ((s= socket(AF_INET,SOCK_STREAM,0)) < 0) /* create socket */
      return(-1);
    n = 1;
    if(setsockopt( s, SOL_SOCKET, SO_REUSEADDR,(void *)&n, sizeof(n) )<0)
      return(-1);                               /* bind address to socket */
    if (bind(s,(struct sockaddr *)&sa,sizeof(sa)) < 0)
      return(-1);                               /* bind address to socket */
    return(s);
}

static Notify_value 
read_pipe(me,fd)
int *me;
int fd;
{
	int i;
	static char c[512];

	MESSAGE("read_pipe");

	openpl(&pc_main);
	/*
	do{
	*/
		MESSAGE("read_pipe: before read");
		i = read(fd, c, 512);
#ifdef DEBUG
		printf("read_pipe: after read %d\n",i);
#endif
		if(i > 0){
			driver(&pc_main,c,i,1);
		}else if(i <= 0){
			MESSAGE("read_pipe - closepl");
			closepl(&pc_main);
			if(!pc_main.cmderr)tell_user("Plot completed");
			fflush(plot_file);
			notify_set_input_func((Notify_client)me_fd,NOTIFY_FUNC_NULL,fd);
			c[0] = HANDSHAKE1;
			c[1] = HANDSHAKE2;
			if(-1 == write(fd,&c[0],2)) tell_user("read_pipe: write error");
			shutdown(fd,2);
			close(fd);
			notify_set_input_func((Notify_client)me_sock,pipe_open,socket_handle);
			i = 0;
		}
	/*
	}while (i > 0);
	*/
	
	return(NOTIFY_DONE);
}


static Notify_value pipe_open(me,fd)
int *me;
int fd;
{
	MESSAGE("pipe_open");

	notify_set_input_func((Notify_client)me_sock,NOTIFY_FUNC_NULL,socket_handle);

	driver_reset(&pc_main);

	socket_fd = get_connection(socket_handle);
	tell_user("Accepting plot");
	if(socket_fd == -1){
		tell_user("accept error");
		return(NOTIFY_DONE);
	}

	notify_set_input_func((Notify_client)me_fd,read_pipe,socket_fd);
	return(NOTIFY_DONE);
}

void notify()
{
	MESSAGE("notify");
	socket_handle = establish();
	if(socket_handle == -1){
		perror("establish socket");
		exit(1);
	}
	
	if(listen(socket_handle,1) == -1){
		perror("rec/listen");
		clean_socket();
	}

	if((plot_file = fopen(plot_name,"w")) == NULL){
		fprintf(stderr,"Can't open %s\n",plot_name);
		exit(1);
	}
	chmod(plot_name,0666);

	notify_set_input_func((Notify_client)me_sock,pipe_open,socket_handle);
}


void 
clean_socket()
{
	MESSAGE("clean_socket");
	fclose(plot_file);
	unlink(plot_name);
	exit(1);
}


void 
mouse_send(x,y)
int x,y;
{
	int h[2];
	char h1,h2;
	int err1,err2,err3;

	h[0] = x;
	h[1] = y;

	h1 = CMD_MOUSE1;
	err1 = write(socket_fd,&h1,sizeof(h1));
	err2 = write(socket_fd,&h2,sizeof(h2));
	err3 = write(socket_fd,h,sizeof(h));
	if(err1 == -1 || err2 == -1 || err3 == -1) tell_user("mouse_send: write error");
}
